#include "HJetGen.h"
#include<complex>
typedef std::complex<double> COM;

SGeneratorFlags Flags;
double jetrapmax(4.5);
double Hrapmax(2.5);
double ysep(0.0);

COM EPSTENSOR(CLHEP::HepLorentzVector p1, CLHEP::HepLorentzVector p2, CLHEP::HepLorentzVector p3, CLHEP::HepLorentzVector p4)
{
  return -COM(0.,1.)*(p1.x()*p2.y()*p3.z()*p4.e()-p1.x()*p2.y()*p3.e()*p4.z()-p1.x()*p2.z()*p3.y()*p4.e()+p1.x()*p2.z()*p3.e()*p4.y()+p1.x()*p2.e()*p3.y()*p4.z()-p1.x()*p2.e()*p3.z()*p4.y()-p1.y()*p2.x()*p3.z()*p4.e()+p1.y()*p2.x()*p3.e()*p4.z()+p1.y()*p2.z()*p3.x()*p4.e()-p1.y()*p2.z()*p3.e()*p4.x()-p1.y()*p2.e()*p3.x()*p4.z()+p1.y()*p2.e()*p3.z()*p4.x()+p1.z()*p2.x()*p3.y()*p4.e()-p1.z()*p2.x()*p3.e()*p4.y()-p1.z()*p2.y()*p3.x()*p4.e()+p1.z()*p2.y()*p3.e()*p4.x()+p1.z()*p2.e()*p3.x()*p4.y()-p1.z()*p2.e()*p3.y()*p4.x()-p1.e()*p2.x()*p3.y()*p4.z()+p1.e()*p2.x()*p3.z()*p4.y()+p1.e()*p2.y()*p3.x()*p4.z()-p1.e()*p2.y()*p3.z()*p4.x()-p1.e()*p2.z()*p3.x()*p4.y()+p1.e()*p2.z()*p3.y()*p4.x());
}

int init(int argn, char* argv[])
{
  ifstream input;
  istringstream sline;
  int iline(0);
  string label,temp,line;

  if (argn!=2) {
    cerr << "# Usage:\n./HJetGen [input.file]\n";
    cerr << "# where input.file is a file with parameters controlling the generation of events.\n";
    cerr << "# More information at http://andersen.web.cern.ch/andersen/MJEV\n";
    if (argn>2)
      return 1;
    else {
      cout << "# Using standard values for generator parameters\n";
      return 0;
    }
  } else {
    // Now open input file:
    input.open(argv[1], ios::in); // Open input file
    
  // Read parameters in from input file:
    while(getline(input,line)) {
      line.push_back(' ');
//       cout << "#line: " << line<<endl;
      sline.str(line);
      sline>>label;
//       cout << "#label: " << label<<"#"<<endl;
      iline++;
//       cout << "#iline: " << iline<<endl;
//       cout << "#sline: " << sline.str()<<endl;
      
      if(label=="Ebeam:") {
	if(!(sline>>Flags.Ebeam)) {
	  cout << "Error in input file at line " <<iline<<endl;
	  cout << "Value of Beam Energy invalid!"<<endl;
	  return 1;
	}
	if(Flags.Ebeam<0) {
	  cout << "Beam Energy<0 at line "<<iline<<" of input file!"<<endl;
	  return 1;
	}
// 	cout << "#Flags.Ebeam: "<<Flags.Ebeam<<"#"<<endl;
      }
      else if(label=="jetptmin:") {
	if(!(sline>>Flags.jetptmin)){
	  cout<<"Error in input file at line "<<iline<<endl;
	  cout<<"Value of jetptmin invalid!"<<endl;
	  return 1;
	}
	if(Flags.jetptmin<0) {
	  cout <<"jetptmin<0 at line "<<iline<<" of input file!"<<endl;
	  return 1;
	}
// 	cout << "#Flags.jetptmin: "<<Flags.jetptmin<<"#"<<endl;
      }
      else if(label=="extpartonptmin:") {
	if(!(sline>>Flags.extpartonptmin)){
	  cout<<"Error in input file at line "<<iline<<endl;
	  cout<<"Value of extpartonptmin invalid!"<<endl;
	  return 1;
	}
	if(Flags.extpartonptmin<0) {
	  cout <<"extpartonptmin<0 at line "<<iline<<" of input file!"<<endl;
	  return 1;
	}
      }
      else if(label=="mH:") {
	if(!(sline>>Flags.mH)){
	  cout<<"Error in input file at line "<<iline<<endl;
	  cout<<"Value of mH invalid!"<<endl;
	  return 1;
	}
	if(Flags.mH<0) {
	  cout<<"mH<0 at line "<<iline<<" of input file!"<<endl;
	  return 1;
	}
      }
      else if(label=="GammaH:") {
	if(!(sline>>Flags.GammaH)){
	  cout<<"Error in input file at line "<<iline<<endl;
	  cout<<"Value of GammaH invalid!"<<endl;
	  return 1;
	}
	if(Flags.GammaH<0) {
	  cout<<"GammaH<0 at line "<<iline<<" of input file!"<<endl;
	  return 1;
	}
      }
      else if(label=="HBRgammagamma:") {
	if(!(sline>>Flags.HBRgammagamma)){
	  cout<<"Error in input file at line "<<iline<<endl;
	  cout<<"Value of HBRgammagamma invalid!"<<endl;
	  return 1;
	}
	if(Flags.HBRgammagamma<0) {
	  cout<<"HBRgammagamma<0 at line "<<iline<<" of input file!"<<endl;
	  return 1;
	}
      }
      else if(label=="DecayHiggs:") {
	if(!(sline>>temp)){
	  cout << "Error in input file at line "<<iline<<endl;
	  cout << "DecayHiggs flag invalid!"<<endl;
	  return 1;
	}
	if(temp=="Y"||temp=="y") Flags.DecayHiggs=true; else Flags.DecayHiggs=false;
      }
      else if(label=="scale:") {
	if(!(sline>>Flags.scale)){
	  cout<<"Error in input file at line "<<iline<<endl;
	  cout<<"Value of scale invalid!"<<endl;
	  return 1;
	}
	if(Flags.scale<0) {
	  cout<<"scale<0 at line "<<iline<<" of input file!"<<endl;
	  return 1;
	}
      }
      else if(label=="scalesetting:") { //0: PTMAX, 1: fixed, 2:ht/2, 3:geom mean
	if(!(sline>>Flags.scalesetting)){
	  cout<<"Error in input file at line "<<iline<<endl;
	  cout<<"Value of scalesetting invalid!"<<endl;
	  return 1;
	}
	if(abs(Flags.scalesetting)>3) {
	  cout<<"scalesetting \notin {0,1,2,3}  at line "<<iline<<" of input file!"<<endl;
	  return 1;
	}
      }
      else if(label=="scalefactor:") {
	if(!(sline>>Flags.scalefactor)){
	  cout<<"Error in input file at line "<<iline<<endl;
	  cout<<"Value of scalefactor invalid!"<<endl;
	  return 1;
	}
	if(Flags.scalefactor<0) {
	  cout<<"scalefactor <0  at line "<<iline<<" of input file!"<<endl;
	  return 1;
	}
      }
      else if(label=="logcorrect:") { 
	if(!(sline>>Flags.logcorrect)){
	  cout<<"Error in input file at line "<<iline<<endl;
	  cout<<"Value of logcorrect invalid!"<<endl;
	  return 1;
	}
	if(Flags.logcorrect!=0&&Flags.logcorrect!=1) {
	  cout<<"logcorrrect \notin {0,1}  at line "<<iline<<" of input file!"<<endl;
	  return 1;
	}
      }
      else if(label=="jetrapmax:") {
	if(!(sline>>jetrapmax)){
	  cout<<"Error in input file at line "<<iline<<endl;
	  cout<<"Value of jetrapmax invalid!"<<endl;
	  return 1;
	}
	if(jetrapmax<0) {
	  cout<<"jetrapmax<0 at line "<<iline<<" of input file!"<<endl;
	  return 1;
	}
      }
      else if(label=="Hrapmax:") {
	if(!(sline>>Hrapmax)){
	  cout<<"Error in input file at line "<<iline<<endl;
	  cout<<"Value of Hrapmax invalid!"<<endl;
	  return 1;
	}
	if(Hrapmax<0) {
	  cout<<"Hrapmax<0 at line "<<iline<<" of input file!"<<endl;
	  return 1;
	}
      }
      else if(label=="NEvents:") {
	if(!(sline>>Flags.NEvents)){
	  cout<<"Error in input file at line "<<iline<<endl;
	  cout<<"Value of NEvents invalid!"<<endl;
	  return 1;
	}
	if(Flags.NEvents<0) {
	  cout<<"NEvents<0 at line "<<iline<<" of input file!"<<endl;
	  return 1;
	}
      }
      else if(label=="rootfile:") {
	if(!(sline>>Flags.rootfile)){
	  cout << "Error in input file at line "<<iline<<endl;
	  cout << "rootfile name invalid!"<<endl;
	  return 1;
	}
      }
      else if(label=="LHAPDFsubset:") {
	if(!(sline>>Flags.LHAPDFsubset)){
	  cout<<"Error in input file at line "<<iline<<endl;
	  cout<<"Value of LHAPDFsubset invalid!"<<endl;
	  return 1;
	}
	if(Flags.LHAPDFsubset<0) {
	  cout<<"LHAPDFsubset<0 at line "<<iline<<" of input file!"<<endl;
	  return 1;
	}
      }
      else if(label=="mstwpdffile:") {
	if(!(sline>>Flags.mstwpdffile)){
	  cout << "Error in input file at line "<<iline<<endl;
	  cout << "mstwpdffile name invalid!"<<endl;
	  return 1;
	}
      }
      else if(label=="LHAPDFName:") {
	if(!(sline>>Flags.LHAPDFName)){
	  cout << "Error in input file at line "<<iline<<endl;
	  cout << "LHAPDFName name invalid!"<<endl;
	  return 1;
	}
      }
      else if(label=="useLHApdf:") {
	if(!(sline>>temp)){
	  cout << "Error in input file at line "<<iline<<endl;
	  cout << "useLHApdf flag invalid!"<<endl;
	  return 1;
	}
	if(temp=="Y"||temp=="y") Flags.useLHApdf=true; else Flags.useLHApdf=false;
      }
      else if(label=="beam1:") {
	if(!(sline>>temp)){
	  cout<<"Error in input file at line "<<iline<<endl;
	  cout<<"Beam 1 type invalid!"<<endl;
	  return 1;
	}
	if(temp=="P") Flags.beam[0]=1;
	else if(temp=="PBAR") Flags.beam[0]=-1;
	else {
	  cout << "Error in input file at line "<<iline<<endl;
	  cout << "Unrecognised beam type for beam 1!"<<endl;
	  return 1;
	}
      }
      else if(label=="beam2:") {
	if(!(sline>>temp)){
	  cout<<"Error in input file at line "<<iline<<endl;
	  cout<<"Beam 2 type invalid!"<<endl;
	  return 1;
	}
	if(temp=="P") Flags.beam[1]=1;
	else if(temp=="PBAR") Flags.beam[1]=-1;
	else {
	  cout <<"Error in input file at line "<<iline<<endl;
	  cout <<"Unrecognised beam type for beam 2!"<<endl;
	  return 1;
	}
      }
      else if(label=="jetalgo:") {
	if(!(sline>>temp)){
	  cout<<"Error in input file at line "<<iline<<endl;
	  cout<<"jetalgo type invalid!"<<endl;
	  return 1;
	}
	if(temp=="antikt") Flags.jetalgo=fastjet::antikt_algorithm;
	else if(temp=="kt") Flags.jetalgo=fastjet::kt_algorithm;
	else {
	  cout << "Error in input file at line "<<iline<<endl;
	  cout << "Unrecognised beam type for beam 1!"<<endl;
	  return 1;
	}
      }
      else if(label=="Rparam:") {
	if(!(sline>>Flags.Rparam)){
	  cout<<"Error in input file at line "<<iline<<endl;
	  cout<<"Value of Rparam invalid!"<<endl;
	  return 1;
	}
	if(Flags.Rparam<0) {
	  cout<<"Rparam<0 at line "<<iline<<" of input file!"<<endl;
	  return 1;
	}
      }
      else if(label=="LHoutput:") {
	if(!(sline>>temp)){
	  cout << "Error in input file at line "<<iline<<endl;
	  cout << "LHoutput flag invalid!"<<endl;
	  return 1;
	}
	if(temp=="Y"||temp=="y") Flags.LHoutput=true; else Flags.LHoutput=false;
      }
      else if(label=="LHoutputfile:") {
	if(!(sline>>Flags.LHoutputfile)){
	  cout << "Error in input file at line "<<iline<<endl;
	  cout << "Eventfile name invalid!"<<endl;
	  return 1;
	}
      }
      else if(label=="match:") {
	if(!(sline>>temp)){
	  cout << "Error in input file at line "<<iline<<endl;
	  cout << "match flag invalid!"<<endl;
	  return 1;
	}
	if(temp=="Y"||temp=="y") Flags.match=true; else Flags.match=false;
      }
      else if(label=="resum:") {
	if(!(sline>>temp)){
	  cout << "Error in input file at line "<<iline<<endl;
	  cout << "match flag invalid!"<<endl;
	  return 1;
	}
	if(temp=="Y"||temp=="y") Flags.resum=true; else Flags.resum=false;
      }
      else if(label=="nonFKLmatch2j:") {
	if(!(sline>>temp)){
	  cout << "Error in input file at line "<<iline<<endl;
	  cout << "match flag invalid!"<<endl;
	  return 1;
	}
	if(temp=="Y"||temp=="y") Flags.nonFKLmatch2j=true; else Flags.nonFKLmatch2j=false;
      }
      else if(label=="nonFKLmatch3j:") {
	if(!(sline>>temp)){
	  cout << "Error in input file at line "<<iline<<endl;
	  cout << "match flag invalid!"<<endl;
	  return 1;
	}
	if(temp=="Y"||temp=="y") Flags.nonFKLmatch3j=true; else Flags.nonFKLmatch3j=false;
      }
      else if(label=="rootevents:") {
	if(!(sline>>temp)){
	  cout << "Error in input file at line "<<iline<<endl;
	  cout << "rootevents flag invalid!"<<endl;
	  return 1;
	}
	if(temp=="Y"||temp=="y") Flags.rootevents=true; else Flags.rootevents=false;
      }
      else if(label=="unweight:") {
	if(!(sline>>temp)){
	  cout << "Error in input file at line "<<iline<<endl;
	  cout << "unweight flag invalid!"<<endl;
	  return 1;
	}
	if(temp=="Y"||temp=="y") Flags.unweight=true; else Flags.unweight=false;
      }
      else if(label=="raninitfile:") {
	if(!(sline>>Flags.raninitfile)){
	  cout << "Error in input file at line "<<iline<<endl;
	  cout << "raninitfile name invalid!"<<endl;
	  return 1;
	}
      }
      else if(label.at(0)==47){
      }
      else{
	cout<<"Unrecognised parameter at line "<<iline<<" in input file:"<<endl;
	cout<<line<<endl;
	return 1;
      }
    }
    
    input.close();
  }
  fastjet::Strategy strategy = fastjet::Best;
  fastjet::RecombinationScheme recomb_scheme = fastjet::E_scheme;
  Flags.jet_def=fastjet::JetDefinition(Flags.jetalgo, Flags.Rparam, recomb_scheme, strategy);

  return 0;
}

int jetcuts(CLHEP::HepLorentzVector pj1,CLHEP::HepLorentzVector pj2)
{
  int scuts(1);
  if (pj1.perp()<Flags.jetptmin)
    scuts=0;
  else if (pj2.perp()<Flags.jetptmin)
    scuts=0;
  //  else if (fabs(pj1.rapidity()-pj2.rapidity())<4.2)
//   else if (fabs(pj1.rapidity()-pj2.rapidity())<2.)
//     scuts=0;
  else if (fabs(pj1.rapidity())>jetrapmax)
    scuts=0;
  else if (fabs(pj2.rapidity())>jetrapmax)
    scuts=0;
//   else if (pj1.rapidity()*pj2.rapidity()>0.)
//     scuts=0;
  return scuts;
}

int main(int argn, char *argv[])
{
  double wt,wts; // The event weight
  double ya,yb,yH,yi;   // ypa and ypb removed from this list
  CLHEP::HepLorentzVector pH,pi;
  int np,nj; // The number of final state partons
  int njets,njetsb,njetst; // The number of final state jets
  CLHEP::HepLorentzVector ja,jb,jthird,jfourth;
  CLHEP::Ranlux64Engine rUnweight;
  
  Float_t rpp[24][4]; // four vectors for parton momenta, for use with ROOT
  Float_t rpH[4]; // the four vector for the Higgs boson, for use with ROOT
  Float_t rpg[2][4]; // the four vector for the Higgs boson, for use with ROOT
  Double_t weights[6]; // weights, including scale variations
  Int_t processtype;

  // Only simple analysis and self-consistency check:
  if ((init(argn,argv))==1) 
    return(-1);

  // Example histogram file:
  TFile hfile(Flags.rootfile.c_str(),"RECREATE","Output from HJetGen");
  TROOT simple("HJetGen","Output from HJetGen");
  TH1D *hmstatus,*hwt,*hwtwt,*hwt2,*htotal,*hya,*hyb,*hyH,*hpt1,*hpt2,*hptH,*hmgg;
  TH1D *hnjets,*hnjetsb,*hnjetst,*hng;
  TH1D *htwojets,*hthreejets;
  TH1D *havgnjets,*havgnjetsnorm;
  TH1D *hphi2,*hphi3,*hphi4,*hphi5;
  TH1D *hAC1,*hAC2,*hAC3;
  TTree * Events;
  Events = new TTree("Events","Events");
  Events->Branch("weight",&wts,"weight/D");
  Events->Branch("weights",&weights,"weights[12]/D");
  Events->Branch("nj",&nj,"nj/I");
  Events->Branch("Pfourv",&rpp,"Pfourv[nj][4]/F");
  Events->Branch("processtype",&processtype,"processtype/I");
  Events->Branch("PH",&rpH,"PH[4]/F");
  Events->Branch("Pgamma",&rpg,"Pgamma[2][4]/F");
  
  htotal = new TH1D("total","total",1,-0.5,.5);
  htotal->Sumw2();
  hmstatus = new TH1D("mstatus","mstatus",25,-.5,24.5);
  hwt = new TH1D("wt","wt",800,-60.,20.);
  //  hwtwt = new TH1D("wtwt","wtwt",3500,-30.,5.);
  hwtwt = new TH1D("wtwt","wtwt",800,-60.,20.);
  hwt2 = new TH1D("wt2","wt2",800,-60.,20.);
  //  hwtwt = new TH1D("wtwt","wtwt",3500,-30.,5.);

  hnjets = new TH1D("njets","njets dist",10,0.5,10.5);
  hnjets->Sumw2();
  hnjetsb = new TH1D("njetsb","njetsb dist",10,0.5,10.5);
  hnjetsb->Sumw2();
  hnjetst = new TH1D("njetst","njetst dist",10,0.5,10.5);
  hnjetst->Sumw2();
  hng = new TH1D("ng","ng dist",35,-.5,34.5);
  hng->Sumw2();
  hya = new TH1D("ya","ya",20,-5.,5.);
  hya->Sumw2();
  hyH = new TH1D("yH","yH",20,-5.,5.);
  hyH->Sumw2();
  hyb = new TH1D("yb","yb",20,-5.,5.);
  hyb->Sumw2();
  hpt1 = new TH1D("pt1","pt1",100,0.,1000.);
  hpt1->Sumw2();
  hpt2 = new TH1D("pt2","pt2",100,0.,1000.);
  hpt2->Sumw2();
  hptH = new TH1D("ptH","ptH",100,0.,1000.);
  hptH->Sumw2();
  htwojets = new TH1D("twojets","twojets",20,0.,10.);
  htwojets->Sumw2();
  hthreejets = new TH1D("threejets","threejets",20,0.,10.);
  hthreejets->Sumw2();
  havgnjets = new TH1D("avgnjets","avgnjets",35,0.0,7.0);
  havgnjets->Sumw2();
  havgnjetsnorm = new TH1D("avgnjetsnorm","avgnjetsnorm",35,0.0,7.0);
  havgnjetsnorm->Sumw2(); 
  hphi2 = new TH1D("phi2","phi2",62,-3.1,3.1);
  hphi2->Sumw2();
  hphi3 = new TH1D("phi3","phi3",62,-3.1,3.1);
  hphi3->Sumw2();
  hphi4 = new TH1D("phi4","phi4",62,-3.1,3.1);
  hphi4->Sumw2();
  hphi5 = new TH1D("phi5","phi5",62,-3.1,3.1);
  hphi5->Sumw2();
  hAC1 = new TH1D("AC1","AC1",1,-.5,.5);
  hAC1->Sumw2();
  hAC2 = new TH1D("AC2","AC2",1,-.5,.5);
  hAC2->Sumw2();
  hAC3 = new TH1D("AC3","AC3",1,-.5,.5);
  hAC3->Sumw2();
  hmgg = new TH1D("mgg","mgg",1000,125.,127.);
  hmgg->Sumw2();

  CHMultijet event(Flags); // use just standard settings

  Events->Branch("pdfta",&event.pdfta,"pdfta/D");
  Events->Branch("pdftb",&event.pdftb,"pdftb/D");
  // Generate the events:
  
  unsigned int evententries(0);
  for (unsigned int i=0;i<Flags.NEvents;i++) {
    wt=event.Event(); // in pico-barn
    wts=wt*1E3/double(Flags.NEvents); // in femto-barn
    hmstatus->Fill(event.getMstatus());
    if (event.getMstatus()==0) { // If event was good
      
      // Number of partons in event:
      np=event.getnp();
      
      std::vector<fastjet::PseudoJet> avec;
      
      //       std::cout<<"====================\n";
      for (int ii=0;ii<np;ii++) {
	pi=event.getpartvn(ii);
	// 	cout << pi<<endl;
	avec.push_back(pi);
      }
      //       std::cout<<"--------------------\n";
	
      
      /** fastjet find clusters */
      fastjet::ClusterSequence ev(avec, Flags.jet_def);
      
      /** Retrieve the final state jets from fastjet Sorted by Rapidity*/
      std::vector<fastjet::PseudoJet> jets = sorted_by_rapidity(ev.inclusive_jets(Flags.jetptmin));
      
      njets=int(jets.size());
      // Now find the tagging jets
      njetst=0;
      njetsb=0;
      ya=10;
      yb=-10;
      // Count how many of the fastjet jets have Et>energy_cut GeV 
      // and calculate largest distance between jets of more than ptcut2
      std::vector<fastjet::PseudoJet>::const_iterator itr = jets.begin();
      std::vector<fastjet::PseudoJet>::const_iterator jai,jbi;
      for( ; itr != jets.end() ; ++itr) {
	if (((*itr).perp()>Flags.jetptmin)&&(fabs((*itr).rapidity())<4.5)) {
	  njetsb++;
	  if ((*itr).perp()>Flags.jetptmin) {
	    njetst++;
	    yi=(*itr).rapidity();
	    if (yi<ya) { ya=yi ; ja.set((*itr)[0],(*itr)[1],(*itr)[2],(*itr)[3]);jai=itr;};
	    if (yi>yb) { yb=yi ; jb.set((*itr)[0],(*itr)[1],(*itr)[2],(*itr)[3]);jbi=itr;};
	  }
	}
      }
      
      if (njetst>=2)
	{
	  // impose jet cuts
	  int pass(1);

	  // Make list of hard jets
	  // Check that none of them are too close to the Higgs boson
	  std::vector<fastjet::PseudoJet> hardjets = sorted_by_rapidity(ev.inclusive_jets(Flags.jetptmin));
	  itr=hardjets.begin();
	  pH=event.getpHv();
	  
	  for( ; itr != hardjets.end() ; ++itr) {
	    if ((fabs((*itr).rapidity())<jetrapmax))
	      if (fabs((*itr).rapidity()-pH.rapidity())<ysep)
		pass=0;
	    if (Flags.DecayHiggs) {
	      CLHEP::HepLorentzVector phot1=event.getPhoton(0);
	      CLHEP::HepLorentzVector phot2=event.getPhoton(1);
	      if (!(fabs(phot1.rapidity())<Hrapmax&&fabs(phot2.rapidity())<Hrapmax))
		pass=0;
	    } else {
	      if (!(fabs(pH.rapidity())<Hrapmax))
		pass=0;
	    }
	  }
	  
	  if (pass) {
	    event.xsec[event.getprocesstype()][0]+=wts/1E3;
	    event.xsec[event.getprocesstype()][1]+=wts*wts/1E6;
	    if (wt>event.maxweight[event.getprocesstype()])
	      event.maxweight[event.getprocesstype()]=wt;
	    if (Flags.LHoutput) {
	      if ((Flags.unweight&&(rUnweight.flat()<wt/event.maxweight[event.getprocesstype()]))||!Flags.unweight) {
		event.LHSaveEvent();
		hwt2->Fill(log(wt));
	      }
	    }
	  
	    if (Flags.rootevents) {
	      if ((Flags.unweight&&(rUnweight.flat()<wt/event.maxweight[event.getprocesstype()]))||!Flags.unweight) {
		// 		if (np>24) std::cout << "Ouch!\n";
		int ii=0;
		std::vector<fastjet::PseudoJet>::const_iterator itr = jets.begin();
		for( ; itr != jets.end() ; ++itr) {
		  rpp[ii][0]=(*itr).e();
		  rpp[ii][1]=(*itr).px();
		  rpp[ii][2]=(*itr).py();
		  rpp[ii][3]=(*itr).pz();
		  ++ii;
		}
		for (int iii=ii;iii<22;iii++) { // reset the rest
		  rpp[iii][0]=0.;
		  rpp[iii][1]=0.;
		  rpp[iii][2]=0.;
		  rpp[iii][3]=0.;
		}

		rpH[0]=pH.e();
		rpH[1]=pH.x();
		rpH[2]=pH.y();
		rpH[3]=pH.z();
		nj=jets.size();

		processtype=event.getprocesstype();
		if (Flags.DecayHiggs) {
		  for (unsigned int iiii=0;iiii<2;iiii++) {
		    CLHEP::HepLorentzVector temp=event.getPhoton(iiii);
		    rpg[iiii][0]=temp.e();
		    rpg[iiii][1]=temp.px();
		    rpg[iiii][2]=temp.py();
		    rpg[iiii][3]=temp.pz();
		  }
		}
		std::vector<double> vweights=event.GetWeights();
		for (unsigned int ii=0;ii<vweights.size();++ii) {
		  weights[ii]=vweights.at(ii)/double(Flags.NEvents);
		  if (isnan(weights[ii])) {
		    std::cerr << "isnan : "<<ii<<std::endl;
		    std::cerr << "weight :" << wts<<std::endl;
		    std::cerr << "weights["<<ii<<"] :" << weights[ii]<<std::endl;
		  }
		}
		Events->Fill();
		++evententries;
	      }
	      if (evententries % 1000000 == 0) Events->AutoSave();
	    }
	    yH=pH.rapidity(); // Calculate the rapidity of the Higgs
		  
	    hwt->Fill(log(fabs(wt)));
	    //		    hwtwt->Fill(log(wt),wt*100.);
	    hwtwt->Fill(log(fabs(wt)),wt);
	    htotal->Fill(0.,wts);
	    hya->Fill(ya,wts*2.);
	    hyb->Fill(yb,wts*2.);
	    hyH->Fill(yH,wts*2.);
	    hpt1->Fill(ja.perp(),wts/10.);
	    hpt2->Fill(jb.perp(),wts/10.);
	    hptH->Fill(pH.perp(),wts/10.);
		    
		
	    hnjetsb->Fill(njetsb,wts);
	    hnjetst->Fill(njetst,wts);
	    hnjets->Fill(njets,wts);
	    hng->Fill(np,wts);
	    havgnjets->Fill(fabs(yb-ya),njetsb*wts*5.);
	    havgnjetsnorm->Fill(fabs(yb-ya),wts*5.);
	    if (Flags.DecayHiggs) 
	      hmgg->Fill(event.getPhoton(0).invariantMass(event.getPhoton(1)));
	    if (njetst>=2)
	      htwojets->Fill(fabs(ya-yb),2.*wts);
	    if (njetst>=3)
	      hthreejets->Fill(fabs(ya-yb),2.*wts);

	    // A little extra calculation: Add jet vectors either side of Higgs boson
	    // and check azimuthal angle between these
		  
	    CLHEP::HepLorentzVector jsa,jsb;
	    itr=hardjets.begin();
	    for( ; itr != hardjets.end() ; ++itr) {
	      CLHEP::HepLorentzVector vtemp;
	      vtemp.set((*itr)[0],(*itr)[1],(*itr)[2],(*itr)[3]);
	      if (fabs(vtemp.rapidity())<jetrapmax&&vtemp.rapidity()<pH.rapidity())
		jsa+=vtemp;
	      else if (fabs(vtemp.rapidity())<jetrapmax&&vtemp.rapidity()>pH.rapidity())
		jsb+=vtemp;
	    }
	    double phi2=acos((jsa.x()*jsb.x()+jsa.y()*jsb.y())/(sqrt(jsa.x()*jsa.x()+jsa.y()*jsa.y())*sqrt(jsb.x()*jsb.x()+jsb.y()*jsb.y())));
	    if (phi2<M_PI/4.) hAC1->Fill(0.,wts);
	    else if (phi2>M_PI/4.&&phi2<3.*M_PI/4.) hAC2->Fill(0.,wts);
	    else if (phi2>3.*M_PI/4.) hAC3->Fill(0.,wts); // always true
	    { // Calculate the parity sign for phi
	      COM epst;
	      CLHEP::HepLorentzVector p1,p2;
		    
	      p1.set(0,0,1,1);
	      p2.set(0,0,-1,1);     
	      epst=EPSTENSOR(p1,jsa,p2,jsb);
	      if (imag(epst)<0.)
		phi2*=-1.;
	    }
	    hphi2->Fill(phi2,10.*wts);
	    if (njetst>2)
	      hphi3->Fill(phi2,10.*wts);
	    if (njetst>3)
	      hphi4->Fill(phi2,10.*wts);

	    if (njetst==3) { // perform extra analysis on three-jet events
	      // Find two hardest jets:
	      std::vector<fastjet::PseudoJet> pthardjets = sorted_by_pt(ev.inclusive_jets(Flags.jetptmin));
	      // CLHEP::HepLorentzVector pj1,pj2,pj3;
	      fastjet::PseudoJet pj1,pj2,pj3;
	      pj1=pthardjets.at(0);
	      pj2=pthardjets.at(1);
	      pj3=pthardjets.at(2);
	      double ydif1,ydif2,ydif3;
	      ydif1=pj1.rapidity()-pH.rapidity();
	      ydif2=pj2.rapidity()-pH.rapidity();
	      ydif3=pj3.rapidity()-pH.rapidity();
	      // Select events where j1,j2<yh<j3 or j1,j2>yh>j3
	      if (((ydif1<-ysep&&ydif2<-ysep&&ydif3>ysep)||(ydif1>ysep&&ydif2>ysep&&ydif3<-ysep)))
		hphi5->Fill(phi2,10.*wts);
		    
	    }
	  }
	}
    }
  }
  
//   event.xsec[1][0]=htotal->GetBinContent(1);
//   event.xsec[1][1]=htotal->GetBinError(1);
  std::cout<<"Cross section passing cuts (fb): "<<htotal->GetBinContent(1)<<" +/- "<<htotal->GetBinError(1)<<std::endl;
  for (int count=1;count<=3;count++) 
    std::cout<<"Cross section passing cuts (pb): "<<event.xsec[count][0]<<" +/- "<<sqrt(event.xsec[count][1])<<std::endl;

  for (int count=1;count<=3;count++) 
    std::cout << event.maxweight[count] << " ";
  std::cout << endl;

  hfile.Write();
  hfile.Close();
  return 0;
}
