//////////////////////////////////////////////////
//////////////////////////////////////////////////
// This source code is Copyright (2012) of      //
//  Jeppe R. Andersen and Jennifer M. Smillie   //
// and is distributed under the                 //
// Gnu Public License version 2                 //
// http://www.gnu.org/licenses/gpl-2.0.html     //
// You are allowed to distribute and alter the  //
// source under the conditions of the GPLv2     //
// as long as this copyright notice             //
// is unaltered and distributed with the source //
// Any use should comply with the               //
//             MCNET GUIDELINES                 //
//    for Event Generator Authors and Users     //
// as distributed with this source code         //
//////////////////////////////////////////////////
//////////////////////////////////////////////////
#define USELHAPDF 0
#include "Jets.h"
#include "Currents.h"
#include "Tree2j.h"
#include "Tree3j.h"
#include "Tree4j.h"
// #include "Matching.h"
#include <fenv.h>
#if USELHAPDF
#include "LHAPDF/LHAPDF.h"
#endif
const double cut=-2.0; // lambda=2: -1; lambda=.2: -2
const double clambda=0.2;
const double ccut=0.2;
const double sf=.5; // lambda=2: 0.25; lambda=.2: .5
//const double cqtsqgen=50.*50.; // lambda=0.5: 50.*50.; lambda=0.2: 25.*25.
const double cqtsqgen=25.*25.; // lambda=0.5: 50.*50.; lambda=0.2: 25.*25.

#define printout 0
#define prettyplot 0
#define NewPhSp 0
#define MirrorTest 0
#define fix 0
#define SmartThingCL 1
#define sijassign 0  // whether to assign with 1 = s_ij, or 0 = pt spread
#define newreshuffle 1 // keep the rapidities of jets fixed, change pt and E
#define printalphas 0
#define dijetLO 0
#define trijetLO 0
#define calcscaleunc 1

double invmass(std::vector<fastjet::PseudoJet> * jets)
{
  fastjet::PseudoJet sum;
  std::vector<fastjet::PseudoJet>::const_iterator itr = jets->begin();
  for( ; itr != jets->end() ; ++itr)
    sum+=(*itr);
  return sum.m();
}

double geometricmean(std::vector<fastjet::PseudoJet> * jets)
{
  double gm(1.);
  std::vector<fastjet::PseudoJet>::const_iterator itr = jets->begin();
  for( ; itr != jets->end() ; ++itr)
    gm*=(*itr).perp();
  gm=pow(gm,1./jets->size());
  return gm;
}

double C2Lipatov(HepLorentzVector qav, HepLorentzVector qbv, HepLorentzVector p1, HepLorentzVector p2)
{
  HepLorentzVector temptrans=-(qav+qbv);
  HepLorentzVector p5=qav-qbv;
  HepLorentzVector CL=temptrans+(qav.m2()/p5.dot(p1)+2.*p5.dot(p2)/p1.dot(p2))*p1-p2*(qbv.m2()/p5.dot(p2)+2.*p5.dot(p1)/p1.dot(p2));
  
#if printoutput
  cout << "#Fadin qa : "<<qav<<endl;
  cout << "#Fadin qb : "<<qbv<<endl;
  cout << "#Fadin p1 : "<<p1<<endl;
  cout << "#Fadin p2 : "<<p2<<endl;
  cout << "#Fadin p5 : "<<p5<<endl;
  cout << "#Fadin Gauge Check : "<< CL.dot(p5)<<endl;
  cout << "#Fadin C2L : "<< -CL.dot(CL)<<" "<<-CL.dot(CL)/(qav.m2()*qbv.m2())/(4./p5.perp2())<<endl;
#endif
#if 0
  if (-CL.dot(CL)<0.)
    //   if (fabs(CL.dot(p5))>fabs(CL.dot(CL))) // not sufficient!
    return 0.;
  else 
#endif
    return -CL.dot(CL);
}

double C2Lipatovots(HepLorentzVector qav, HepLorentzVector qbv, HepLorentzVector p1, HepLorentzVector p2)
{
  double kperp=(qav-qbv).perp();
  if (kperp>clambda)
    return C2Lipatov(qav, qbv, p1, p2)/(qav.m2()*qbv.m2());
  else {
    double Cls=(C2Lipatov(qav, qbv, p1, p2)/(qav.m2()*qbv.m2()));
    double temp=Cls-4./(kperp*kperp);
    // std::cout <<kperp <<" "<<temp<<" "<<4./(kperp*kperp)<<" "<<(C2Lipatov(qav, qbv, pa, pb, p1, p2)/(qav.m2()*qbv.m2()))<<std::endl;
    return temp;
  }
}

double C2Lipatov(CLHEP::HepLorentzVector qav, CLHEP::HepLorentzVector qbv, CLHEP::HepLorentzVector pim, CLHEP::HepLorentzVector pip,CLHEP::HepLorentzVector pom, CLHEP::HepLorentzVector pop) // B
{
  CLHEP::HepLorentzVector temptrans=-(qav+qbv);
  CLHEP::HepLorentzVector p5=qav-qbv;
  CLHEP::HepLorentzVector CL=temptrans+qav.m2()*(1./p5.dot(pip)*pip+1./p5.dot(pop)*pop)/2.-qbv.m2()*(1./p5.dot(pim)*pim+1./p5.dot(pom)*pom)/2.+(pip*(p5.dot(pim)/pip.dot(pim)+p5.dot(pom)/pip.dot(pom)) + pop*(p5.dot(pim)/pop.dot(pim)+p5.dot(pom)/pop.dot(pom)) - pim*(p5.dot(pip)/pip.dot(pim) + p5.dot(pop)/pop.dot(pim)) - pom*(p5.dot(pip)/pip.dot(pom) + p5.dot(pop)/pop.dot(pom)) )/2.;

    return -CL.dot(CL);
}

double C2Lipatovots(CLHEP::HepLorentzVector qav, CLHEP::HepLorentzVector qbv, CLHEP::HepLorentzVector pa, CLHEP::HepLorentzVector pb, CLHEP::HepLorentzVector p1, CLHEP::HepLorentzVector p2)
{
  double kperp=(qav-qbv).perp();
  if (kperp>clambda)
    return C2Lipatov(qav, qbv, pa, pb, p1, p2)/(qav.m2()*qbv.m2());
  else {
    double Cls=(C2Lipatov(qav, qbv, pa, pb, p1, p2)/(qav.m2()*qbv.m2()));
    double temp=Cls-4./(kperp*kperp);
    // std::cout <<kperp <<" "<<temp<<" "<<4./(kperp*kperp)<<" "<<(C2Lipatov(qav, qbv, pa, pb, p1, p2)/(qav.m2()*qbv.m2()))<<std::endl;
    return temp;
  }
}

// The constructor
CMultijet::CMultijet(void) 
  // : parton("Grids/mstw2008nlo.00.dat"),
  : ran(new Ranlux64Engine())
{
  // set the default values
  lambda=clambda;
  ymax=5.5;// the maximum (and -minimum) rapidity generated for any particle
  wt05=wt2=0.;
  cout << "************************************************************\n";
  cout << "*                                                          *\n";
  cout << "*                #     #  #######        #                 *\n";
  cout << "*                #     #  #              #                 *\n";
  cout << "*                #     #  #              #                 *\n";
  cout << "*                #######  #####          #                 *\n";
  cout << "*                #     #  #        #     #                 *\n";
  cout << "*                #     #  #        #     #                 *\n";
  cout << "*                #     #  #######   #####                  *\n";
  cout << "*                                                          *\n";
  cout << "* http://cern.ch/hej                      Version 1.4.3    *\n";
  cout << "* by Jeppe R. Andersen, jeppe.andersen@cern.ch             *\n";
  cout << "*    Jennifer M. Smillie, j.m.smillie@ed.ac.uk             *\n";
  cout << "* If you use this software, please cite the following      *\n";
  cout << "* papers:                                                  *\n";
  cout << "* arXiv:1101.5394 (JHEP 1106:010,2011)                     *\n";
  cout << "* arXiv:0910.5113 (Phys.Rev. D81 (2010) 114021)            *\n";
  cout << "* arXiv:0908.2786 (JHEP 1001:039,2010)                     *\n";
  cout << "************************************************************\n";
  std::cout<<"# ";
  parton = new c_mstwpdf("Grids/mstw2008nlo.00.dat"); // default: warn=false, fatal=true
  // init alpha_s
  double FR2 = 1e0, Q0 = 1e0, mTop = 1.e10;
  InitAlphaS(parton->alphaSorder,FR2,Q0,parton->alphaSQ0,parton->mCharm,parton->mBottom,mTop);
  // Setup_Currents();
}

CMultijet::CMultijet(string filename) 
  // :  parton("Grids/mstw2008nlo.00.dat"),
  : ran(new Ranlux64Engine())
{
  // set the default values
  lambda=clambda;
  ymax=5.5;// the maximum (and -minimum) rapidity generated for any particle
  wt05=wt2=0.;
  cout << "************************************************************\n";
  cout << "*                                                          *\n";
  cout << "*                #     #  #######        #                 *\n";
  cout << "*                #     #  #              #                 *\n";
  cout << "*                #     #  #              #                 *\n";
  cout << "*                #######  #####          #                 *\n";
  cout << "*                #     #  #        #     #                 *\n";
  cout << "*                #     #  #        #     #                 *\n";
  cout << "*                #     #  #######   #####                  *\n";
  cout << "*                                                          *\n";
  cout << "* http://cern.ch/hej                      Version 1.4.3    *\n";
  cout << "* by Jeppe R. Andersen, jeppe.andersen@cern.ch             *\n";
  cout << "*    Jennifer M. Smillie, j.m.smillie@ed.ac.uk             *\n";
  cout << "* If you use this software, please cite the following      *\n";
  cout << "* papers:                                                  *\n";
  cout << "* arXiv:1101.5394 (JHEP 1106:010,2011)                     *\n";
  cout << "* arXiv:0910.5113 (Phys.Rev. D81 (2010) 114021)            *\n";
  cout << "* arXiv:0908.2786 (JHEP 1001:039,2010)                     *\n";
  cout << "************************************************************\n";
  std::cout<<"# ";
  parton = new c_mstwpdf(filename.c_str()); // default: warn=false, fatal=true
  // init alpha_s
  double FR2 = 1e0, Q0 = 1e0, mTop = 1.e10;
  InitAlphaS(parton->alphaSorder,FR2,Q0,parton->alphaSQ0,parton->mCharm,parton->mBottom,mTop);
  // Setup_Currents();

  LHoutput.open(filename.c_str());
  LHoutput << "<LesHouchesEvents version=\"1.0\">\n";
  LHoutput << "<!--\n";
  LHoutput << "File generated with HEJ Jets vers. 1.4.3\n";
  LHoutput << "-->\n";
  LHoutput << "<init>\n";
  LHoutput << "  2212   2212  7.000000E+03  7.000000E+03     0     0     7     7     4     1\n";
  LHoutput << "  1E+00  1E-02  1.000000E+00    1\n";
  LHoutput << "</init>\n";
}

CMultijet::CMultijet(SGeneratorFlags GeneratorFlagsa) 
  // : parton(GeneratorFlagsa.mstwpdffile),
  : ran(new Ranlux64Engine())
{
  flags=GeneratorFlagsa;
  // set the default values
  lambda=clambda;
  ymax=5.5;// the maximum (and -minimum) rapidity generated for any particle
  wt05=wt2=0.;
  cout << "************************************************************\n";
  cout << "*                                                          *\n";
  cout << "*                #     #  #######        #                 *\n";
  cout << "*                #     #  #              #                 *\n";
  cout << "*                #     #  #              #                 *\n";
  cout << "*                #######  #####          #                 *\n";
  cout << "*                #     #  #        #     #                 *\n";
  cout << "*                #     #  #        #     #                 *\n";
  cout << "*                #     #  #######   #####                  *\n";
  cout << "*                                                          *\n";
  cout << "* http://cern.ch/hej                      Version 1.4.3    *\n";
  cout << "* by Jeppe R. Andersen, jeppe.andersen@cern.ch             *\n";
  cout << "*    Jennifer M. Smillie, j.m.smillie@ed.ac.uk             *\n";
  cout << "* If you use this software, please cite the following      *\n";
  cout << "* papers:                                                  *\n";
  cout << "* arXiv:1101.5394 (JHEP 1106:010,2011)                     *\n";
  cout << "* arXiv:0910.5113 (Phys.Rev. D81,114021 (2010))            *\n";
  cout << "* arXiv:0908.2786 (JHEP 1001:039,2010)                     *\n";
  cout << "************************************************************\n";
#if USELHAPDF
  LHAPDF::initPDFSet(GeneratorFlagsa.LHAPDFName, LHAPDF::LHGRID, GeneratorFlagsa.LHAPDFsubset);
#endif
  std::cout<<"# ";
  parton = new c_mstwpdf(GeneratorFlagsa.mstwpdffile); // default: warn=false, fatal=true
  // init alpha_s
  double FR2 = 1e0, Q0 = 1e0, mTop = 1.e10;
  InitAlphaS(parton->alphaSorder,FR2,Q0,parton->alphaSQ0,parton->mCharm,parton->mBottom,mTop);

  if (flags.raninitfile.length()!=0) {
    cout << "# Restoring Random Engine Status from : "<<flags.raninitfile<<std::endl;
    ran.restoreEngineStatus(flags.raninitfile.c_str());
  }
  maxweight[0]=0;
  maxweight[1]=1.5E4;
  maxweight[2]=5E3;
  maxweight[3]=1E4;
  maxweight[4]=1E4;
  xsec[0][0]=xsec[1][0]=xsec[2][0]=xsec[3][0]=xsec[4][0]=0.;
  xsec[0][1]=xsec[1][1]=xsec[2][1]=xsec[3][1]=xsec[4][1]=0.;
  if (flags.LHoutput) {
    LHoutput.open(flags.LHoutputfile.c_str());
    LHoutput << "<LesHouchesEvents version=\"1.0\">\n";
    LHoutput << "<!--\n";
    LHoutput << "File generated with HEJ Jets vers. 1.4.3\n";
    LHoutput << "-->\n";
    LHoutput << "<init>\n";
    LHoutput << "  ";
    if (flags.beam[0]==1)
      LHoutput<<"2212";
    else 
      LHoutput<<"-2212";
    LHoutput << "   ";
    if (flags.beam[1]==1)
      LHoutput<<"2212 ";
    else 
      LHoutput<<"-2212 ";

    LHoutput <<scientific<<setprecision(7)<<setw(12)<<flags.Ebeam<<"  "<<flags.Ebeam<<" 0     0     7     7     ";
    if (flags.unweight)
      LHoutput << "2     ";
    else
      LHoutput << "4     ";
    // The trick is that we can first write some arbitrary data to the file, 
    // save a pointer, and then return later to overwrite the fields! 
    // lean and mean!
    xsecpos=LHoutput.tellp();
    if (flags.match) { // if matching
      LHoutput <<"3\n"; // How many processes form part of the matching?
      LHoutput << "  "<<setprecision(7)<<setw(12)<<1.21<<" "<<1E-02<<" "<<1<<" 1\n"; 
      LHoutput << "  "<<setprecision(7)<<setw(12)<<1.21<<" "<<1E-02<<" "<<1<<" 2\n"; 
      LHoutput << "  "<<setprecision(7)<<setw(12)<<1.21<<" "<<1E-02<<" "<<1<<" 3\n"; 
    } else { // if no matching
      LHoutput <<"1\n";
      LHoutput << "  "<<setprecision(7)<<setw(12)<<1.21<<" "<<1E-02<<" "<<1<<" 1\n"; 
    }
    // 

    LHoutput << "</init>\n";
  }
}

CMultijet::~CMultijet(void)
  // The destructor
{
  if(LHoutput.is_open()) {
    LHoutput << "</LesHouchesEvents>\n";

    LHoutput.seekp(xsecpos);
    int processcount;
    if (flags.match) // if matching
      processcount=nprocesstype;
    else // if no matching
      processcount=1;
    LHoutput <<processcount<<"\n"; // How many processes form part of the matching?
    for (int tcount=1;tcount<=processcount;tcount++ )
      LHoutput << "  "<<setprecision(7)<<setw(12)<<xsec[tcount][0]<<" "<<sqrt(xsec[tcount][1])<<" "<<1<<" "<<tcount<<"\n"; 
    
    LHoutput.close();
  }
  delete(parton);
}

// double CMultijet::getwt05()
// {
//   return wt05;
// }

// double CMultijet::getwt2()
// {
//   return wt2;
// }

double CMultijet::getmatchreweight()
{
  return matchreweight;
}

double CMultijet::getxa()
{
  return xa;
}

double CMultijet::getxb()
{
  return xb;
}

double CMultijet::getQa()
{
  return Qa;
}

double CMultijet::getQb()
{
  return Qb;
}

double CMultijet::getpdfta()
{
  return pdfta;
}

double CMultijet::getpdftb()
{
  return pdftb;
}

int CMultijet::getaptype()
{
  return aptype;
}

int CMultijet::getbptype()
{
  return bptype;
}

void CMultijet::LHSaveEvent(void)
{
  LHoutput << "<event>\n";
  
  if (flags.unweight==false) {
    LHoutput << " " << LHparticles.size()<< " 1 " <<scientific<<setprecision(10)<<setw(19)<<wt<<" "<<flags.scale<<" "<<1./128.<< " "<<Halphas(flags.scale)<<endl;
  }
  else {
    LHoutput << " " << LHparticles.size()<< " "<< getprocesstype()<<" " <<scientific<<setprecision(2)<<setw(4)<<1<<" "<<scientific<<setprecision(10)<<setw(19)<<flags.scale<<" "<<1./128.<< " "<<Halphas(flags.scale)<<endl;
  }
  for (unsigned int i=0;i<LHparticles.size();i++) {
    LHoutput << " " ;
    LHoutput << setw(3) << LHparticles.at(i).PID<< " " ;
    LHoutput << setw(2) << LHparticles.at(i).ISTUP<< " " ;
    LHoutput << setw(2) << LHparticles.at(i).MOT1<< " " ;
    LHoutput << setw(2) << LHparticles.at(i).MOT2<< " " ;
    LHoutput << setw(3) << LHparticles.at(i).COL1<< " " ;
    LHoutput << setw(3) << LHparticles.at(i).COL2<< " " ;
    LHoutput << scientific<<setprecision(12)<<setw(19) << LHparticles.at(i).p.x()<< " " ;
    LHoutput << scientific<<setprecision(12)<<setw(19) << LHparticles.at(i).p.y()<< " " ;
    LHoutput << scientific<<setprecision(12)<<setw(19) << LHparticles.at(i).p.z()<< " " ;
    LHoutput << scientific<<setprecision(12)<<setw(19) << LHparticles.at(i).p.e()<< " " ;
    if (LHparticles.at(i).PID!=25)
      LHoutput << scientific<<setprecision(12)<<setw(19) << 0. << " " ;
    else
      LHoutput << scientific<<setprecision(12)<<setw(19) << LHparticles.at(i).p.m()<< " " ;
    LHoutput << "0. 9.\n" ;
  }
  LHoutput << "</event>\n";
}

int CMultijet::getMstatus(void)
{
  return status;
}

int CMultijet::getstatus(void)
{
  return status;
}

void CMultijet::setstatus(int i)
{
  status=i;
}

double CMultijet::getcmsenergy(void)
{
  return 2.*flags.Ebeam;
}

int CMultijet::getnp(void)
{
  return partonmomenta.size(); // The total number of partons in final state
}

int CMultijet::getprocesstype(void)
{
  return processtype;
}

HepLorentzVector CMultijet::getpartvn(int ni)
{

  if ((ni<=int(partonmomenta.size()-1))&&(ni>=0))
    {
      return partonmomenta.at(ni);
    }
  else {
    cerr << "Parton number out of range : "<<ni<<endl;
    exit(0);
  }
}

vector<CLHEP::HepLorentzVector> CMultijet::getorigpartons(void)
{
  return origpartons;
}

vector<CLHEP::HepLorentzVector> CMultijet::getfjjets(void)
{
  return fjjets;
}

vector<CLHEP::HepLorentzVector> CMultijet::getfinaljets(void)
{
  return finaljets;
}

double CMultijet::pdfpt(int beam,double x, double uf, int pt)
{
  if (x<1e-5)
    return 0.;
  if (pt==21) // gluon
#if USELHAPDF
    {
      if (flags.useLHApdf)
	return max(0.,LHAPDF::xfx(x,uf,0));
      else
	return max(0.,CMultijet::parton->parton(0,x,uf));
    }
#else
    return max(0.,CMultijet::parton->parton(0,x,uf));
#endif
  else if (abs(pt)<7)
#if USELHAPDF
    {
      if (flags.useLHApdf)
	return max(0.,LHAPDF::xfx(x,uf,pt*flags.beam[beam]));
      else
	return max(0.,parton->parton(pt*flags.beam[beam],x,uf));
    }
#else
    return max(0.,parton->parton(pt*flags.beam[beam],x,uf));
#endif
// if (pt*flags.beam[beam]==-1) //dbar of proton, d of anti-proton
//     return max(0.,CMultijet::parton->parton(-1,x,uf));
//   else if (pt*flags.beam[beam]==1) /
//     return max(0.,CMultijet::parton->parton(1,x,uf));
//   else if (pt*flags.beam[beam]==-2)
//     return max(0.,CMultijet::parton->parton(-2,x,uf));
//   else if (pt*flags.beam[beam]==2)
//     return max(0.,CMultijet::parton->parton(2,x,uf));
//   else if (pt==3||pt==-3) {
//     return max(0.,CMultijet::parton->parton(6,x,uf));}
//   else if (pt==4||pt==-4)
//     return max(0.,CMultijet::parton->parton(5,x,uf));
//   else if (pt==5||pt==-5)
//     return max(0.,CMultijet::parton->parton(7,x,uf));
  else {
    cout << "particle type unknown: "<<pt<<endl;
    return 0.;
  }
}

double CMultijet::Halphas(double q)
{
  double as;
#if USELHAPDF
  if (flags.useLHApdf)
    as=LHAPDF::alphasPDF(q);
  else
    as=AlphaS(q);
#else
  as=AlphaS(q);
#endif
  if (isnan(as)||as>.5)
    as=0.5;
  return as;
}

double CMultijet::GenerateFKLIncomingParticles(double xa,double ufa,double xb,double ufb)
{
  double r1,lwt(1.),sum;

//   parton->mrst(xa,ufa);
//   pdfa=fabs(parton->cont.glu)+4.0/9.0*(fabs(parton->cont.upv) + fabs(parton->cont.dnv) + 2.0*(fabs(parton->cont.usea)+fabs(parton->cont.dsea)+fabs(parton->cont.str)+fabs(parton->cont.chm)+fabs(parton->cont.bot)));
//   pdfa=max(0.,parton->cont.glu)+4.0/9.0*(max(0.,parton->cont.upv+parton->cont.usea) + max(0.,parton->cont.dnv+parton->cont.dsea) + max(0.,parton->cont.usea)+max(0.,parton->cont.dsea) + 2.0*(max(0.,parton->cont.str)+max(0.,parton->cont.chm)+max(0.,parton->cont.bot)));
  pdfga=fabs(pdfpt(0,xa,ufa,21));
  pdfua=fabs(pdfpt(0,xa,ufa,2));
  pdfda=fabs(pdfpt(0,xa,ufa,1));
  pdfuxa=fabs(pdfpt(0,xa,ufa,-2));
  pdfdxa=fabs(pdfpt(0,xa,ufa,-1));
  pdfca=fabs(pdfpt(0,xa,ufa,4));
  pdfsa=fabs(pdfpt(0,xa,ufa,3));
  pdfsxa=fabs(pdfpt(0,xa,ufa,-3));
  pdfba=fabs(pdfpt(0,xa,ufa,5));

  pdfa=pdfga+4.0/9.0*(pdfua + pdfda + pdfuxa + pdfdxa + pdfsa + pdfsxa + 2.0*(pdfca +pdfba ));
  Qa=ufa;

  r1=pdfa*ran.flat();

#if 1
  if (r1<(sum=pdfga)) {
    aptype=21;
    lwt*=pdfa/pdfga;}
  else if (r1<(sum+=(4./9.)*pdfua)) {
    aptype=2;
    lwt*=pdfa/((4./9.)*pdfua); }
  else if (r1<(sum+=(4./9.)*pdfda)) {
    aptype=1;
    lwt*=pdfa/((4./9.)*pdfda); }
  else if (r1<(sum+=(4./9.)*pdfuxa)) {
    aptype=-2;
    lwt*=pdfa/((4./9.)*pdfuxa); }
  else if (r1<(sum+=(4./9.)*pdfdxa)) {
    aptype=-1;
    lwt*=pdfa/((4./9.)*pdfdxa); }
  else if (r1<(sum+=(4./9.)*pdfca)) {
    aptype=4;
    lwt*=pdfa/((4./9.)*pdfca); }
  else if (r1<(sum+=(4./9.)*pdfca)){
    aptype=-4;
    lwt*=pdfa/((4./9.)*pdfca); }
  else if (r1<(sum+=(4./9.)*pdfsa)) {
    aptype=3;
    lwt*=pdfa/((4./9.)*pdfsa); }
  else if (r1<(sum+=(4./9.)*pdfsxa)) {
    aptype=-3;
    lwt*=pdfa/((4./9.)*pdfsxa); }
  else if (r1<(sum+=(4./9.)*pdfba)) {
    aptype=5;
    lwt*=pdfa/((4./9.)*pdfba); }
  else if (r1<=(sum+=(4./9.)*pdfba)) {
    aptype=-5;
    lwt*=pdfa/((4./9.)*pdfba); }
  else {
    cout << "Error in choosing incoming parton a : "<<xa<<" "<<ufa<<" "<<sum<<" "<<pdfa<<" "<<r1;
    cout << " "<<pdfga+4./9.*(pdfua+pdfuxa+pdfda+pdfdxa+pdfsa+pdfsxa+2.*(pdfca+pdfba))<<endl;
    setstatus(15);
    return 0;
  }
#else
  aptype=21;
#endif

  pdfta=pdfpt(0,xa,ufa,aptype);
  // std::cout<<xa<<" "<<ufa<<" "<<aptype<<" "<<pdfta<<std::endl;

//   parton->mrst(xb,ufb);
//   pdfb=max(0.,parton->cont.glu)+4.0/9.0*(max(0.,parton->cont.upv+parton->cont.usea) + max(0.,parton->cont.dnv+parton->cont.dsea) + max(0.,parton->cont.usea)+max(0.,parton->cont.dsea) + 2.0*(max(0.,parton->cont.str)+max(0.,parton->cont.chm)+max(0.,parton->cont.bot)));
//   pdfb=fabs(parton->cont.glu)+4.0/9.0*(fabs(parton->cont.upv) + fabs(parton->cont.dnv) + 2.0*(fabs(parton->cont.usea)+fabs(parton->cont.dsea)+fabs(parton->cont.str)+fabs(parton->cont.chm)+fabs(parton->cont.bot)));
  pdfgb=fabs(pdfpt(1,xb,ufb,21));
  pdfub=fabs(pdfpt(1,xb,ufb,2));
  pdfdb=fabs(pdfpt(1,xb,ufb,1));
  pdfuxb=fabs(pdfpt(1,xb,ufb,-2));
  pdfdxb=fabs(pdfpt(1,xb,ufb,-1));
  pdfcb=fabs(pdfpt(1,xb,ufb,4));
  pdfsb=fabs(pdfpt(1,xb,ufb,3));
  pdfsxb=fabs(pdfpt(1,xb,ufb,-3));
  pdfbb=fabs(pdfpt(1,xb,ufb,5));

  pdfb=pdfgb+4.0/9.0*(pdfub + pdfdb + pdfuxb + pdfdxb + pdfsb + pdfsxb + 2.0*(pdfcb +pdfbb ));
  Qb=ufb;

  r1=pdfb*ran.flat();

#if 1
  if (r1<(sum=pdfgb)) {
    bptype=21;
    lwt*=pdfb/pdfgb; }
  else if (r1<(sum+=(4./9.)*pdfub)) {
    bptype=2;
    lwt*=pdfb/((4./9.)*pdfub); }
  else if (r1<(sum+=(4./9.)*pdfdb)) {
    bptype=1;
    lwt*=pdfb/((4./9.)*pdfdb); }
  else if (r1<(sum+=(4./9.)*pdfuxb)) {
    bptype=-2;
    lwt*=pdfb/((4./9.)*pdfuxb); }
  else if (r1<(sum+=(4./9.)*pdfdxb)) {
    bptype=-1;
    lwt*=pdfb/((4./9.)*pdfdxb); }
  else if (r1<(sum+=(4./9.)*pdfcb)) {
    bptype=4;
    lwt*=pdfb/((4./9.)*pdfcb); }
  else if (r1<(sum+=(4./9.)*pdfcb)){
    bptype=-4;
    lwt*=pdfb/((4./9.)*pdfcb); }
  else if (r1<(sum+=(4./9.)*pdfsb)) {
    bptype=3;
    lwt*=pdfb/((4./9.)*pdfsb); }
  else if (r1<(sum+=(4./9.)*pdfsxb)) {
    bptype=-3;
    lwt*=pdfb/((4./9.)*pdfsxb); }
  else if (r1<(sum+=(4./9.)*pdfbb)) {
    bptype=5;
    lwt*=pdfb/((4./9.)*pdfbb); }
  else if (r1<(sum+=(4./9.)*pdfbb)) {
    bptype=-5;
    lwt*=pdfb/((4./9.)*pdfbb); }
  else {
    cout << "Error in choosing incoming partons b : "<<xb<<" "<<ufb<<" "<<sum<<" "<<pdfb<<" "<<r1;
    cout << " "<<pdfgb+4./9.*(pdfub+pdfuxb+pdfdb+pdfdxb+pdfsb+pdfsxb+2.*(pdfcb+pdfbb))<<endl;
    setstatus(15);
    return 0;
  }

#else
 bptype=21;
#endif

  pdftb=pdfpt(1,xb,ufb,bptype);

  if (pdfta<1e-20||pdftb<1e-20) {
    setstatus(21);
  }
  return lwt;
}

double CMultijet::GenerateIncomingParticles(double xa,double ufa,double xb,double ufb)
{
  double r1,sum,lwt(1.);

//   parton->mrst(xa,ufa);
//   pdfa=fabs(parton->cont.glu)+(fabs(parton->cont.upv) + fabs(parton->cont.dnv) + 2.0*(fabs(parton->cont.usea)+fabs(parton->cont.dsea)+fabs(parton->cont.str)+fabs(parton->cont.chm)+fabs(parton->cont.bot)));

  pdfga=fabs(pdfpt(0,xa,ufa,21));
  pdfua=fabs(pdfpt(0,xa,ufa,2));
  pdfda=fabs(pdfpt(0,xa,ufa,1));
  pdfuxa=fabs(pdfpt(0,xa,ufa,-2));
  pdfdxa=fabs(pdfpt(0,xa,ufa,-1));
  pdfca=fabs(pdfpt(0,xa,ufa,4));
  pdfsa=fabs(pdfpt(0,xa,ufa,3));  
  pdfsxa=fabs(pdfpt(0,xa,ufa,-3));  
  pdfba=fabs(pdfpt(0,xa,ufa,5));

  pdfa=pdfga+(pdfua + pdfda + pdfuxa + pdfdxa + pdfsa + pdfsxa + 2.0*(pdfca +pdfba ));
  Qa=ufa;
#if 1
  r1=pdfa*ran.flat();

  if (r1<(sum=pdfga)) {
    aptype=21;
    lwt*=pdfa/pdfga;}
  else if (r1<(sum+=pdfua)) {
    aptype=2;
    lwt*=pdfa/(pdfua); }
  else if (r1<(sum+=pdfda)) {
    aptype=1;
    lwt*=pdfa/(pdfda); }
  else if (r1<(sum+=pdfuxa)) {
    aptype=-2;
    lwt*=pdfa/(pdfuxa); }
  else if (r1<(sum+=pdfdxa)) {
    aptype=-1;
    lwt*=pdfa/(pdfdxa); }
  else if (r1<(sum+=pdfca)) {
    aptype=4;
    lwt*=pdfa/(pdfca); }
  else if (r1<(sum+=pdfca)){
    aptype=-4;
    lwt*=pdfa/(pdfca); }
  else if (r1<(sum+=pdfsa)) {
    aptype=3;
    lwt*=pdfa/(pdfsa); }
  else if (r1<(sum+=pdfsxa)) {
    aptype=-3;
    lwt*=pdfa/(pdfsxa); }
  else if (r1<(sum+=pdfba)) {
    aptype=5;
    lwt*=pdfa/(pdfba); }
  else if (r1<(sum+=pdfba)) {
    aptype=-5;
    lwt*=pdfa/(pdfba); }
  else {
    cout << "Error in choosing incoming partons: "<<sum<<" "<<pdfa<<" "<<r1;
    cout << " "<<pdfga+(pdfua+pdfuxa+pdfda+pdfdxa+pdfsa+pdfsxa+2.*(pdfca+pdfba))<<endl;
    setstatus(15);
    return 0;
  }  
#else
  aptype=21;
#endif

  pdfta=pdfpt(0,xa,ufa,aptype);

//   parton->mrst(xb,ufb);

//   pdfb=fabs(parton->cont.glu)+(fabs(parton->cont.upv) + fabs(parton->cont.dnv) + 2.0*(fabs(parton->cont.usea)+fabs(parton->cont.dsea)+fabs(parton->cont.str)+fabs(parton->cont.chm)+fabs(parton->cont.bot)));
  pdfgb=fabs(pdfpt(1,xb,ufb,21));
  pdfub=fabs(pdfpt(1,xb,ufb,2));
  pdfdb=fabs(pdfpt(1,xb,ufb,1));
  pdfuxb=fabs(pdfpt(1,xb,ufb,-2));
  pdfdxb=fabs(pdfpt(1,xb,ufb,-1));
  pdfcb=fabs(pdfpt(1,xb,ufb,4));
  pdfsb=fabs(pdfpt(1,xb,ufb,3));
  pdfsxb=fabs(pdfpt(1,xb,ufb,-3));
  pdfbb=fabs(pdfpt(1,xb,ufb,5));  
  
  pdfb=pdfgb+(pdfub + pdfdb + pdfuxb + pdfdxb + pdfsb + pdfsxb + 2.0*(pdfcb +pdfbb ));
  Qb=ufb;

  r1=pdfb*ran.flat();

#if 1
  if (r1<(sum=pdfgb)) {
    bptype=21;
    lwt*=pdfb/pdfgb; }
  else if (r1<(sum+=pdfub)) {
    bptype=2;
    lwt*=pdfb/(pdfub); }
  else if (r1<(sum+=pdfdb)) {
    bptype=1;
    lwt*=pdfb/(pdfdb); }
  else if (r1<(sum+=pdfuxb)) {
    bptype=-2;
    lwt*=pdfb/(pdfuxb); }
  else if (r1<(sum+=pdfdxb)) {
    bptype=-1;
    lwt*=pdfb/(pdfdxb); }
  else if (r1<(sum+=pdfcb)) {
    bptype=4;
    lwt*=pdfb/(pdfcb); }
  else if (r1<(sum+=pdfcb)){
    bptype=-4;
    lwt*=pdfb/(pdfcb); }
  else if (r1<(sum+=pdfsb)) {
    bptype=3;
    lwt*=pdfb/(pdfsb); }
  else if (r1<(sum+=pdfsxb)) {
    bptype=-3;
    lwt*=pdfb/(pdfsxb); }
  else if (r1<(sum+=pdfbb)) {
    bptype=5;
    lwt*=pdfb/(pdfbb); }
  else if (r1<(sum+=pdfbb)) {
    bptype=-5;
    lwt*=pdfb/(pdfbb); }
  else {
    cout << "Error in choosing incoming partons dd: "<<sum<<" "<<pdfb<<" "<<r1;
    cout << " "<<pdfgb+(pdfub+pdfuxb+pdfdb+pdfdxb+pdfsb+pdfsxb+2.*(pdfcb+pdfbb))<<endl;
    setstatus(15);
    return 0;
  }

#else
  bptype=21;
#endif

  pdftb=pdfpt(1,xb,ufb,bptype);

  if (pdfta<1e-10||pdftb<1e-10) {
    setstatus(19);
  }
  //  std::cout<<"pdfs: "<<pdfta<<"  "<<pdftb<<"\n";
  return lwt;
}

std::vector<double> CMultijet::GetWeights(void)
{
  return weights;
}

void CMultijet::SetupLHInformation()
{
  std::list<int> vecl,vecr;

  if (aptype>0) { // if "a" is a gluon or quark
    vecl.push_front(-1);
    vecr.push_back(0);
  } else { // if a is an anti-quark
    vecl.push_front(0); 
    vecr.push_back(-1);
  }

  for (unsigned int i=1;i<partonmomenta.size()-1;i++) { // don't deal with the last gluon
    if (ran.flat()<.5) 
      vecr.push_back(i);
    else vecl.push_front(i);
  }
  
  if (bptype==21) {
    if (ran.flat()<.5) {
      vecr.push_back(partonmomenta.size()-1);
      vecl.push_front(-2);
    } else {
      vecl.push_front(partonmomenta.size()-1);
      vecr.push_back(-2);
    } 
  } else if (bptype<0) { // if b is an anti-quark
    vecr.push_back(partonmomenta.size()-1);
    vecl.push_front(-2);
  } else { // if b is a quark
    vecl.push_front(partonmomenta.size()-1);
    vecr.push_back(-2);
  }
  
  // now generate the colour connection information to go in the LH records
  int ctypea(1); // ctype: the colour start type: anti- or colour?
                 // when we swap the side of particles/anti-particles we should
                 // always start with a colour line
  
  particle tparticle; // a temp. particle
  {
    // setup vector of particles
    tparticle.PID=0;
    tparticle.COL1=0;
    tparticle.COL2=0;
    tparticle.MOT1=0;
    tparticle.MOT2=0;
    tparticle.ISTUP=0;
    tparticle.p.set(0,0,0,0);
    for (unsigned int i=0;i<partonmomenta.size()+2;i++)
      LHparticles.push_back(tparticle);
  }
  
  int ctype,col(501);
  ctype=ctypea;    // which of the colour columns should we fill?

//   cout << "LHparticles size : "<<LHparticles.size()<<endl;
  int n=vecr.front()+2;
  int np1;
  vecr.pop_front();
  while (vecr.size()!=0) {
    int np1=vecr.front()+2;
    vecr.pop_front();
//     cerr << "n, np1, ctype : "<<n<< ", "<<np1<<" "<<ctype<<endl;
    if (ctype>0) {
      LHparticles.at(n).COL1=col;
      LHparticles.at(np1).COL2=col;
    } else {
      LHparticles.at(n).COL2=col;
      LHparticles.at(np1).COL1=col;
    }
    col++;
    n=np1;
//     ctype*=-1;
  }
  
  // Now we are at the end of the right hand side.
  // Should we connect to the left hand side?
  
//   cout << "  --\n";
  if (bptype==21) {
    np1=vecl.front()+2;
    vecl.pop_front();
//     cerr << "n, np1, ctype : "<<n<< ", "<<np1<<" "<<ctype<<endl;
    if (ctype>0) {
      LHparticles.at(n).COL1=col;
      LHparticles.at(np1).COL2=col;
    } else {
      LHparticles.at(n).COL2=col;
      LHparticles.at(np1).COL1=col;
    }
    col++;
    n=np1;
  } else {
    n=vecl.front()+2;
    vecl.pop_front();
  }
//   cout << "  --\n";
  while (vecl.size()!=0) {
    int np1=vecl.front()+2;
    vecl.pop_front();
//     cerr << "n, np1, ctype : "<<n<< ", "<<np1<<" "<<ctype<<endl;
    if (ctype>0) {
      LHparticles.at(n).COL1=col;
      LHparticles.at(np1).COL2=col;
    } else {
      LHparticles.at(n).COL2=col;
      LHparticles.at(np1).COL1=col;
    }
    col++;
    n=np1;
//     ctype*=-1;
  }
  
  // if particle a is gluon we need to make another connections
  if (aptype==21) {
    // we know that if a is a gluon, then we started with particle 0 
    // and have ended with particle -1. We also know that the last colour line 
    // is opposite that of ctype - so connect from -1 to 0 with a ctype!
    if (ctype>0) {
      LHparticles.at(1).COL1=col;
      LHparticles.at(2).COL2=col;
    } else {
      LHparticles.at(1).COL2=col;
      LHparticles.at(2).COL1=col;
    }
    col++;
    n=np1;
  }
  
  // swap colours for particles of incoming type
  {  
    int itemp1,itemp2;
    itemp1=LHparticles.at(0).COL1;
    itemp2=LHparticles.at(0).COL2;
    LHparticles.at(0).COL1=itemp2;
    LHparticles.at(0).COL2=itemp1;
    itemp1=LHparticles.at(1).COL1;
    itemp2=LHparticles.at(1).COL2;
    LHparticles.at(1).COL1=itemp2;
    LHparticles.at(1).COL2=itemp1;
  }

  // incoming particles
  LHparticles.at(0).p=pb;
  LHparticles.at(0).PID=bptype;
  LHparticles.at(0).MOT1=0;
  LHparticles.at(0).MOT2=0;
  LHparticles.at(0).ISTUP=-1;
  LHparticles.at(1).p=pa;
  LHparticles.at(1).PID=aptype;
  LHparticles.at(1).MOT1=0;
  LHparticles.at(1).MOT2=0;
  LHparticles.at(1).ISTUP=-1;
    
  // outgoing particles
  for (unsigned int i=2;i<partonmomenta.size()+2;i++) { 
    LHparticles.at(i).p=partonmomenta.at(i-2);
    if (i==2)
      LHparticles.at(i).PID=aptype;
    else if (i==partonmomenta.size()+1)
      LHparticles.at(i).PID=bptype;
    else LHparticles.at(i).PID=21;
    LHparticles.at(i).MOT1=1;
    LHparticles.at(i).MOT2=2;
    LHparticles.at(i).ISTUP=1;
  }

  // // setup vector of particles
  // tparticle.PID=25;// Higgs boson
  // tparticle.COL1=0;
  // tparticle.COL2=0;
  // tparticle.MOT1=1;
  // tparticle.MOT2=2;
  // tparticle.ISTUP=1;
  // tparticle.p=getpHv();
  // LHparticles.push_back(tparticle);
}

double CMultijet::omega0(double q2, double lambda2)
{
  double result1(0.);
  if (flags.logcorrect==0) {
      result1=-Halphas(renscale)*3./M_PI*log(q2/lambda2);
      // renscale=PTMAX if necessary in weight() before Virtuals
  }
  else {   // flags.logcorrect==1
    //    result1=-Halphas(sqrt(sqrt(q2*lambda2))*flags.scalefactor)*3./M_PI*log(q2/lambda2);
    //  Instead, leave log explicitly
    result1=-Halphas(renscale)*3./M_PI*log(q2/lambda2)*(1.+Halphas(renscale)*beta0*log(pow(renscale,4)/(q2*lambda2))/(8.*M_PI));
#if printalphas
    std::cout<<"omega0: "<<Halphas(sqrt(sqrt(q2*lambda2)))<<"  "<<Halphas(renscale)*(1.+Halphas(renscale)*beta0*log(pow(renscale,4)/(q2*lambda2))/(8.*M_PI))<<"\n";
#endif
  }
  return result1;
}


void CMultijet::AddParton(double p[4])
{
  HepLorentzVector pi(p[1],p[2],p[3],p[0]);
  partonmomenta.push_back(pi);
}

double CMultijet::GeneratePHSPpoint()
// Generates a phase space point with unit weight
// ng partonmomenta, 2 extremal partons.
{
  int ng;
  double y,ysum,p[4],pt,r1,temp1,temp2,phi, maxptsofar;
  double tscale,kt[2];
//   static const double ptmax(2000.);
//   static const double ptSQgen(40.*40.);
  double omega0n,qtSQgen,qtSQ,qtmod;
//   double minv2after;
  double ptpar(40.);
  double sqrts;

  double pta[GLUEINT+2][2];
  double deltay[GLUEINT+2];
  // double phia[GLUEINT+2];

  lambda=clambda; // trouble is we don't know the value of lambda until after the whole generation
  //  lambda=.25; // trouble is we don't know the value of lambda until after the whole generation

  double wt=1.;
  sqrts=2.*flags.Ebeam;

  // The parton of lowest rapidity:
#if 1
  // Generate the transverse momentum of parton a
  temp1=atan((flags.partonptmax-flags.extpartonptmin)/ptpar);
  r1=ran.flat();
  pt=flags.extpartonptmin+ptpar*tan(r1*temp1);
  temp2=cos(r1*temp1);
  wt*=2.0*M_PI*pt*ptpar*temp1/temp2/temp2;
  Ht+=pt;
#else
  {
    double ptSQ;
    r1=ran.flat();
    ptSQ=ptSQgen/(exp(r1*log(1.+ptSQgen))-1.);
    wt*=2.*M_PI*(ptSQ)*(1.+(ptSQ)/ptSQgen)*log(1.+ptSQgen)/2.;
    pt=sqrt(ptSQ);
    renscale2=ptSQ;
  }
#endif

  phi=2.*M_PI*ran.flat(); // Random angle in x,y-plane

  y=0.;
  ysum=0.;
  // phia[0]=phi;
  
  p[1]=pt*cos(phi);
  p[2]=pt*sin(phi);
  pta[0][0]=pt*cos(phi);
  pta[0][1]=pt*sin(phi);
  deltay[0]=0.;

  kt[0]=-p[1];
  kt[1]=-p[2];


  tscale=kt[0]*kt[0]+kt[1]*kt[1];
  maxptsofar=pt;
  // Generate a rapidity span
  //  omega0n=0.6*min(omega0(tscale,lambda*lambda,renscale2),cut);
  //  if (flags.scalesetting==1||flags.scalesetting==2)
  renscale=maxptsofar*flags.scalefactor;   // This is corrected below, necessary for now.
  omega0n=min(sf*omega0(tscale,lambda*lambda),cut);
  y=log(1.-ran.flat())/omega0n;
  wt*=-1./omega0n/exp(omega0n*y); // take out the exponential factor which was included 
                                  // in the generation of y - we want this in the 
                                  // FKL evaluation
  ysum+=y;
  deltay[1]=ysum;

  // Assume we have already chosen a regulater mu
  // then generate the gluon emissions

#if dijetLO
  ng=0;    // Fix final number of gluons here
#else
  ng=int((GLUEINT)*ran.flat()); // a random int number of partonmomenta between 0 and GLUEINT-1
  wt*=double(GLUEINT);
#endif

  for (int ii=1;ii<=ng;ii++) {
    // Generate a random (weighted) qti min: 0, max: ptmaxsq 
    r1=ran.flat();
    //      qtSQgen=pt1SQ;
//     qtSQgen=12.*12.;
//    qtSQgen=15.*15.;
    // qtSQgen=20.*20.;
    qtSQgen=cqtsqgen;
    // qtSQgen=30.*30.;
    //    qtSQgen=40.*40.;
    // qtSQgen=50.*50.;
    //    qtSQgen=60.*60.;

    // qtSQ=qtSQgen/(exp(r1*log(1.+qtSQgen))-1.)-1.;
    // wt*=2.*M_PI*(qtSQ+1.)*(1.+(qtSQ+1.)/qtSQgen)*log(1.+qtSQgen)/2.;
    
    double t2=log(1.+qtSQgen);
    double t=exp(r1*t2);
    qtSQ=qtSQgen/(t-1.)-(1.-ccut*ccut);
    wt*=2.*M_PI*(qtSQ+1.-ccut*ccut)*(qtSQ+1.-ccut*ccut)*t*t2/qtSQgen/2.;

    wt/=16.*pow(M_PI,3);
    qtmod=sqrt(qtSQ);    
    Ht+=qtmod;
    if (qtmod>maxptsofar)
      maxptsofar=qtmod;

    // Generate random angle 
    phi=2.*M_PI*ran.flat();
    // phia[ii]=phi;
    
    p[1]=pta[ii][0]=qtmod*cos(phi);
    p[2]=pta[ii][1]=qtmod*sin(phi);

    // keep track of the transverse t-channel scale in the evolution
    kt[0]-=p[1];
    kt[1]-=p[2];

    tscale=kt[0]*kt[0]+kt[1]*kt[1];

    // Generate a rapidity span after the emission
    //    omega0n=0.6*min(omega0(tscale,lambda*lambda,renscale2),cut);
    //    if (flags.scalesetting==1)
    renscale=maxptsofar*flags.scalefactor;  // This is corrected below, necessary for now.

    omega0n=min(sf*omega0(tscale,lambda*lambda),cut);
    y=log(1.-ran.flat())/omega0n;
    ysum+=y;
    deltay[ii+1]=ysum;
    wt*=-1./omega0n/exp(omega0n*y); // take out the exponential factor which was included 
                                    // in the generation of y - we want this in the 
                                    // FKL evaluation
  }

  // calculate the resulting outgoing momentum of the last gluon (largest rapidity)
  pta[ng+1][0]=kt[0];
  pta[ng+1][1]=kt[1];
  // wt/=16.*pow(M_PI,3);
  // check that is is above the ptmin 
  pt=sqrt(pta[ng+1][0]*pta[ng+1][0]+pta[ng+1][1]*pta[ng+1][1]);
  if (pt<flags.extpartonptmin) {
    setstatus(8);
    wt=0.;
    return wt;
  }
  Ht+=pt;

  // Calculate rapidity centre of partons
  // Calculate shift
  {
    double yoffset;
    double ycenter=ysum/(2.);
    {
      double lninvr1,r1,r2,temp,a(1.6);
      r1=ran.flat();
      r2=ran.flat();
      lninvr1=-log(r1);
      temp=a*sqrt(2.*lninvr1)*cos(2.*M_PI*r2);
      yoffset=(temp-ycenter);
      wt=wt*(exp(temp*temp/2./a/a))*sqrt(2.*M_PI)*a;
    }
    // Reset parton rapidities
    // add to partonmomenta and xa,xb
    for (int ii=0;ii<=ng+1;ii++) {
      double yi=deltay[ii]+yoffset;
      if (fabs(yi)>ymax) {
	setstatus(2);
	wt=0.;
	return wt;
      }
      
      pt=sqrt(pta[ii][0]*pta[ii][0]+pta[ii][1]*pta[ii][1]);

      p[0]=pt*cosh(yi); // the energy component
      p[1]=pta[ii][0];
      p[2]=pta[ii][1];
      p[3]=pt*sinh(yi); // the z-component

      xa+=(p[0]-p[3])/sqrts;
      xb+=(p[0]+p[3])/sqrts;
      AddParton(p);
    }
    
  }
  pa.set(0,0,-xa*sqrts/2.,xa*sqrts/2.);
  pb.set(0,0,xb*sqrts/2.,xb*sqrts/2.);
  
#if 0
  // Check of energy and momentum conservation
  std::vector<CLHEP::HepLorentzVector>::iterator itemissions;
  HepLorentzVector outsum(0.,0.,0.,0.);
  for (itemissions=partonmomenta.begin();itemissions!=partonmomenta.end();itemissions++) {
    outsum+=*itemissions;
  }
  cout << outsum <<" "<<pa+pb<<" "<<outsum-pa-pb<<endl;
#endif
  return wt;
}

double CMultijet::MEt2(std::vector<CLHEP::HepLorentzVector> partons, CLHEP::HepLorentzVector pa, CLHEP::HepLorentzVector pb,bool checkmomenta,fastjet::ClusterSequence * ev)
{
  double wt(0.);
    // std::cout<<"Momenta in MEt2: \n";
    // std::cout<<pb<<"   "<<pa<<std::endl;

  if (aptype==21&&bptype==21) {
    wt=jM2gg(partons.back(),pb,partons.front(),pa);
  } else if (aptype==21&&bptype!=21) {
    if (bptype > 0)
      wt=jM2qg(partons.back(),pb,partons.front(),pa);
    else
      //      wt*=1;
      wt=jM2qbarg(partons.back(),pb,partons.front(),pa);
  }
  else if (bptype==21&&aptype!=21) { // ----- || -----
    if (aptype > 0) 
      wt=jM2qg(partons.front(),pa,partons.back(),pb);
    else
      wt=jM2qbarg(partons.front(),pa,partons.back(),pb);
  }
  else { // they are both quark
    if (bptype>0) {
      if (aptype>0)
	wt=jM2qQ(partons.back(),pb,partons.front(),pa);
      else
	wt=jM2qQbar(partons.back(),pb,partons.front(),pa);
    }
    else {
      if (aptype>0)
	wt=jM2qQbar(partons.front(),pa,partons.back(),pb);
      else
	wt=jM2qbarQbar(partons.back(),pb,partons.front(),pa);
    }
  }
  //  std::cout<<"  c: "<<wt;
  // Multiply by couplings etc.
  wt/=(4.*8.);    // (Nc^2-1)*numb_in_hel_states

  double gs2=0.;
  if (flags.logcorrect==0) {
      gs2=4.*M_PI*Halphas(renscale);    // renscale=PTMAX if nec. in weight() before MEt2
      double couplings=pow(gs2,2);
      wt*=couplings;
  } else {  // flags.logcorrect==1
    //    wt*=pow(4.*M_PI,2)*Halphas(partons.front().perp()*flags.scalefactor)*Halphas(partons.back().perp()*flags.scalefactor);
    //  Leave log explicitly
      
      wt*=pow(4.*M_PI*Halphas(renscale),2)*(1.-Halphas(renscale)*beta0*log(partons.front().perp()/renscale)/(2.*M_PI))*(1.-Halphas(renscale)*beta0*log(partons.back().perp()/renscale)/(2.*M_PI));
#if printalphas
      std::cout<<"extremal: "<<Halphas(partons.front().perp())<<"  "<<Halphas(renscale)*(1.-Halphas(renscale)*beta0*log(partons.front().perp()/renscale)/(2.*M_PI))<<"  "<<Halphas(partons.back().perp())<<"  "<<Halphas(renscale)*(1.-Halphas(renscale)*beta0*log(partons.back().perp()/renscale)/(2.*M_PI))<<"\n";
#endif
  }
  // Now multiply by the number of Lipatov vertices and couplings
  CLHEP::HepLorentzVector qi, qip1,g;
  qi=pb-partons.back();

  for (unsigned gcount=partons.size()-2;gcount>0;gcount--) {
    //    std::cout<<"doing extra jets"<<std::endl;
    g=partons.at(gcount);
    qip1=qi-g;

#if SmartThingCL // decide according to membership of extremal jets
    if (checkmomenta) {
      bool jetmember(false);
      // Is the parton part of the jet of lowest rapidity?
      std::vector<fastjet::PseudoJet> constituents = (*ev).constituents(jets[0]);
      for (unsigned int ll=0;ll<constituents.size();ll++)
	if (gcount==unsigned(constituents[ll].user_index())) 
	  jetmember=true;
      // Is the parton part of the jet of highest rapidity?
      constituents = (*ev).constituents(jets[jets.size()-1]);      
      for (unsigned int ll=0;ll<constituents.size();ll++)
	if (gcount==unsigned(constituents[ll].user_index())) 
	  jetmember=true;
      //      std::cout<<"  e: "<<jetmember;
      if (flags.logcorrect==1) {// include extra logs
	//	gs2=4.*M_PI*Halphas(flags.scalefactor*g.perp());
	//  Keeping log explicitly
	// renscale=PTMAX if necessary set in weight() before MEt2()
	gs2=4.*M_PI*Halphas(renscale)*(1.-Halphas(renscale)*beta0*log(g.perp()/renscale)/(2.*M_PI));
#if printalphas
      std::cout<<"middles: "<<Halphas(g.perp())<<"  "<<Halphas(renscale)*(1.-Halphas(renscale)*beta0*log(g.perp()/renscale)/(2.*M_PI))<<"\n";
#endif
      }
      if (jetmember)
	wt*=C2Lipatovots(qi,qip1,pa,pb)*3.*gs2;   // *Ca
      else
	wt*=C2Lipatovots(qi,qip1,pa,pb,partons.front(),partons.back())*3.*gs2;   // *Ca
    } else {   // Don't check momenta
      if (flags.logcorrect==1) {// in case we use a dynamic scale choice
	//	gs2=4.*M_PI*Halphas(flags.scalefactor*g.perp());
	gs2=4.*M_PI*Halphas(renscale)*(1.-Halphas(renscale)*beta0*log(g.perp()/renscale)/(2.*M_PI));
#if printalphas
	std::cout<<"middle uncheckeds: "<<Halphas(g.perp())<<"  "<<Halphas(renscale)*(1.-Halphas(renscale)*beta0*log(g.perp()/renscale)/(2.*M_PI))<<"\n";
#endif
      }
      wt*=C2Lipatovots(qi,qip1,pa,pb,partons.front(),partons.back())*3.*gs2;   // *Ca
    }
#else
    if (flags.logcorrect==1) {// in case we use a dynamic scale choice
      //      gs2=4.*M_PI*Halphas(flags.scalefactor*g.perp());
      gs2=4.*M_PI*Halphas(renscale)*(1.-Halphas(renscale)*beta0*log(g.perp()/renscale)/(2.*M_PI));
#if printalphas
	std::cout<<"not smart thing: "<<Halphas(g.perp())<<"  "<<Halphas(renscale)*(1.-Halphas(renscale)*beta0*log(g.perp()/renscale)/(2.*M_PI))<<"\n";
#endif
    }
    wt*=C2Lipatovots(qi,qip1,pa,pb)*3.*gs2;   // *Ca
#endif
    qi=qip1;
  }
  //  std::cout<<"  d: "<<wt<<"\n";
  return wt;
}

double CMultijet::VirtualCorrections()
// oh - and to all orders
// well, at least the log piece
{
  double vt(0.0);
  
  HepLorentzVector tiv,p1,p2;
  double y1,y2;
  tiv=pa-partonmomenta.front();
  
  std::vector<CLHEP::HepLorentzVector>::iterator itpartonmomenta;
  
  itpartonmomenta=partonmomenta.begin();
  p1=*(itpartonmomenta);
  y1=p1.rapidity();
  itpartonmomenta++;
  for (;itpartonmomenta!=partonmomenta.end();itpartonmomenta++) {
    p2=*(itpartonmomenta);
    y2=p2.rapidity();
    vt+=omega0(tiv.perp2(),lambda*lambda)*(y2-y1);
    tiv-=p2;
    p1=p2;
    y1=y2;
  }

  return exp(vt);
}


double CMultijet::weight()
{
  //  double temp1,temp2,pttemp[2],ptHvertex[2],r1,pt1SQ,pt2SQ,pt1,pt2,q1SQ,q2SQ,t1,t2;
  double ufa(0.),ufb(0.),ura(0.),urb(0.); // factorisation scale
  
  //  double ptH,mHperp;
  //  double Smatch,shat2,HElimit;
  double shat2;

#if printoutput
  cout << "--------------------\n";
#endif
  setstatus(0);
  wt=1;
  
#if dijetLO
  wt*=Generate2jphsppoint();
#else
#if trijetLO 
  wt*=Generate3jphsppoint();
#else
  wt*=GeneratePHSPpoint();
#endif
#endif

  if (getstatus()!=0)
    return 0.;

  shat2=xa*xa*xb*xb*pow(getcmsenergy(),4);
  if (xa>1.0||xb>1.0) {
    setstatus(5);
    wt=0.;
    return wt;
  } 

  // Cluster into jets already at this stage, so MEt2 knows whether a
  // parton is inside an extremal jet
  // Count how many hard jets we have
  std::vector<CLHEP::HepLorentzVector>::iterator itpm,ijpm;
  std::vector<fastjet::PseudoJet>::iterator fijpm;
  
  std::vector<fastjet::PseudoJet> avec;
  
  unsigned int i(0);
  for (itpm=partonmomenta.begin();itpm!=partonmomenta.end();itpm++) {
    avec.push_back(*(itpm));
    avec.at(i).set_user_index(i); // keep track of the partons. These are now numbered 
    // according to their rapidity (increasing)
    i++;
  }
  // /**set fastjet flags */
  // double Rparam = 0.6;
  // fastjet::Strategy strategy = fastjet::Best;
  // fastjet::RecombinationScheme recomb_scheme = fastjet::E_scheme;
  
  // /** Define jet for fastjet*/
  // fastjet::JetDefinition jet_def(fastjet::antikt_algorithm, Rparam, recomb_scheme, strategy);
  
  /** fastjet find clusters */
  fastjet::ClusterSequence ev(avec, flags.jet_def);
  
  /** Retrieve the final state jets from fastjet Sorted by Rapidity*/
  jets = sorted_by_rapidity(ev.inclusive_jets(flags.jetptmin));
  //
  
  if (jets.size()>=2) {
    bool jetmember1(false);
    // Check to see if the extremal partons are included in the extremal jets
    std::vector<fastjet::PseudoJet> constituents = (ev).constituents(jets[0]);
    for (unsigned int ll=0;ll<constituents.size();ll++)
      if (0==unsigned(constituents[ll].user_index())) 
	jetmember1=true;
    
    bool jetmember2(false);
    // Check to see if the extremal partons are included in the extremal jets
    constituents = (ev).constituents(jets[jets.size()-1]);
    for (unsigned int ll=0;ll<constituents.size();ll++)
      if ((partonmomenta.size()-1)==unsigned(constituents[ll].user_index())) 
	jetmember2=true;
    if (!(jetmember2&&jetmember1)) {
      wt=0;
      setstatus(18);
      return wt;
    }
  } 
  else { // we didn't even have two jets - reject outright!
    wt=0;
    setstatus(17);
    return wt;
  }

  std::vector<fastjet::PseudoJet> ptjets = sorted_by_pt(jets);  

#if calcscaleunc
  renscale=ptjets.at(0).perp()*flags.scalefactor;
  scales.push_back(renscale);
  scales.push_back(.5*renscale);
  scales.push_back(2.*renscale);
  renscale=flags.scale*flags.scalefactor;
  scales.push_back(renscale);
  scales.push_back(.5*renscale);
  scales.push_back(2.*renscale);
  renscale=Ht/2.*flags.scalefactor;
  scales.push_back(renscale);
  scales.push_back(.5*renscale);
  scales.push_back(2.*renscale);
  renscale=geometricmean(&ptjets)*flags.scalefactor;
  scales.push_back(renscale);
  scales.push_back(.5*renscale);
  scales.push_back(2.*renscale);
  renscale=invmass(&ptjets)*flags.scalefactor;
  scales.push_back(renscale);
  scales.push_back(.5*renscale);
  scales.push_back(2.*renscale);
  renscale=ptjets.at(0).perp()*flags.scalefactor;
#else
  if (flags.scalesetting==0) //max pt(jets)
    renscale=ptjets.at(0).perp()*flags.scalefactor;
  else if (flags.scalesetting==1) //fixed scale
    renscale=flags.scale;
  else if (flags.scalesetting==2) // Ht/2
    renscale=Ht/2.*flags.scalefactor;
  else if (flags.scalesetting==3) //geometric mean of pt(jets)
    renscale=geometricmean(&ptjets)*flags.scalefactor;
#endif

  ufa=ura=renscale;    // PTMAX set a few lines above if necessary
  ufb=urb=renscale;
  
  // Implement LH interface:
  // Choose incoming parton species and assign pdfs accordingly
  wt*=GenerateFKLIncomingParticles(xa,ufa,xb,ufb);
  if (status!=0) { // if we failed in picking a sensible incoming state
    wt=0.;
    return wt;
  }

  // wtmatch=wt;
  if (flags.LHoutput)
    SetupLHInformation();

  //////////////////////////////////////////////////
  // Choose lambda based on q_perps               //
  // reject phase space point if k_perp_i<lambda  //
  //////////////////////////////////////////////////
#if 0
  // choose lambda
  {
    HepLorentzVector tiv,g;
    tiv=pb-partonmomenta.back();
    
    double minqp2=tiv.perp2();
    for (int gcount=partonmomenta.size()-2;gcount>0;gcount--) {
      g=partonmomenta.at(gcount);
      tiv=tiv-g; // the t-channel momentum AFTER emission
      double temp=tiv.perp2();
      if (temp<minqp2)
	minqp2=temp;
    }
    {
      double lambda_temp;
      lambda=clambda;
      //      lambda=.25;
      lambda_temp=sqrt(minqp2)/20.;
      
      if (lambda_temp<lambda) {
	lambda=lambda_temp;
	// cout << "lambda : "<<lambda<<endl;
      }
    }
  }
#endif

  // Reject if k_perp_i<lambda; the extremal partons have to be yet harder
  // and are thus not checked here
  for (int gcount=partonmomenta.size()-2;gcount>0;gcount--) {
    if(partonmomenta.at(gcount).perp()<ccut) {
      setstatus(9);
      wt=0.;
      return wt;
    }
  }
  //  std::cout<<"a: "<<wt;
  ////////////////////////////////////////////////////////////
  // Calculate the factors making up the cross section      //
  ////////////////////////////////////////////////////////////


#if calcscaleunc
  // double MEsq05,MEsq2;
  // {
  //   double orgscale=renscale;
  //   renscale=0.5*orgscale;
  //   MEsq05=MEt2(partonmomenta,pa,pb,true,&ev);
  //   renscale=2.*orgscale;
  //   MEsq2=MEt2(partonmomenta,pa,pb,true,&ev);
  //   renscale=orgscale; // put things back to normal
  // }
    std::vector<double> MEs;
    {
      double orgscale=renscale;
      for (unsigned int ii=0;ii<scales.size();ii++) {
	renscale=scales.at(ii);
	MEs.push_back(MEt2(partonmomenta,pa,pb,true,&ev));
      }
      renscale=orgscale; // put things back to normal
    }
#endif
  double MEsq=MEt2(partonmomenta,pa,pb,true,&ev); // true: use jet-member test

#if MirrorTest
  std::cout<<"--------------------\n";
  std::cout<<"ng,aptype,bptype : "<<partonmomenta.size()<<","<<aptype<<","<<bptype<<std::endl;
  std::cout<<MEsq<<std::endl;

  // Mirror momenta:
  {
    std::vector<CLHEP::HepLorentzVector> partonmomenta2;
    CLHEP::HepLorentzVector pa2,pb2,*t;
    int typesave;
    wc=!wc;
    typesave=aptype;
    aptype=bptype;
    bptype=typesave;
    pa2.set(pb.x(),pb.y(),-pb.z(),pb.e());
    pb2.set(pa.x(),pa.y(),-pa.z(),pa.e());
  
    std::vector<CLHEP::HepLorentzVector>::iterator itemissions;
    HepLorentzVector g,vtemp;
    for (int gcount=partonmomenta.size()-1;gcount>=0;gcount--) {
      g=partonmomenta.at(gcount);
      vtemp.set(g.x(),g.y(),-g.z(),g.e());
      partonmomenta2.push_back(vtemp);
    }

    std::cout<<MEt2(partonmomenta2,pa2,pb2)<<std::endl;
  
    // Put things back to normal again:
    wc=!wc;
    typesave=aptype;
    aptype=bptype;
    bptype=typesave;
  }
  
#endif
  //  double gs2t=4.*M_PI*AlphaS(flags.scale);

  wt*=MEsq/(16.*shat2*pow(M_PI,2));

  // std::cout<<"Compare: "<<JEWEL<<"  "<<"  "<<Tree3jenu(0.5,0.5,pa,pb,partonmomenta.front(),partonmomenta.at(1),partonmomenta.back(),alepton.p,lepton.p)*gs2t*gs2t*gs2t<<"  "<<MG3(partonmomenta,pa,pb,alepton.p,lepton.p)<<std::endl;  
  // REMINDER   double MEsq=MEt2(partonmomenta,pa,pb,true,&ev); 
  //  if (aptype<0 & bptype<0) 
  //  std::cerr<<"Compare: "<<aptype<<"  "<<bptype<<"  "<<MEsq<<"  "<<Tree4j(xa,xb,pa,pb,partonmomenta.front(),partonmomenta.at(1),partonmomenta.at(2),partonmomenta.back(),flags.scale)*gs2t*gs2t*gs2t*gs2t<<"  "<<MG4(partonmomenta,pa,pb)<<std::endl;  
  //  std::cerr<<"Compare: "<<aptype<<"  "<<bptype<<"  "<<MEsq<<"  "<<MG4(partonmomenta,pa,pb)<<"  "<<MEsq/MG4(partonmomenta,pa,pb)<<std::endl;  
#if dijetLO==0&&trijetLO==0
#if calcscaleunc
  std::vector<double> virtuals;
  {
    double orgscale=renscale;
    for (unsigned int ii=0;ii<scales.size();ii++) {
      renscale=scales.at(ii);
      virtuals.push_back(VirtualCorrections());
    }
    renscale=orgscale;
  }
  // double vt05,vt2;
  // {
  //   double orgscale=renscale;
  //   renscale=0.5*orgscale;
  //   vt05=VirtualCorrections();
  //   renscale=2.*orgscale;
  //   vt2=VirtualCorrections();
  //   renscale=orgscale; // put things back to normal
  // }
#endif

  double vt=VirtualCorrections();
  wt*=vt;
  
#else
  double vt(1.),vt2(1.),vt05(1.);
#endif
  //////////////////////////////////////////////////

  if (flags.match) {
      
    int njets=int(jets.size());
    if (njets>1&&njets<5) { // if the right number of jets are detected, i.e. not too many

      // cluster all partons into the identified jets
      std::vector<CLHEP::HepLorentzVector> jetlist,fastjetlist,nojetlist;
      CLHEP::HepLorentzVector zerovec;
      zerovec.set(0.,0.,0.,0.);

      for (fijpm=jets.begin();fijpm!=jets.end();fijpm++) {
	jetlist.push_back(zerovec);
	CLHEP::HepLorentzVector tempfast;
	tempfast.set((*(fijpm)).px(),(*(fijpm)).py(),(*(fijpm)).pz(),(*(fijpm)).e());
	fastjetlist.push_back(tempfast);
      }

      // Add all the partons to either jet-list or no-jet-list

      for (unsigned int ii=0;ii<partonmomenta.size();ii++) {
	bool included(false);
	// check to see if ii is included in one of the already existing jets
	for (unsigned int jj=0;jj<jets.size();jj++) {
	  // get constituents:
	  std::vector<fastjet::PseudoJet> constituents = ev.constituents(jets[jj]);
	  for (unsigned int ll=0;ll<constituents.size();ll++) {
	    if (ii==unsigned(constituents[ll].user_index())) {
#if printout
	      CLHEP::HepLorentzVector tempfast;
	      tempfast.set(constituents[ll].px(),constituents[ll].py(),constituents[ll].pz(),constituents[ll].e());
	      std::cout << "Found one vector in jet : "<<tempfast<<std::endl;
#endif
	      // add to jet vector
	      jetlist[jj]+=partonmomenta[ii];
	      included=true;
	    }
	  }
	}
	if (!included) { // add to list of non-included particles
	  nojetlist.push_back(partonmomenta[ii]);
	}
      }

#if printout
      std::cout<< "Fastjet Jets: "<<jets.size()<<std::endl;
      std::cout<< "partonmomenta size : "<<partonmomenta.size()<<std::endl;
      std::cout<< "nojetlist size before: "<<nojetlist.size()<<std::endl;
#endif
      CLHEP::HepLorentzVector tempadd;	
      // Now cluster the rest according to min inv. mass or pt compared to existing jets
#if sijassign
      while (nojetlist.size()>0) {
	// cycle through the list of unclustered partons and the list of jets to find 
	// the pair of minimum invariant mass
	double smin=1e10;
	unsigned int ii_save(10),jj_save(10);
	for (unsigned int ii=0;ii<nojetlist.size();ii++) {
	  for (unsigned int jj=0;jj<jetlist.size();jj++) {
	    double temp;
	    temp=jetlist[jj].invariantMass2(nojetlist[ii]);
	    if (temp<smin) {
	      smin=temp;
	      ii_save=ii;
	      jj_save=jj;
	    }
	  }
	} // now we have found the smallest smin, and saved info in ii_save, jj_save;
	jetlist[jj_save]+=nojetlist[ii_save];
	nojetlist.erase(nojetlist.begin()+ii_save);
      }
#else   // Just distribute pt instead

      tempadd.set(0.,0.,0.,0.);
      for (unsigned int ii=0; ii<nojetlist.size(); ii++) {
	tempadd+=nojetlist[ii];
      }


#if printout
	CLHEP::HepLorentzVector tempadd2;
	tempadd2.set(0.,0.,0.,0.);
	for (unsigned int hh=0; hh<jetlist.size(); hh++) {
	  tempadd2+=jetlist[hh];
	}
	std::cout<<"Sum jets+left over partons = "<<tempadd+tempadd2<<"\n";
#endif
	double jetperptotal(0.);
	for (unsigned int jj=0; jj<jetlist.size(); jj++) 
	  jetperptotal+=(jetlist.at(jj)).perp();
	
	for (unsigned int kk=0; kk<jetlist.size(); kk++) {
	  jetlist[kk]+=tempadd*(jetlist.at(kk)).perp()/jetperptotal;
	}
#if printout
	tempadd2.set(0.,0.,0.,0.);
	for (unsigned int hh=0; hh<jetlist.size(); hh++) 
	  tempadd2+=jetlist[hh];
	std::cout<<"Sum jets after = "<<tempadd2<<"\n";
#endif

#endif
#if printout
      std::cout<< "nojetlist size after: "<<nojetlist.size()<<std::endl;
#endif

#if printout
      std::cout<<"FastJet versus OurJet"<<std::endl;
      std::cout<<"Length: "<<fastjetlist.size()<<"  "<<jetlist.size()<<std::endl;
      std::cout<<"Number of partons: "<< partonmomenta.size()<<std::endl;
#endif
      double xE(0.),xz(0.);
      for (int icount=0;icount<njets;icount++) {
	CLHEP::HepLorentzVector temp=fastjetlist.at(icount);
	CLHEP::HepLorentzVector diff=temp-jetlist.at(icount);
#if printout
	double difflength=diff.x()*diff.x()+diff.y()*diff.y()+diff.z()*diff.z();
	double templength=temp.x()*temp.x()+temp.y()*temp.y()+temp.z()*temp.z();
	std::cout<<"Diff Length "<<icount<<": "<<difflength<<" ratio= "<<difflength/templength<<" "<<(ev.constituents(jets[icount])).size()<<" "<<" "<<jets[icount].m2()<<" "<<jetlist.at(icount).m2()<<std::endl;
#endif
	temp=jetlist.at(icount);
#if newreshuffle
	// keep the rapidities fixed 
	double mperp=temp.perp();
	double rapidity=jets.at(icount).rapidity();
	jetlist.at(icount).set(temp.px(),temp.py(),mperp*sinh(rapidity),mperp*cosh(rapidity));
#else
	//  Put our jets onshell
	jetlist.at(icount).set(temp.x(),temp.y(),temp.z(),sqrt(temp.x()*temp.x()+temp.y()*temp.y()+temp.z()*temp.z()));
#endif
#if printout
	std::cout << "Our Jet "<<icount<<" : "<<jetlist.at(icount)<<std::endl;
	std::cout << "Fastjet Jet "<<icount<<" : "<<fastjetlist.at(icount)<<std::endl;
#endif
	//  Now we calculate new incoming particles
	xE+=(jetlist.at(icount)).e();
	xz+=(jetlist.at(icount)).z();
	  
      }
      CLHEP::HepLorentzVector npa,npb;
      npa.set(0,0,-(xE-xz)/2.,(xE-xz)/2.);
      npb.set(0,0,(xE+xz)/2.,(xE+xz)/2.);
#if printout
      std::cout<<"'pa's:"<<pa<<"  "<<npa<<std::endl;
      std::cout<<"'pb's:"<<pb<<"  "<<npb<<std::endl;
      
      // std::cout<<"New multiplier: "<<MG(jetlist,npa,npb,alepton.p,lepton.p)/MEt2(jetlist,npa,npb)<<" "<<MG(jetlist,npa,npb,alepton.p,lepton.p)<<" "<< MEt2(jetlist,npa,npb)<<" "<<JEWEL<<std::endl;
#endif

#if prettyplot
      if (partonmomenta.size() > 7) {
	std::cout<<"# Output from a 4j event from many partons\n";
	for (unsigned int ii=0; ii<jetlist.size(); ii++) {
	  //	  std::cout<<(partonmomenta.at(ii)).rapidity()<<"  "<<(partonmomenta.at(ii)).phi()<<"  "<<0<<"  "<<(fastjetlist.at(ii)).rapidity()<<"  "<<(fastjetlist.at(ii)).phi()<<"  "<<0<<"  "<<(jetlist.at(ii)).rapidity()<<"  "<<(jetlist.at(ii)).phi()<<"  "<<0<<"\n";
	  std::cout<<(partonmomenta.at(ii)).rapidity()<<"  "<<(partonmomenta.at(ii)).phi()<<"  "<<(partonmomenta.at(ii)).perp()/100.<<"  "<<(fastjetlist.at(ii)).rapidity()<<"  "<<(fastjetlist.at(ii)).phi()<<"  "<<(fastjetlist.at(ii)).perp()/100.<<"  "<<(jetlist.at(ii)).rapidity()<<"  "<<(jetlist.at(ii)).phi()<<"  "<<(jetlist.at(ii)).perp()/100.<<"\n";
	}
	for (unsigned int ii=jetlist.size(); ii<partonmomenta.size(); ii++) {
	  //	  std::cout<<(partonmomenta.at(ii)).rapidity()<<"  "<<(partonmomenta.at(ii)).phi()<<"  "<<0<<"\n";
	  std::cout<<(partonmomenta.at(ii)).rapidity()<<"  "<<(partonmomenta.at(ii)).phi()<<"  "<<(partonmomenta.at(ii)).perp()/100.<<"\n";
	}
	// origpartons=partonmomenta;
	// fjjets=fastjetlist;
	// finaljets=jetlist;
      }
#endif

      switch (jetlist.size()) {
      case 2 : 
	matchreweight=MG(jetlist,npa,npb)/MEt2(jetlist,npa,npb,false,&ev); //ev never used
	wt*=matchreweight; 
	break;
      case 3 : 
	matchreweight=MG3(jetlist,npa,npb)/MEt2(jetlist,npa,npb,false,&ev); //ev never used
	wt*=matchreweight;
      	break;
      case 4 : 
	matchreweight=MG4(jetlist,npa,npb)/MEt2(jetlist,npa,npb,false,&ev); //ev never used
	wt*=matchreweight;
      	break;
      default:
	break;
      }
    }
  }
  double pdf1=pdfta*pdftb;
  wt*=pdf1;
#if calcscaleunc
  // double pdf2,pdf05;
  // {
  //   pdf05=pdfpt(0,xa,.5*ufa,aptype)*pdfpt(1,xb,.5*ufb,bptype);
  //   pdf2=pdfpt(0,xa,2.*ufa,aptype)*pdfpt(1,xb,2.*ufb,bptype);
  // }
  std::vector<double> pdfs;
  {
    double orgscale=renscale;
    for (unsigned int ii=0;ii<scales.size();ii++) {
      renscale=scales.at(ii);
      pdfs.push_back(pdfpt(0,xa,renscale,aptype)*pdfpt(1,xb,renscale,bptype));
    }
    renscale=orgscale;
  }
#endif


  // GeV^(-2)->micro barn
  wt*=389.379292;
  // wt*=1E6; // micro barn-> pico barn

  // if (isnan(wt)) {
  //   setstatus(6);
  // }

  // if (wt<0) {
  //   setstatus(13);
  // }

#if 1
  //  if (fabs(wt)>4.85E8) { // cuts off at 20
  //  if (fabs(wt)>6.56E7) { // cuts off at 18
  // if (fabs(wt)>8.89E6) { // cuts off at 16
  if (fabs(wt)>3.27E6) { // cuts off at 15
  //  if (fabs(wt)>59875.) { // cuts off at 11
  //if(fabs(wt)>442413.392009) {  // cuts off at 13
   setstatus(7);
  }
#endif  
#if calcscaleunc
  // wt2=wt*MEsq2/MEsq*pdf2/pdf1*vt2/vt;
  // wt05=wt*MEsq05/MEsq*pdf05/pdf1*vt05/vt;
  for (unsigned int ii=0;ii<scales.size();ii++) {
    double temp=wt*MEs.at(ii)/MEsq*pdfs.at(ii)/pdf1*virtuals.at(ii)/vt;
    weights.push_back(temp);
  }
#endif
  return wt;
}

// Use CGenerator2jenu from FixedOrder with enu removed
double CMultijet::Generate2jphsppoint()
{
  double p[4];
  double sqrts=2.*flags.Ebeam;

  double temp1,temp2,pt,r1;
  double wt(1.);
  double y;
  double pta[2][2];
  double deltay[2];
  double ptpar(40.);

  CLHEP::HepLorentzVector p1,p2;

  xa=xb=0.;
  //  partons.clear();
  status=0;
#if 1 // Generate as normal
  // generate momenta
  // Generate the transverse momentum of p3
  temp1=atan((flags.partonptmax-flags.jetptmin)/ptpar);
  r1=ran.flat();
  pt=flags.extpartonptmin+ptpar*tan(r1*temp1);
  temp2=cos(r1*temp1);
  wt*=2.0*M_PI*pt*ptpar*temp1/temp2/temp2;
  Ht+=pt;

  temp1=2.*M_PI*ran.flat(); // Random angle in x,y-plane
  pta[0][0]=pt*cos(temp1);
  pta[0][1]=pt*sin(temp1);
  deltay[0]=0.;

  // Calculate the transverse momentum of p4
  pta[1][0]=-(pta[0][0]);
  pta[1][1]=-(pta[0][1]);
  Ht+=pt;

  y=ran.flat()*10;
  wt*=10.;
  deltay[1]=y;
  // Calculate rapidity centre of partons
  // Calculate shift
  {
    double ycenter=deltay[1]/2.;
    double yoffset;
    
    {
      double lninvr1,r1,r2,temp,a(2.);
      r1=ran.flat();
      r2=ran.flat();
      lninvr1=-log(r1);
      temp=a*sqrt(2.*lninvr1)*cos(2.*M_PI*r2);
      yoffset=(temp-ycenter);
      wt=wt*(exp(temp*temp/2./a/a))*sqrt(2.*M_PI)*a;
    }
    // Reset parton rapidities
    // add to gluons and xa,xb
    double yi=deltay[0]+yoffset;
    if (fabs(yi)>ymax) {
      status=2;
      wt=0.;
      return wt;
    }
    pt=sqrt(pta[0][0]*pta[0][0]+pta[0][1]*pta[0][1]);
    
    p[0]=pt*cosh(yi); // the energy component
    p[1]=pta[0][0];
    p[2]=pta[0][1];
    p[3]=pt*sinh(yi); // the z-component

    xa+=(p[0]-p[3])/sqrts;
    xb+=(p[0]+p[3])/sqrts;
    AddParton(p);
    //    p1.set(p[1],p[2],p[3],p[0]);

    yi=deltay[1]+yoffset;
    if (fabs(yi)>ymax) {
      status=2;
      wt=0.;
      return wt;
    } 
    pt=sqrt(pta[1][0]*pta[1][0]+pta[1][1]*pta[1][1]);
    
    p[0]=pt*cosh(yi); // the energy component
    p[1]=pta[1][0];
    p[2]=pta[1][1];
    p[3]=pt*sinh(yi); // the z-component
    
    AddParton(p);
    xa+=(p[0]-p[3])/sqrts;
    xb+=(p[0]+p[3])/sqrts;
    //    p2.set(p[1],p[2],p[3],p[0]);
  }

  // /**set fastjet flags */
  // double Rparam = 0.6;
  // fastjet::Strategy strategy = fastjet::Best;
  // fastjet::RecombinationScheme recomb_scheme = fastjet::E_scheme;
  
  // /** Define jet for fastjet*/
  // fastjet::JetDefinition jet_def(fastjet::antikt_algorithm, Rparam, recomb_scheme, strategy);
  
  /** fastjet find clusters */
  fastjet::ClusterSequence ev(partonmomenta, flags.jet_def);
  
  /** Retrieve the final state jets from fastjet Sorted by Pt*/
  std::vector<fastjet::PseudoJet> jets = sorted_by_pt(ev.inclusive_jets(flags.jetptmin));
  
  int njets=int(jets.size());

  //  std::cout << "Phase space njets : "<<njets<<std::endl;
  if (njets!=2) {
    status=11;
    wt=0;
    return wt;
  } 

  //  PTMAX=(jets.at(0)).perp();  
  // if (flags.scalesetting==1)
  //   renscale=(jets.at(0)).perp()*flags.scalefactor;  
  // else if (flags.scalesetting==2)
  //   renscale=geometricmean(&jets)*flags.scalefactor;
  std::vector<fastjet::PseudoJet> ptjets = sorted_by_pt(jets);  
#if calcscaleunc
  renscale=ptjets.at(0).perp()*flags.scalefactor;
  scales.push_back(renscale);
  scales.push_back(.5*renscale);
  scales.push_back(2.*renscale);
  renscale=flags.scale*flags.scalefactor;
  scales.push_back(renscale);
  scales.push_back(.5*renscale);
  scales.push_back(2.*renscale);
  renscale=Ht/2.*flags.scalefactor;
  scales.push_back(renscale);
  scales.push_back(.5*renscale);
  scales.push_back(2.*renscale);
  renscale=geometricmean(&ptjets)*flags.scalefactor;
  scales.push_back(renscale);
  scales.push_back(.5*renscale);
  scales.push_back(2.*renscale);
  renscale=invmass(&ptjets)*flags.scalefactor;
  scales.push_back(renscale);
  scales.push_back(.5*renscale);
  scales.push_back(2.*renscale);
  renscale=ptjets.at(0).perp()*flags.scalefactor;
#else
  if (flags.scalesetting==0) //max pt(jets)
    renscale=ptjets.at(0).perp()*flags.scalefactor;
  else if (flags.scalesetting==1) //fixed scale
    renscale=flags.scale;
  else if (flags.scalesetting==2) // Ht/2
    renscale=Ht/2.*flags.scalefactor;
  else if (flags.scalesetting==3) //geometric mean of pt(jets)
    renscale=geometricmean(&ptjets)*flags.scalefactor;
#endif

#else 
  p1.set(500,0,0,500);
  p2.set(-500,0, 0, 500);

  xa=500.*2./sqrts;
  xb=xa;
  wt=1.;

  p[0]=p1.e();
  p[1]=p1.x();
  p[2]=p1.y();
  p[3]=p1.z();
  AddParton(p);
  p[0]=p2.e();
  p[1]=p2.x();
  p[2]=p2.y();
  p[3]=p2.z();
  AddParton(p);

#endif
  pa.set(0,0,-xa*sqrts/2.,xa*sqrts/2.);
  pb.set(0,0,xb*sqrts/2.,xb*sqrts/2.);
  return wt;
}

double CMultijet::weight1() // The two-jet FKL cross section
{
  wt=1.;

#if 1
  wt*=Generate2jphsppoint();
  //  std::cout<<"t1: "<<wt<<"\n";
  if (status!=0) { // if we failed in generating a 2-jet configuration
    wt=0.;
    return wt;
  }
  double shat2=xa*xa*xb*xb*pow(getcmsenergy(),4);

  // now evaluate the process and setup the flavour and colour info
  if (xa>1.||xb>1.) {
    setstatus(5);
    wt=0.;
    return 0.;
  }
  // choose incoming flavours based on pdfs only
  double ufa(0.),ufb(0.);
  ufa=renscale;  // Set to PTMAX in Generate2jphsppoint() if necessary
  ufb=renscale;

  wt*=GenerateIncomingParticles(xa,ufa,xb,ufb);
  //  std::cout<<"t2: "<<wt<<"\n";
  double pmg[4][4];
  pmg[0][1]=pb.x();
  pmg[0][2]=pb.y();
  pmg[0][3]=pb.z();
  pmg[0][0]=pb.e();

  pmg[1][1]=pa.x();
  pmg[1][2]=pa.y();
  pmg[1][3]=pa.z();
  pmg[1][0]=pa.e();

  for (int ii=0;ii<2;ii++) {
    pmg[ii+2][1]=partonmomenta.at(1-ii).x();
    pmg[ii+2][2]=partonmomenta.at(1-ii).y();
    pmg[ii+2][3]=partonmomenta.at(1-ii).z();
    pmg[ii+2][0]=partonmomenta.at(1-ii).e();
  }

  wt*=Generate2JNonFKLflavourconf_and_eval(pmg);
  //  std::cout<<"t3: "<<wt<<"\n";
  wt/=(16.*shat2*pow(M_PI,2));
  double pdf1=pdfta*pdftb;
  wt*=pdf1;
  //  std::cout<<"t4: "<<wt<<"  "<<pdfta<<"  "<<pdftb<<"\n";
  wt*=389.379292;
  //  wt*=1E6; // micro barn-> pico barn
  //  std::cout<<"t5: "<<wt<<"\n";
#endif
  //std::cout<<wt<<std::endl;

#if calcscaleunc
  // {
  //   double pdf2,pdf05;
  //   pdf05=pdfpt(0,xa,.5*ufa,aptype)*pdfpt(1,xb,.5*ufb,bptype);
  //   pdf2=pdfpt(0,xa,2.*ufa,aptype)*pdfpt(1,xb,2.*ufb,bptype);
  //   wt2=wt*pdf2/pdf1*pow(Halphas(2.*renscale)/Halphas(renscale),2);
  //   wt05=wt*pdf05/pdf1*pow(Halphas(0.5*renscale)/Halphas(renscale),2);
  // }
  for (unsigned int ii=0;ii<scales.size();ii++) {
    double reweight=pdfpt(0,xa,scales.at(ii),aptype)*pdfpt(1,xb,scales.at(ii),bptype)/pdf1*pow(Halphas(scales.at(ii))/Halphas(renscale),2);
    if (flags.logcorrect==1) {
      reweight*=(1.-Halphas(scales.at(ii))*beta0*log(partonmomenta.front().perp()/scales.at(ii))/(2.*M_PI))*(1.-Halphas(scales.at(ii))*beta0*log(partonmomenta.back().perp()/scales.at(ii))/(2.*M_PI));
      reweight/=(1.-Halphas(renscale)*beta0*log(partonmomenta.front().perp()/renscale)/(2.*M_PI))*(1.-Halphas(renscale)*beta0*log(partonmomenta.back().perp()/renscale)/(2.*M_PI));
    }
    weights.push_back(wt*reweight);
  }
#endif

  return wt;
}

double CMultijet::Generate3jphsppoint()
{
  double sqrts=2.*flags.Ebeam;
  double p[4],weight(1.);

#if 1 // generate
  double temp1,temp2,pt,r1;
  double y;
  double pta[3][2];
  double deltay[3];
  double ptpar(40.);

  // generate momenta
  // Generate the transverse momentum of p1
  temp1=atan((flags.partonptmax-flags.jetptmin)/ptpar);
  r1=ran.flat();
  pt=flags.jetptmin+ptpar*tan(r1*temp1);
  temp2=cos(r1*temp1);
  weight*=2.0*M_PI*pt*ptpar*temp1/temp2/temp2;

  temp1=2.*M_PI*ran.flat(); // Random angle in x,y-plane
  // phia[0]=temp1;
  pta[0][0]=pt*cos(temp1);
  pta[0][1]=pt*sin(temp1);
  deltay[0]=0.;

  // Generate the transverse momentum of p3
  temp1=atan((flags.partonptmax-flags.jetptmin)/ptpar);
  r1=ran.flat();
  pt=flags.jetptmin+ptpar*tan(r1*temp1);
  temp2=cos(r1*temp1);
  weight*=2.0*M_PI*pt*ptpar*temp1/temp2/temp2;
  
  y=ran.flat()*10;
  weight*=10.;
  deltay[2]=y;
  temp1=2.*M_PI*ran.flat(); // Random angle in x,y-plane
  // phia[2]=temp1;
  pta[2][0]=pt*cos(temp1);
  pta[2][1]=pt*sin(temp1);

  // Calculate the transverse momentum of p2
  pta[1][0]=-(pta[0][0]+pta[2][0]);
  pta[1][1]=-(pta[0][1]+pta[2][1]);
  // Combined with 1/(16 s^2 pi^2) in weight2() gives 2^{4-4n} pi^{4-3n}
  weight/=16.*pow(M_PI,3); 
  
  y=deltay[2]*ran.flat(); // generate rapidity evenly distributed between the two other jets
  weight*=deltay[2];
  deltay[1]=y;

  // Calculate rapidity centre of partons
  // Calculate shift
  {
    double ycenter=deltay[2]/2.;
    double yoffset;
    
    {
      double lninvr1,r1,r2,temp,a(2.);
      r1=ran.flat();
      r2=ran.flat();
      lninvr1=-log(r1);
      temp=a*sqrt(2.*lninvr1)*cos(2.*M_PI*r2);
      yoffset=(temp-ycenter);
      weight=weight*(exp(temp*temp/2./a/a))*sqrt(2.*M_PI)*a;
    }
    // Reset parton rapidities
    // add to gluons and xa,xb
    double yi=deltay[0]+yoffset;
    if (fabs(yi)>ymax) {
      status=2;
      weight=0.;
      return weight;
    }
    pt=sqrt(pta[0][0]*pta[0][0]+pta[0][1]*pta[0][1]);
    Ht+=pt;
    
    p[0]=pt*cosh(yi); // the energy component
    p[1]=pta[0][0];
    p[2]=pta[0][1];
    p[3]=pt*sinh(yi); // the z-component
    
    xa+=(p[0]-p[3])/sqrts;
    xb+=(p[0]+p[3])/sqrts;
    AddParton(p);

    yi=deltay[1]+yoffset;
    if (fabs(yi)>ymax) {
      status=2;
      weight=0.;
      return weight;
    } 
    pt=sqrt(pta[1][0]*pta[1][0]+pta[1][1]*pta[1][1]);
    Ht+=pt;
    
    p[0]=pt*cosh(yi); // the energy component
    p[1]=pta[1][0];
    p[2]=pta[1][1];
    p[3]=pt*sinh(yi); // the z-component
    
    xa+=(p[0]-p[3])/sqrts;
    xb+=(p[0]+p[3])/sqrts;
    AddParton(p);
    
    yi=deltay[2]+yoffset;
    if (fabs(yi)>ymax) {
      status=2;
      weight=0.;
      return weight;
    } 
    pt=sqrt(pta[2][0]*pta[2][0]+pta[2][1]*pta[2][1]);
    Ht+=pt;

    p[0]=pt*cosh(yi); // the energy component
    p[1]=pta[2][0];
    p[2]=pta[2][1];
    p[3]=pt*sinh(yi); // the z-component
    
    xa+=(p[0]-p[3])/sqrts;
    xb+=(p[0]+p[3])/sqrts;

    AddParton(p);

  }
  //     cout <<"--------------------\n";
  //     cout << p1 << endl<<p2<<endl;
  
  // //   std::cout << "Phase space xa, xb : "<<xa<<" , "<<xb<<std::endl;
  if (xa>1.||xb>1.) {
    status=1;
    weight=0.;
    return 0.;
  }
#else
  xa=500.*2./sqrts;
  xb=xa;
  //  p1.set(1000./3.,0.,0.,1000./3.);
  p[0]=1000./3.; // the energy component
  p[1]=1000./3.;
  p[2]=0.;
  p[3]=0.; // the z-component
  AddParton(p);
  //  p2.set(-500.,sqrt(3.)*500,0,1000./3.);
  p[0]=1000./3.;
  p[1]=-500./3.;
  p[2]=sqrt(3.)*500/3.;
  p[3]=0.;
  AddParton(p);
  //  p3.set(-500.,-sqrt(3.)*500,0,1000./3.);
  p[0]=1000./3.;
  p[1]=-500./3.;
  p[2]=-sqrt(3.)*500/3.;
  p[3]=0.;
  AddParton(p);

#endif
  pa.set(0,0,-xa*sqrts/2.,xa*sqrts/2.);
  pb.set(0,0,xb*sqrts/2.,xb*sqrts/2.);
    //    cout<< "----------\n" << pa<<"\n"<<pb<<"\n"<<pW<<"\n"<<p1<<"\n"<<p2<<"\n";
  
  // /**set fastjet flags */
  // double Rparam = 0.6;
  // fastjet::Strategy strategy = fastjet::Best;
  // fastjet::RecombinationScheme recomb_scheme = fastjet::E_scheme;
    
  // /** Define jet for fastjet*/
  // fastjet::JetDefinition jet_def(fastjet::antikt_algorithm, Rparam, recomb_scheme, strategy);
    
  /** fastjet find clusters */
  fastjet::ClusterSequence ev(partonmomenta, flags.jet_def);

  /** Retrieve the final state jets from fastjet Sorted by Rapidity */
  std::vector<fastjet::PseudoJet> jets = sorted_by_pt(ev.inclusive_jets(flags.jetptmin));
    
  /** Save the number of final state jets */
  int njets=int(jets.size());

  //   std::cout << "Phase space njets : "<<njets<<std::endl;
  if (njets!=3) {
    status=11;
    weight=0;
    return weight;
  } 
    
  //  PTMAX=(jets.at(0)).perp();  
  // if (flags.scalesetting==1)
  //   renscale=(jets.at(0)).perp()*flags.scalefactor;  
  // else if (flags.scalesetting==2)
  //   renscale=geometricmean(&jets)*flags.scalefactor;
  std::vector<fastjet::PseudoJet> ptjets = sorted_by_pt(jets);  
#if calcscaleunc
  renscale=ptjets.at(0).perp()*flags.scalefactor;
  scales.push_back(renscale);
  scales.push_back(.5*renscale);
  scales.push_back(2.*renscale);
  renscale=flags.scale*flags.scalefactor;
  scales.push_back(renscale);
  scales.push_back(.5*renscale);
  scales.push_back(2.*renscale);
  renscale=Ht/2.*flags.scalefactor;
  scales.push_back(renscale);
  scales.push_back(.5*renscale);
  scales.push_back(2.*renscale);
  renscale=geometricmean(&ptjets)*flags.scalefactor;
  scales.push_back(renscale);
  scales.push_back(.5*renscale);
  scales.push_back(2.*renscale);
  renscale=invmass(&ptjets)*flags.scalefactor;
  scales.push_back(renscale);
  scales.push_back(.5*renscale);
  scales.push_back(2.*renscale);
  renscale=ptjets.at(0).perp()*flags.scalefactor;
#else
  if (flags.scalesetting==0) //max pt(jets)
    renscale=ptjets.at(0).perp()*flags.scalefactor;
  else if (flags.scalesetting==1) //fixed scale
    renscale=flags.scale;
  else if (flags.scalesetting==2) // Ht/2
    renscale=Ht/2.*flags.scalefactor;
  else if (flags.scalesetting==3) //geometric mean of pt(jets)
    renscale=geometricmean(&ptjets)*flags.scalefactor;
#endif


  //   std::cout << "Phase space weight : "<<weight<<std::endl;
  return weight; 
}

double CMultijet::weight2() // The three-jet cross section
{
  wt=1.;
  wt*=Generate3jphsppoint();

  if (status!=0) { // if we failed in generating a 3-jet configuration
    wt=0.;
    return wt;
  }

  double shat2=xa*xa*xb*xb*pow(getcmsenergy(),4);

  // now evaluate the process and setup the flavour and colour info
  if (xa>1.||xb>1.) {
    setstatus(5);
    wt=0.;
    return 0.;
  }
  // choose incoming flavours based on pdfs only
  double ufa(0.),ufb(0.);

  ufa=renscale; // Set to PTMAX in Generate3jphsppoint() if necessary
  ufb=renscale;

  wt*=GenerateIncomingParticles(xa,ufa,xb,ufb);
  double pmg[5][4];
  pmg[0][1]=pb.x();
  pmg[0][2]=pb.y();
  pmg[0][3]=pb.z();
  pmg[0][0]=pb.e();

  pmg[1][1]=pa.x();
  pmg[1][2]=pa.y();
  pmg[1][3]=pa.z();
  pmg[1][0]=pa.e();

  for (int ii=0;ii<3;ii++) {
    pmg[ii+2][1]=partonmomenta.at(2-ii).x();
    pmg[ii+2][2]=partonmomenta.at(2-ii).y();
    pmg[ii+2][3]=partonmomenta.at(2-ii).z();
    pmg[ii+2][0]=partonmomenta.at(2-ii).e();
  }

  wt*=Generate3JNonFKLflavourconf_and_eval(pmg);

  wt/=(16.*shat2*pow(M_PI,2));
  double pdf1=pdfta*pdftb;
  wt*=pdf1;
  wt*=389.379292;
  //  wt*=1E6; // micro barn-> pico barn

#if calcscaleunc
  // {
  //   double pdf2,pdf05;
  //   pdf05=pdfpt(0,xa,.5*ufa,aptype)*pdfpt(1,xb,.5*ufb,bptype);
  //   pdf2=pdfpt(0,xa,2.*ufa,aptype)*pdfpt(1,xb,2.*ufb,bptype);
  //   wt2=wt*pdf2/pdf1*pow(Halphas(2.*renscale)/Halphas(renscale),3);
  //   wt05=wt*pdf05/pdf1*pow(Halphas(0.5*renscale)/Halphas(renscale),3);
  // }
  for (unsigned int ii=0;ii<scales.size();ii++) {
    double reweight=pdfpt(0,xa,scales.at(ii),aptype)*pdfpt(1,xb,scales.at(ii),bptype)/pdf1*pow(Halphas(scales.at(ii))/Halphas(renscale),3);
    if (flags.logcorrect==1) {
      reweight*=(1.-Halphas(scales.at(ii))*beta0*log(partonmomenta.front().perp()/scales.at(ii))/(2.*M_PI))*(1.-Halphas(scales.at(ii))*beta0*log(partonmomenta.back().perp()/scales.at(ii))/(2.*M_PI))*(1.-Halphas(scales.at(ii))*beta0*log(partonmomenta.at(1).perp()/scales.at(ii))/(2.*M_PI));
      reweight/=(1.-Halphas(renscale)*beta0*log(partonmomenta.front().perp()/renscale)/(2.*M_PI))*(1.-Halphas(renscale)*beta0*log(partonmomenta.back().perp()/renscale)/(2.*M_PI))*(1.-Halphas(renscale)*beta0*log(partonmomenta.at(1).perp()/renscale)/(2.*M_PI));
    }
    weights.push_back(wt*reweight);
  }
#endif

  return wt;
}

double CMultijet::Generate4jphsppoint()
{
  double weight(1.);
  double sqrts=2.*flags.Ebeam;
  double p[4];
#if 1
  double temp1,temp2,pt,r1;
  double y;
  double pta[4][2];
  double deltay[4];
  double ptpar(40.);

  // generate momenta
  // Generate the transverse momentum of p1
  temp1=atan((flags.partonptmax-flags.jetptmin)/ptpar);
  r1=ran.flat();
  pt=flags.jetptmin+ptpar*tan(r1*temp1);
  temp2=cos(r1*temp1);
  weight*=2.0*M_PI*pt*ptpar*temp1/temp2/temp2;

  temp1=2.*M_PI*ran.flat(); // Random angle in x,y-plane
  // phia[0]=temp1;
  pta[0][0]=pt*cos(temp1);
  pta[0][1]=pt*sin(temp1);
  deltay[0]=0.;

  // Generate the transverse momentum of p4
  temp1=atan((flags.partonptmax-flags.jetptmin)/ptpar);
  r1=ran.flat();
  pt=flags.jetptmin+ptpar*tan(r1*temp1);
  temp2=cos(r1*temp1);
  weight*=2.0*M_PI*pt*ptpar*temp1/temp2/temp2;
  
  y=ran.flat()*10;
  weight*=10.;
  deltay[3]=y;
  temp1=2.*M_PI*ran.flat(); // Random angle in x,y-plane
  // phia[3]=temp1;
  pta[3][0]=pt*cos(temp1);
  pta[3][1]=pt*sin(temp1);

  // Generate the transverse momentum of p3
  temp1=atan((flags.partonptmax-flags.jetptmin)/ptpar);
  r1=ran.flat();
  pt=flags.jetptmin+ptpar*tan(r1*temp1);
  temp2=cos(r1*temp1);
  weight*=2.0*M_PI*pt*ptpar*temp1/temp2/temp2;
  weight/=16.*pow(M_PI,3);
  
  y=deltay[3]*ran.flat(); // generate rapidity evenly distributed between the two other jets
  weight*=deltay[3];
  deltay[2]=y;
  temp1=2.*M_PI*ran.flat(); // Random angle in x,y-plane
  // phia[2]=temp1;
  pta[2][0]=pt*cos(temp1);
  pta[2][1]=pt*sin(temp1);

  // Calculate the transverse momentum of p2
  pta[1][0]=-(pta[0][0]+pta[2][0]+pta[3][0]);
  pta[1][1]=-(pta[0][1]+pta[2][1]+pta[3][1]);
  weight/=16.*pow(M_PI,3);
  
  y=deltay[2]*ran.flat(); // generate rapidity evenly distributed between the two other jets
  weight*=deltay[2];
  deltay[1]=y;


  // Calculate rapidity centre of partons
  // Calculate shift
  {
    double ycenter=deltay[3]/2.;
    double yoffset;
    
    {
      double lninvr1,r1,r2,temp,a(2.);
      r1=ran.flat();
      r2=ran.flat();
      lninvr1=-log(r1);
      temp=a*sqrt(2.*lninvr1)*cos(2.*M_PI*r2);
      yoffset=(temp-ycenter);
      weight=weight*(exp(temp*temp/2./a/a))*sqrt(2.*M_PI)*a;
    }
    // Reset parton rapidities
    // add to gluons and xa,xb
    double yi=deltay[0]+yoffset;
    if (fabs(yi)>ymax) {
      status=2;
      weight=0.;
      return weight;
    }
    pt=sqrt(pta[0][0]*pta[0][0]+pta[0][1]*pta[0][1]);
    Ht+=pt;
    
    p[0]=pt*cosh(yi); // the energy component
    p[1]=pta[0][0];
    p[2]=pta[0][1];
    p[3]=pt*sinh(yi); // the z-component
    
    xa+=(p[0]-p[3])/sqrts;
    xb+=(p[0]+p[3])/sqrts;
    AddParton(p);

    yi=deltay[1]+yoffset;
    if (fabs(yi)>ymax) {
      status=2;
      weight=0.;
      return weight;
    } 
    pt=sqrt(pta[1][0]*pta[1][0]+pta[1][1]*pta[1][1]);
    Ht+=pt;
    
    p[0]=pt*cosh(yi); // the energy component
    p[1]=pta[1][0];
    p[2]=pta[1][1];
    p[3]=pt*sinh(yi); // the z-component
    
    xa+=(p[0]-p[3])/sqrts;
    xb+=(p[0]+p[3])/sqrts;
    AddParton(p);
    
    yi=deltay[2]+yoffset;
    if (fabs(yi)>ymax) {
      status=2;
      weight=0.;
      return weight;
    } 
    pt=sqrt(pta[2][0]*pta[2][0]+pta[2][1]*pta[2][1]);
    Ht+=pt;
    
    p[0]=pt*cosh(yi); // the energy component
    p[1]=pta[2][0];
    p[2]=pta[2][1];
    p[3]=pt*sinh(yi); // the z-component
    
    xa+=(p[0]-p[3])/sqrts;
    xb+=(p[0]+p[3])/sqrts;

    AddParton(p);

    yi=deltay[3]+yoffset;
    if (fabs(yi)>ymax) {
      status=2;
      weight=0.;
      return weight;
    } 
    pt=sqrt(pta[3][0]*pta[3][0]+pta[3][1]*pta[3][1]);
    Ht+=pt;
    
    p[0]=pt*cosh(yi); // the energy component
    p[1]=pta[3][0];
    p[2]=pta[3][1];
    p[3]=pt*sinh(yi); // the z-component
    
    xa+=(p[0]-p[3])/sqrts;
    xb+=(p[0]+p[3])/sqrts;
    AddParton(p);
  }
  //     cout <<"--------------------\n";
  //     cout << p1 << endl<<p2<<endl;
  
  // //   std::cout << "Phase space xa, xb : "<<xa<<" , "<<xb<<std::endl;
  if (xa>1.||xb>1.) {
    status=1;
    weight=0.;
    return 0.;
  }
#else
  xa=500.*2./sqrts;
  xb=xa;

#endif
  pa.set(0,0,-xa*sqrts/2.,xa*sqrts/2.);
  pb.set(0,0,xb*sqrts/2.,xb*sqrts/2.);

  // Check if it is a 4-jet phase space point:
  /**set fastjet flags */
  // double Rparam = 0.6;
  // fastjet::Strategy strategy = fastjet::Best;
  // fastjet::RecombinationScheme recomb_scheme = fastjet::E_scheme;
    
  // /** Define jet for fastjet*/
  // fastjet::JetDefinition jet_def(fastjet::antikt_algorithm, Rparam, recomb_scheme, strategy);
    
  /** fastjet find clusters */
  fastjet::ClusterSequence ev(partonmomenta, flags.jet_def);
    
  /** Retrieve the final state jets from fastjet Sorted by Rapidity*/
  std::vector<fastjet::PseudoJet> jets = sorted_by_pt(ev.inclusive_jets(flags.jetptmin));
    
  /** Save the number of final state jets */
  int njets=int(jets.size());

//   std::cout << "Phase space njets : "<<njets<<std::endl;
  if (njets!=4) {
    status=11;
    weight=0;
    return weight;
  } 

  //  PTMAX=(jets.at(0)).perp();  
  // if (flags.scalesetting==1)
  //   renscale=(jets.at(0)).perp()*flags.scalefactor;  
  // else if (flags.scalesetting==2)
  //   renscale=geometricmean(&jets)*flags.scalefactor;
  std::vector<fastjet::PseudoJet> ptjets = sorted_by_pt(jets);  
#if calcscaleunc
  renscale=ptjets.at(0).perp()*flags.scalefactor;
  scales.push_back(renscale);
  scales.push_back(.5*renscale);
  scales.push_back(2.*renscale);
  renscale=flags.scale*flags.scalefactor;
  scales.push_back(renscale);
  scales.push_back(.5*renscale);
  scales.push_back(2.*renscale);
  renscale=Ht/2.*flags.scalefactor;
  scales.push_back(renscale);
  scales.push_back(.5*renscale);
  scales.push_back(2.*renscale);
  renscale=geometricmean(&ptjets)*flags.scalefactor;
  scales.push_back(renscale);
  scales.push_back(.5*renscale);
  scales.push_back(2.*renscale);
  renscale=invmass(&ptjets)*flags.scalefactor;
  scales.push_back(renscale);
  scales.push_back(.5*renscale);
  scales.push_back(2.*renscale);
  renscale=ptjets.at(0).perp()*flags.scalefactor;
#else
  if (flags.scalesetting==0) //max pt(jets)
    renscale=ptjets.at(0).perp()*flags.scalefactor;
  else if (flags.scalesetting==1) //fixed scale
    renscale=flags.scale;
  else if (flags.scalesetting==2) // Ht/2
    renscale=Ht/2.*flags.scalefactor;
  else if (flags.scalesetting==3) //geometric mean of pt(jets)
    renscale=geometricmean(&ptjets)*flags.scalefactor;
#endif

  return weight;
}

double CMultijet::weight3() // The four-jet cross section
{
  wt=1.;
  wt*=Generate4jphsppoint();
  if (status!=0) { // if we failed in generating a 4-jet configuration
    wt=0.;
    return wt;
  }

  double shat2=xa*xa*xb*xb*pow(getcmsenergy(),4);

#if 1
  // now evaluate the process and setup the flavour and colour info
  if (xa>1.||xb>1.) {
    setstatus(5);
    wt=0.;
    return 0.;
  }
  // choose incoming flavours based on pdfs only
  double ufa,ufb;
  ufa=renscale;  // Set to PTMAX in Generate4jphsppoint() if necessary
  ufb=renscale;

  wt*=GenerateIncomingParticles(xa,ufa,xb,ufb);

  double pmg[8][4];
  pmg[0][1]=pb.x();
  pmg[0][2]=pb.y();
  pmg[0][3]=pb.z();
  pmg[0][0]=pb.e();

  pmg[1][1]=pa.x();
  pmg[1][2]=pa.y();
  pmg[1][3]=pa.z();
  pmg[1][0]=pa.e();

  for (int ii=0;ii<4;ii++) {
    pmg[ii+2][1]=partonmomenta.at(3-ii).x();
    pmg[ii+2][2]=partonmomenta.at(3-ii).y();
    pmg[ii+2][3]=partonmomenta.at(3-ii).z();
    pmg[ii+2][0]=partonmomenta.at(3-ii).e();
  }

  wt*=Generate4JNonFKLflavourconf_and_eval(pmg);

#endif
  wt/=(16.*shat2*pow(M_PI,2));
  double pdf1=pdfta*pdftb;
  wt*=pdf1;
  wt*=389.379292;
  //  wt*=1E6; // micro barn-> pico barn

#if calcscaleunc
  // {
  //   double pdf2,pdf05;
  //   pdf05=pdfpt(0,xa,.5*ufa,aptype)*pdfpt(1,xb,.5*ufb,bptype);
  //   pdf2=pdfpt(0,xa,2.*ufa,aptype)*pdfpt(1,xb,2.*ufb,bptype);
  //   wt2=wt*pdf2/pdf1*pow(Halphas(2.*renscale)/Halphas(renscale),4);
  //   wt05=wt*pdf05/pdf1*pow(Halphas(0.5*renscale)/Halphas(renscale),4);
  // }
  for (unsigned int ii=0;ii<scales.size();ii++) {
    double reweight=pdfpt(0,xa,scales.at(ii),aptype)*pdfpt(1,xb,scales.at(ii),bptype)/pdf1*pow(Halphas(scales.at(ii))/Halphas(renscale),4);
    if (flags.logcorrect==1) {
      reweight*=(1.-Halphas(scales.at(ii))*beta0*log(partonmomenta.front().perp()/scales.at(ii))/(2.*M_PI))*(1.-Halphas(scales.at(ii))*beta0*log(partonmomenta.back().perp()/scales.at(ii))/(2.*M_PI))*(1.-Halphas(scales.at(ii))*beta0*log(partonmomenta.at(1).perp()/scales.at(ii))/(2.*M_PI))*(1.-Halphas(scales.at(ii))*beta0*log(partonmomenta.at(2).perp()/scales.at(ii))/(2.*M_PI));
      reweight/=(1.-Halphas(renscale)*beta0*log(partonmomenta.front().perp()/renscale)/(2.*M_PI))*(1.-Halphas(renscale)*beta0*log(partonmomenta.back().perp()/renscale)/(2.*M_PI))*(1.-Halphas(renscale)*beta0*log(partonmomenta.at(1).perp()/renscale)/(2.*M_PI))*(1.-Halphas(renscale)*beta0*log(partonmomenta.at(2).perp()/renscale)/(2.*M_PI));
    }
    weights.push_back(wt*reweight);
  }
#endif
  return wt;
}

double CMultijet::Event()
{
  renscale=flags.scale*flags.scalefactor;  // if scalesetting==1, this is changed later

  static double processc[4]; // how often to choose the different type of processes (cummulative)

  static bool initialisearray(true);
  static double sum(0.);
  // run only once:
  if (initialisearray) {
    if (flags.resum) {
      sum+=0.9;
      processc[0]=sum;
    }
    if (flags.nonFKLmatch2j) {
      sum+=0.05;
      processc[1]=sum;
    }
    if (flags.nonFKLmatch3j) {
      sum+=0.05;
      processc[2]=sum;
    }
    if (flags.nonFKLmatch4j) {
      sum+=0.05;
      processc[3]=sum;
    }
    initialisearray=false; // don't do this again!
  }

  setstatus(0);
  partonmomenta.clear();
  LHparticles.clear();
  jets.clear();
  scales.clear();
  weights.clear();
  xa=xb=0.;
  Ht=0.;
  aptype=bptype=0;
  matchreweight=0;
  // Clear Exceptions
  feclearexcept(FE_ALL_EXCEPT);

  static int test_excepts=(FE_DIVBYZERO & FE_INEXACT & FE_INVALID & FE_OVERFLOW & FE_UNDERFLOW);
  double rtemp=sum*ran.flat();
  double tempw;
  int set_excepts;

  if (rtemp<processc[0]) {
    processtype=1;
    tempw=weight()*sum/processc[0];
    wt2*=sum/processc[0];
    wt05*=sum/processc[0];
    set_excepts = fetestexcept(test_excepts );
    if (set_excepts & test_excepts)
      setstatus(20);
    return tempw;
  } else if (rtemp<processc[1]) {
    processtype=2;
    tempw=weight1()*sum/(processc[1]-processc[0]);
    wt2*=sum/(processc[1]-processc[0]);
    wt05*=sum/(processc[1]-processc[0]);
    set_excepts = fetestexcept(test_excepts );
    if (set_excepts & test_excepts) {
      setstatus(21);
      cout << "Bugger, 2jet\n";
    }      
    return tempw;
  } else if (rtemp<processc[2]) {
    processtype=3;
    tempw=weight2()*sum/(processc[2]-processc[1]);
    wt2*=sum/(processc[2]-processc[1]);
    wt05*=sum/(processc[2]-processc[1]);
    //    std::cout<<"3jNonFKL: "<<tempw<<std::endl;
    set_excepts = fetestexcept(test_excepts );
    if (set_excepts & test_excepts) {
      setstatus(22);
      cout << "Bugger, 3jet\n";
    }      
    return tempw;
  } else { // remember to add test when 4j-nonFKL matching is included
    processtype=4;
    tempw=weight3()*sum/(processc[3]-processc[2]);
    wt2*=sum/(processc[3]-processc[2]);
    wt05*=sum/(processc[3]-processc[2]);

    set_excepts = fetestexcept(test_excepts );
    if (set_excepts & test_excepts) {
      setstatus(23);
      cout << "Bugger, 4jet\n";
    }
    return tempw;
  }
}
