//////////////////////////////////////////////////
//////////////////////////////////////////////////
// This source code is Copyright (2012) of      //
//  Jeppe R. Andersen and Jennifer M. Smillie   //
// and is distributed under the                 //
// Gnu Public License version 2                 //
// http://www.gnu.org/licenses/gpl-2.0.html     //
// You are allowed to distribute and alter the  //
// source under the conditions of the GPLv2     //
// as long as this copyright notice             //
// is unaltered and distributed with the source //
// Any use should comply with the               //
//             MCNET GUIDELINES                 //
//    for Event Generator Authors and Users     //
// as distributed with this source code         //
//////////////////////////////////////////////////
//////////////////////////////////////////////////
#include "Hjets.h"
#include "LOjets.h"

double CHMultijet::MG(std::vector<CLHEP::HepLorentzVector> partons, CLHEP::HepLorentzVector pH, CLHEP::HepLorentzVector pa, CLHEP::HepLorentzVector pb)
{

  // Add matching here:
  // multiply by 
  // (as^4 me2/(256. shat2 pi^5) / HElimit)
  // where me2 is the full matrix element for the partonic process

  double Smatch;
  double shat2=pow(pa.invariantMass2(pb),2);

  // matching for 2-jet case
  FREAL p[5][4],p2[5][4]; // we only need the particle momenta according to the 
  // HE limit in this case

  p[0][0]=pb.e();
  p[0][1]=pb.x();
  p[0][2]=pb.y();
  p[0][3]=pb.z();

  p[1][0]=pa.e();
  p[1][1]=pa.x();
  p[1][2]=pa.y();
  p[1][3]=pa.z();

  p[2][0]=pH.e();
  p[2][1]=pH.x();
  p[2][2]=pH.y();
  p[2][3]=pH.z();

  p[3][0]=partons.at(1).e(); // positive rapidity gluon
  p[3][1]=partons.at(1).x();
  p[3][2]=partons.at(1).y();
  p[3][3]=partons.at(1).z();

  p[4][0]=partons.at(0).e(); // negative rapidity gluon
  p[4][1]=partons.at(0).x();
  p[4][2]=partons.at(0).y();
  p[4][3]=partons.at(0).z();

  // Define a momentum ordering with the extremal particles interchanged:
  p2[0][0]=pa.e();
  p2[0][1]=pa.x();
  p2[0][2]=pa.y();
  p2[0][3]=pa.z();

  p2[1][0]=pb.e();
  p2[1][1]=pb.x();
  p2[1][2]=pb.y();
  p2[1][3]=pb.z();

  p2[2][0]=pH.e();
  p2[2][1]=pH.x();
  p2[2][2]=pH.y();
  p2[2][3]=pH.z();

  p2[3][0]=partons.at(0).e(); // positive rapidity gluon
  p2[3][1]=partons.at(0).x();
  p2[3][2]=partons.at(0).y();
  p2[3][3]=partons.at(0).z();

  p2[4][0]=partons.at(1).e(); // negative rapidity gluon
  p2[4][1]=partons.at(1).x();
  p2[4][2]=partons.at(1).y();
  p2[4][3]=partons.at(1).z();	

  if (aptype==21&&bptype==21)  {
    smatrixggthgg_(p[0],&Smatch);
    Smatch*=2.;
    Smatch*=pow(Halphas(flags.mH),2)*pow(Halphas(renscale),2)/(256.*shat2*pow(M_PI,5))*pdfta*pdftb;
  }

  else if ((aptype<=5)&&(aptype>=1)&&(bptype==21)) { // If ugthug
    smatrixugthug_(p2[0],&Smatch);
    Smatch*=pow(Halphas(flags.mH),2)*pow(Halphas(renscale),2)/(256.*shat2*pow(M_PI,5))*pdfta*pdftb;
  }
  else if ((bptype<=5)&&(bptype>=1)&&(aptype==21)) { // If guthgu
    smatrixugthug_(p[0],&Smatch);
    Smatch*=pow(Halphas(flags.mH),2)*pow(Halphas(renscale),2)/(256.*shat2*pow(M_PI,5))*pdfta*pdftb;
  }
  else if ((aptype>=-5)&&(aptype<=-1)&&(bptype==21)) { // If uxgthuxg
    smatrixuxgthuxg_(p2[0],&Smatch);
    Smatch*=pow(Halphas(flags.mH),2)*pow(Halphas(renscale),2)/(256.*shat2*pow(M_PI,5))*pdfta*pdftb;
  }
  else if ((bptype>=-5)&&(bptype<=-1)&&(aptype==21)) { // If guxthgux
    smatrixuxgthuxg_(p[0],&Smatch);
    Smatch*=pow(Halphas(flags.mH),2)*pow(Halphas(renscale),2)/(256.*shat2*pow(M_PI,5))*pdfta*pdftb;
  }
  else if((aptype<=5)&&(aptype>=1)&&(bptype==aptype)) { // If uuthuu
    smatrixuuthuu_(p[0],&Smatch);
    Smatch*=2.; // Two identical particles in final state
    Smatch*=pow(Halphas(flags.mH),2)*pow(Halphas(renscale),2)/(256.*shat2*pow(M_PI,5))*pdfta*pdftb;
  }
  else if((aptype>=-5)&&(aptype<=-1)&&(bptype==aptype)) { // If uxuxthuxux
    smatrixuxuxthuxux_(p[0],&Smatch);
    Smatch*=2.; // Two identical particles in final state
    Smatch*=pow(Halphas(flags.mH),2)*pow(Halphas(renscale),2)/(256.*shat2*pow(M_PI,5))*pdfta*pdftb;
  }
  else if((aptype<=5)&&(aptype>=1)&&(bptype<=5)&&(bptype>=1)&&(aptype!=bptype)) { // If udthud
    smatrixudthud_(p[0],&Smatch);
    Smatch*=pow(Halphas(flags.mH),2)*pow(Halphas(renscale),2)/(256.*shat2*pow(M_PI,5))*pdfta*pdftb;	  
  }
  else if((aptype>=-5)&&(aptype<=-1)&&(bptype>=-5)&&(bptype<=-1)&&(aptype!=bptype)) { // If uxdxthuxdx
    smatrixuxdxthuxdx_(p[0],&Smatch);
    Smatch*=pow(Halphas(flags.mH),2)*pow(Halphas(renscale),2)/(256.*shat2*pow(M_PI,5))*pdfta*pdftb;	  
  }	
	
  else if((aptype>=-5)&&(aptype<=-1)&&(bptype<=5)&&(bptype>=1)) { // If uxdthuxd
    smatrixuxdthuxd_(p2[0],&Smatch);
    Smatch*=pow(Halphas(flags.mH),2)*pow(Halphas(renscale),2)/(256.*shat2*pow(M_PI,5))*pdfta*pdftb;	  
  }
	
  else if((bptype>=-5)&&(bptype<=-1)&&(aptype<=5)&&(aptype>=1)) { // If udxthudx
    smatrixuxdthuxd_(p[0],&Smatch);
    Smatch*=pow(Halphas(flags.mH),2)*pow(Halphas(renscale),2)/(256.*shat2*pow(M_PI,5))*pdfta*pdftb;	  
  }
  else {
    cout << "Unrecognised initial state for hjj FKL configuration: "<<aptype <<" "<<bptype<<endl;
    Smatch=0.;
  }

  if (flags.logcorrect==1)
    Smatch*=(1.-Halphas(renscale)*beta0*log(partonmomenta.front().perp()/renscale)/(2.*M_PI))*(1.-Halphas(renscale)*beta0*log(partonmomenta.back().perp()/renscale)/(2.*M_PI));

  return Smatch;
}

double CHMultijet::MG3(std::vector<CLHEP::HepLorentzVector> partons, CLHEP::HepLorentzVector pH, CLHEP::HepLorentzVector pa, CLHEP::HepLorentzVector pb)
{
  HepLorentzVector p4v,p2v,pav,pbv,p1v;
  double Smatch;
  double shat2=pow(pa.invariantMass2(pb),2);
  // Matching for three-jet case
  FREAL p[6][4],p2[6][4];

  p4v=pb;
  p2v=pa;
  pav=partons.at(0);
  pbv=partons.at(2);
  p1v=partons.at(1);

  // Momentum assignment for gg and qg channels:

  p[0][0]=p4v.e();   // Incoming parton with positive momentum
  p[0][1]=p4v.x();
  p[0][2]=p4v.y();
  p[0][3]=p4v.z();

  p[1][0]=p2v.e();   // Incoming parton with negative momentum
  p[1][1]=p2v.x();
  p[1][2]=p2v.y();
  p[1][3]=p2v.z();

  p[2][0]=pH.e();   // Higgs
  p[2][1]=pH.x();
  p[2][2]=pH.y();
  p[2][3]=pH.z();

  p[3][0]=pbv.e();   // Positive rapidity final state parton
  p[3][1]=pbv.x();
  p[3][2]=pbv.y();
  p[3][3]=pbv.z();

  p[5][0]=p1v.e();   // Extra parton from BFKL radiation
  p[5][1]=p1v.x();
  p[5][2]=p1v.y();
  p[5][3]=p1v.z();

  p[4][0]=pav.e();   // Negative rapidity final state parton
  p[4][1]=pav.x();
  p[4][2]=pav.y();
  p[4][3]=pav.z();

  // Define a momentum assignment with the extremal partons interchanged:

  p2[0][0]=p2v.e();   // Incoming parton with negative momentum
  p2[0][1]=p2v.x();
  p2[0][2]=p2v.y();
  p2[0][3]=p2v.z();

  p2[1][0]=p4v.e();   // Incoming parton with positive  momentum
  p2[1][1]=p4v.x();
  p2[1][2]=p4v.y();
  p2[1][3]=p4v.z();

  p2[2][0]=pH.e();   // Higgs
  p2[2][1]=pH.x();
  p2[2][2]=pH.y();
  p2[2][3]=pH.z();

  p2[3][0]=pav.e();   // Negative rapidity final state parton
  p2[3][1]=pav.x();
  p2[3][2]=pav.y();
  p2[3][3]=pav.z();

  p2[5][0]=p1v.e();   // Extra parton from BFKL radiation
  p2[5][1]=p1v.x();
  p2[5][2]=p1v.y();
  p2[5][3]=p1v.z();

  p2[4][0]=pbv.e();   // Positive rapidity final state parton
  p2[4][1]=pbv.x();
  p2[4][2]=pbv.y();
  p2[4][3]=pbv.z();	
	
  
  // Calculate the matrix element
  if (aptype==21&&bptype==21) {
    smatrixggthggg_(p[0],&Smatch);
    Smatch*=6.;
  }
  else if ((aptype<=5)&&(aptype>=1)&&(bptype==21)) { // If ugthugg
    smatrixugthugg_(p2[0],&Smatch);
    Smatch*=2.;
  }
  else if ((aptype<=5)&&(aptype>=1)&&(bptype==-aptype)) { // If uuxthugux
    smatrixuuxthuuxg_(p2[0],&Smatch);
  }
  else if ((bptype<=5)&&(bptype>=1)&&(bptype==-aptype)) { // If uuxthugux
    smatrixuuxthuuxg_(p[0],&Smatch);
  }
  else if ((bptype<=5)&&(bptype>=1)&&(aptype==21)) { // If guthggu
    smatrixugthugg_(p[0],&Smatch);
    Smatch*=2.;
  }
  else if ((aptype>=-5)&&(aptype<=-1)&&(bptype==21)) { // If uxgthuxgg
    smatrixuxgthuxgg_(p2[0],&Smatch);
    Smatch*=2.;
  }
  else if ((bptype>=-5)&&(bptype<=-1)&&(aptype==21)) { // If guxthggux
    smatrixuxgthuxgg_(p[0],&Smatch);
    Smatch*=2.;
  }
  else if((aptype<=5)&&(aptype>=1)&&(bptype==aptype)) { // If uuthugu
    smatrixuuthuug_(p[0],&Smatch);
    Smatch*=2.; // Two identical particles in final state
    // 	  cout << "uuthugu : "<<Smatch<<endl;
  }
  else if((aptype>=-5)&&(aptype<=-1)&&(bptype==aptype)) { // If uxuxthuxux
    smatrixuxuxthuxuxg_(p[0],&Smatch);
    Smatch*=2.; // Two identical particles in final state
  }
  else if((aptype<=5)&&(aptype>=1)&&(bptype<=5)&&(bptype>=1)&&(aptype!=bptype)) { // If udthud
    smatrixudthudg_(p[0],&Smatch);
  }
  else if((aptype>=-5)&&(aptype<=-1)&&(bptype>=-5)&&(bptype<=-1)&&(aptype!=bptype)) { // If uxdxthuxdx
    smatrixuxdxthuxdxg_(p[0],&Smatch);
  }	
	
  else if((aptype>=-5)&&(aptype<=-1)&&(bptype<=5)&&(bptype>=1)) { // If uxdthuxd
    smatrixuxdthuxdg_(p2[0],&Smatch);
  }
	
  else if((bptype>=-5)&&(bptype<=-1)&&(aptype<=5)&&(aptype>=1)) { // If udxthudx
    smatrixuxdthuxdg_(p[0],&Smatch);
  }
  else{
    cout << "Unrecognised initial state in hjjj: "<<aptype <<" "<<bptype<<endl;
  }

  Smatch*=pow(Halphas(renscale),3)*pow(Halphas(flags.mH),2)/(256.*shat2*pow(M_PI,5))*pdfta*pdftb;

  if (flags.logcorrect==1)
    Smatch*=(1.-Halphas(renscale)*beta0*log(partonmomenta.front().perp()/renscale)/(2.*M_PI))*(1.-Halphas(renscale)*beta0*log(partonmomenta.back().perp()/renscale)/(2.*M_PI))*(1.-Halphas(renscale)*beta0*log(partonmomenta.at(1).perp()/renscale)/(2.*M_PI));

  return Smatch;
}

double CHMultijet::Generate2jNonFKLflavourconf_and_eval()
{
  double weight=1.;
  double ufa,ufb;
  double me2; // the square of the matrix element
  HepLorentzVector p0,p1;
  // now evaluate the process and setup the flavour and colour info

  // choose incoming flavours based on pdfs only
  ufb=ufa=renscale;  // Set to PTMAX in Generate2jphsppoint() if necessary
  weight*=GenerateIncomingParticles(xa,ufa,xb,ufb);
  if (status!=0) { // if we failed in picking a sensible incoming state
    //  if (status==11) { // if we failed in picking a sensible incoming state
    wt=0.;
    return wt;
  }

  // setup arrays and LH information:
  FREAL p[5][4],p12[5][4],p22[5][4],p32[5][4];
  p[0][0]=pb.e();
  p[0][1]=pb.x();
  p[0][2]=pb.y();
  p[0][3]=pb.z();
  
  p[1][0]=pa.e();
  p[1][1]=pa.x();
  p[1][2]=pa.y();
  p[1][3]=pa.z();
  
  p[2][0]=pHv.e();
  p[2][1]=pHv.x();
  p[2][2]=pHv.y();
  p[2][3]=pHv.z();
  
  {
    p0=partonmomenta.at(0);
    p[3][0]=p0.e(); // "negative" rapidity gluon
    p[3][1]=p0.x();
    p[3][2]=p0.y();
    p[3][3]=p0.z();
    
    p1=partonmomenta.at(1);
    p[4][0]=p1.e(); // "positive" rapidity parton
    p[4][1]=p1.x();
    p[4][2]=p1.y();
    p[4][3]=p1.z();
  }

  if (flags.LHoutput) {
    particle tparticle; // a temp. particle
    {
      // setup vector of particles
      tparticle.PID=0;
      tparticle.COL1=0;
      tparticle.COL2=0;
      tparticle.MOT1=0;
      tparticle.MOT2=0;
      tparticle.ISTUP=0;
      tparticle.p.set(0,0,0,0);
      for (unsigned int i=0;i<partonmomenta.size()+2;i++)
	LHparticles.push_back(tparticle);
    }
    
    // incoming particles
    LHparticles.at(0).p=pb;
    LHparticles.at(0).PID=bptype;
    LHparticles.at(0).MOT1=0;
    LHparticles.at(0).MOT2=0;
    LHparticles.at(0).ISTUP=-1;
    LHparticles.at(1).p=pa;
    LHparticles.at(1).PID=aptype;
    LHparticles.at(1).MOT1=0;
    LHparticles.at(1).MOT2=0;
    LHparticles.at(1).ISTUP=-1;
  }

  // Useful array of momenta ordered according to rapidity configuration
  for (int ii=0;ii<4;ii++) {
    p12[0][ii]=p[1][ii];
    p12[1][ii]=p[0][ii];
    p12[2][ii]=p[2][ii];
    p12[3][ii]=p[4][ii];
    p12[4][ii]=p[3][ii];
    
    p22[0][ii]=p[0][ii];
    p22[1][ii]=p[1][ii];
    p22[2][ii]=p[2][ii];
    p22[3][ii]=p[4][ii];
    p22[4][ii]=p[3][ii];
    
    p32[0][ii]=p12[0][ii];
    p32[1][ii]=p12[1][ii];
    p32[2][ii]=p12[2][ii];
    p32[3][ii]=p12[4][ii];
    p32[4][ii]=p12[3][ii];	
  }

  // choose a matrix element to calculate and set the relevant flavours


  if (((aptype>=1)&&(aptype<=5))&&(bptype==-aptype)) { // if a q-qbar initial state
    int procflag=int(3.*ran.flat())+1; // 0, 1 or 2: decides type of process
    wt*=3.;

    if (procflag==1) {   // If qqxthgg

      if (flags.LHoutput) {
	// outgoing particles
	LHparticles.at(0).COL1=0;
	LHparticles.at(0).COL2=502;
	LHparticles.at(1).COL1=501;
	LHparticles.at(1).COL2=0;
	LHparticles.at(2).COL1=501;
	LHparticles.at(2).COL2=503;
	LHparticles.at(2).p=p0;
	LHparticles.at(2).PID=21;
	LHparticles.at(2).MOT1=1;
	LHparticles.at(2).MOT2=2;
	LHparticles.at(2).ISTUP=1;
	LHparticles.at(3).COL1=503;
	LHparticles.at(3).COL2=502;
	LHparticles.at(3).p=p1;
	LHparticles.at(3).PID=21;
	LHparticles.at(3).MOT1=1;
	LHparticles.at(3).MOT2=2;
	LHparticles.at(3).ISTUP=1;
      }
      smatrixuuxthgg_(p12[0],&me2);
      me2*=2.; // Identical gluons in final state
    }

    else if (procflag==2) { // If uuxthuxu (i.e. non-FKL!) 
      
      if(flags.LHoutput) {
	LHparticles.at(0).COL1=0;
	LHparticles.at(0).COL2=501;
	LHparticles.at(1).COL1=501;
	LHparticles.at(1).COL2=0;
	LHparticles.at(2).COL1=0;
	LHparticles.at(2).COL2=503;
	LHparticles.at(2).p=p0;
	LHparticles.at(2).PID=bptype;
	LHparticles.at(2).MOT1=1;
	LHparticles.at(2).MOT2=2;
	LHparticles.at(2).ISTUP=1;
	LHparticles.at(3).COL1=503;
	LHparticles.at(3).COL2=0;
	LHparticles.at(3).p=p1;
	LHparticles.at(3).PID=aptype;
	LHparticles.at(3).MOT1=1;
	LHparticles.at(3).MOT2=2;
	LHparticles.at(3).ISTUP=1;
      }
      smatrixuuxthuux_(p12[0],&me2);
    }

    else {  // If uuxthddx type

    int flav=((int(4.*ran.flat())+aptype)%5)+1; // Choose a random flavour that is not the input flavour
    int process=int(2.*ran.flat())+1; // rapidity config (0 or 1)
    weight*=8.; // I.e. 8 possible rapidity / flavour combinations

    if (process==1) {
      if (flags.LHoutput) {
	// outgoing particles
	LHparticles.at(0).COL1=0;        // pb (incoming parton in +z direction)
	LHparticles.at(0).COL2=501;
	LHparticles.at(1).COL1=502;
	LHparticles.at(1).COL2=0;      // pa (incoming parton in -z direction)
	LHparticles.at(2).COL1=502;    // p(0) (outgoing particle with negative rapidity)
	LHparticles.at(2).COL2=0;
	LHparticles.at(2).p=p0;
	LHparticles.at(2).PID=flav;
	LHparticles.at(2).MOT1=1;
	LHparticles.at(2).MOT2=2;
	LHparticles.at(2).ISTUP=1;
	LHparticles.at(3).COL1=0;      // p(1) (outgoing particle with positive rapidity)
	LHparticles.at(3).COL2=501;
	LHparticles.at(3).p=p1;
	LHparticles.at(3).PID=-flav;
	LHparticles.at(3).MOT1=1;
	LHparticles.at(3).MOT2=2;
	LHparticles.at(3).ISTUP=1;
      }
      smatrixuuxthddx_(p32[0],&me2);
    } else {
      if (flags.LHoutput) {
	LHparticles.at(0).COL1=0;
	LHparticles.at(0).COL2=501;
	LHparticles.at(1).COL1=502;
	LHparticles.at(1).COL2=0;
	LHparticles.at(2).COL1=0;
	LHparticles.at(2).COL2=501;
	LHparticles.at(2).p=p0;
	LHparticles.at(2).PID=-flav;
	LHparticles.at(2).MOT1=1;
	LHparticles.at(2).MOT2=2;
	LHparticles.at(2).ISTUP=1;
	LHparticles.at(3).COL1=502;
	LHparticles.at(3).COL2=0;
	LHparticles.at(3).p=p1;
	LHparticles.at(3).PID=flav;
	LHparticles.at(3).MOT1=1;
	LHparticles.at(3).MOT2=2;
	LHparticles.at(3).ISTUP=1;
      }
      smatrixuuxthddx_(p12[0],&me2);
    }
    }
  }


  else if (((aptype<=-1)&&(aptype>=-5))&&(bptype==-aptype)) { // if a qbar-q initial state
    
    int procflag=int(3.*ran.flat()); // 0, 1 or 2: controls process type
    wt*=3;

    if (procflag==1) {   // If qxqthgg
      if(flags.LHoutput) {
	LHparticles.at(0).COL1=502;
	LHparticles.at(0).COL2=0;
	LHparticles.at(1).COL1=0;
	LHparticles.at(1).COL2=501;
	LHparticles.at(2).COL1=503;
	LHparticles.at(2).COL2=501;
	LHparticles.at(2).p=p0;
	LHparticles.at(2).PID=21;
	LHparticles.at(2).MOT1=1;
	LHparticles.at(2).MOT2=2;
	LHparticles.at(2).ISTUP=1;
	LHparticles.at(3).COL1=502;
	LHparticles.at(3).COL2=503;
	LHparticles.at(3).p=p1;
	LHparticles.at(3).PID=21;
	LHparticles.at(3).MOT1=1;
	LHparticles.at(3).MOT2=2;
	LHparticles.at(3).ISTUP=1;
      }

      smatrixuuxthgg_(p[0],&me2);
      me2*=2.; // Two identical gluons in final state 
    }

    else if (procflag==2) { // If qxqthqqx (i.e. non-FKL!)
      if(flags.LHoutput) {
	LHparticles.at(0).COL1=501;
	LHparticles.at(0).COL2=0;
	LHparticles.at(1).COL1=0;
	LHparticles.at(1).COL2=501;
	LHparticles.at(2).COL1=503;
	LHparticles.at(2).COL2=0;
	LHparticles.at(2).p=p0;
	LHparticles.at(2).PID=bptype;
	LHparticles.at(2).MOT1=1;
	LHparticles.at(2).MOT2=2;
	LHparticles.at(2).ISTUP=1;
	LHparticles.at(3).COL1=0;
	LHparticles.at(3).COL2=503;
	LHparticles.at(3).p=p1;
	LHparticles.at(3).PID=aptype;
	LHparticles.at(3).MOT1=1;
	LHparticles.at(3).MOT2=2;
	LHparticles.at(3).ISTUP=1;
      }
      
      smatrixuuxthuux_(p[0],&me2);
    }

    else { // If uxuthddx type

      int flav=((int(4.*ran.flat())+bptype)%5)+1; // Choose a random flavour that is not the input flavour
      int process=int(2.*ran.flat())+1; // rapidity config (0 or 1)
      weight*=8.; // I.e. 8 possible rapidity / flavour combinations
      
      if(process==1) {
	if(flags.LHoutput) {
	  // outgoing particles
	  LHparticles.at(0).COL1=501;    // pb (incoming particle in +z direction)
	  LHparticles.at(0).COL2=0;
	  LHparticles.at(1).COL1=0;      // pa (incoming particle in -z direction)
	  LHparticles.at(1).COL2=502;
	  LHparticles.at(2).p=p0;        // p0 (outgoing particle with negative rapidity)
	  LHparticles.at(2).PID=flav;
	  LHparticles.at(2).MOT1=1;
	  LHparticles.at(2).MOT2=2;
	  LHparticles.at(2).COL1=501;
	  LHparticles.at(2).COL2=0;
	  LHparticles.at(2).ISTUP=1;
	  LHparticles.at(3).p=p1;        // p1 (outgoing particle with positive rapidity)
	  LHparticles.at(3).PID=-flav;
	  LHparticles.at(3).MOT1=1;
	  LHparticles.at(3).MOT2=2;
	  LHparticles.at(3).COL1=0;
	  LHparticles.at(3).COL2=502;
	  LHparticles.at(3).ISTUP=1;
	}
	smatrixuuxthddx_(p[0],&me2);
      }
      else {
	if(flags.LHoutput) {
	  // outgoing particles
	  LHparticles.at(0).COL1=501;
	  LHparticles.at(0).COL2=0;
	  LHparticles.at(1).COL1=0;
	  LHparticles.at(1).COL2=502;
	  LHparticles.at(2).p=p1;
	  LHparticles.at(2).PID=flav;
	  LHparticles.at(2).MOT1=1;
	  LHparticles.at(2).MOT2=2;
	  LHparticles.at(2).COL1=501;
	  LHparticles.at(2).COL2=0;
	  LHparticles.at(2).ISTUP=1;
	  LHparticles.at(3).p=p0;
	  LHparticles.at(3).PID=-flav;
	  LHparticles.at(3).MOT1=1;
	  LHparticles.at(3).MOT2=2;
	  LHparticles.at(3).COL1=0;
	  LHparticles.at(3).COL2=502;
	  LHparticles.at(3).ISTUP=1;
	}
	smatrixuuxthddx_(p22[0],&me2);
      }
    }
  }

  else if (aptype==21&&bptype==21) { // if both partons are gluons
    // 5 types of FS flavours and two momentum configurations
    // choose at random
    int flav=int(5.*ran.flat())+1; // 1,...,5
    int process=int(2.*ran.flat())+1; // 1 or 2
    weight*=10.;

    if (process==1) {
      if (flags.LHoutput) {
	// outgoing particles
	LHparticles.at(0).COL1=501;
	LHparticles.at(0).COL2=502;
	LHparticles.at(1).COL1=503;
	LHparticles.at(1).COL2=501;
	LHparticles.at(2).p=p0;
	LHparticles.at(2).PID=flav;
	LHparticles.at(2).MOT1=1;
	LHparticles.at(2).MOT2=2;
	LHparticles.at(2).COL1=503;
	LHparticles.at(2).COL2=0;
	LHparticles.at(2).ISTUP=1;
	LHparticles.at(3).p=p1;
	LHparticles.at(3).PID=-flav;
	LHparticles.at(3).MOT1=1;
	LHparticles.at(3).MOT2=2;
	LHparticles.at(3).COL1=0;
	LHparticles.at(3).COL2=502;
	LHparticles.at(3).ISTUP=1;
      }
      smatrixggthuux_(p[0],&me2);
//       cout << "----------\n"<< me2<<"\n";
    } else {
      if (flags.LHoutput) {
	// outgoing particles
	LHparticles.at(0).COL1=502;
	LHparticles.at(0).COL2=501;
	LHparticles.at(1).COL1=501;
	LHparticles.at(1).COL2=503;
	LHparticles.at(2).p=p0;
	LHparticles.at(2).PID=-flav;
	LHparticles.at(2).MOT1=1;
	LHparticles.at(2).MOT2=2;
	LHparticles.at(2).COL1=0;
	LHparticles.at(2).COL2=503;
	LHparticles.at(2).ISTUP=1;
	LHparticles.at(3).p=p1;
	LHparticles.at(3).PID=flav;
	LHparticles.at(3).MOT1=1;
	LHparticles.at(3).MOT2=2;
	LHparticles.at(3).COL1=502;
	LHparticles.at(3).COL2=0;
	LHparticles.at(3).ISTUP=1;
      }
      smatrixggthuux_(p12[0],&me2);
//       cout << "===========\n"<< me2<<"\n";
    }

  } 
  else if((abs(aptype)>=1)&&(abs(aptype)<=5)&&(abs(bptype)>=1)&&(abs(bptype)<=5)&&(abs(aptype)!=abs(bptype))) { // if a udthud-like state (incl. antiquarks)


    if((aptype<0)&&(bptype<0)) { // if a qx-Qx initial state
      if(flags.LHoutput) {
	LHparticles.at(0).COL1=0;
	LHparticles.at(0).COL2=501;
	LHparticles.at(1).COL1=0;
	LHparticles.at(1).COL2=502;
	LHparticles.at(2).COL1=0;
	LHparticles.at(2).COL2=501;
	LHparticles.at(2).p=p0;
	LHparticles.at(2).PID=bptype;
	LHparticles.at(2).MOT1=1;
	LHparticles.at(2).MOT2=2;
	LHparticles.at(2).ISTUP=1;
	LHparticles.at(3).COL1=0;
	LHparticles.at(3).COL2=502;
	LHparticles.at(3).p=p1;
	LHparticles.at(3).PID=aptype;
	LHparticles.at(3).MOT1=1;
	LHparticles.at(3).MOT2=2;
	LHparticles.at(3).ISTUP=1;
      }
    }
    else if((aptype>0)&&(bptype>0)) { // If a q-Q initial state
      if(flags.LHoutput) {
	LHparticles.at(0).COL1=501;   // Incoming + z
	LHparticles.at(0).COL2=0;
	LHparticles.at(1).COL1=502;   // Incoming -z
	LHparticles.at(1).COL2=0;
	LHparticles.at(2).COL1=501;   // Outgoing -z
	LHparticles.at(2).COL2=0;
	LHparticles.at(2).p=p0;
	LHparticles.at(2).PID=bptype;
	LHparticles.at(2).MOT1=1;
	LHparticles.at(2).MOT2=2;
	LHparticles.at(2).ISTUP=1;
	LHparticles.at(3).COL1=502;   // Outgoing +z
	LHparticles.at(3).COL2=0;
	LHparticles.at(3).p=p1;
	LHparticles.at(3).PID=aptype;
	LHparticles.at(3).MOT1=1;
	LHparticles.at(3).MOT2=2;
	LHparticles.at(3).ISTUP=1;
      }
    }
    else if((aptype<0)&&(bptype>0)) { // If a qx-Q initial state
      if(flags.LHoutput){
	LHparticles.at(0).COL1=502;
	LHparticles.at(0).COL2=0;
	LHparticles.at(1).COL1=0;
	LHparticles.at(1).COL2=502;
	LHparticles.at(2).COL1=503;
	LHparticles.at(2).COL2=0;
	LHparticles.at(2).p=p0;
	LHparticles.at(2).PID=bptype;
	LHparticles.at(2).MOT1=1;
	LHparticles.at(2).MOT2=2;
	LHparticles.at(2).ISTUP=1;
	LHparticles.at(3).COL1=0;
	LHparticles.at(3).COL2=503;
	LHparticles.at(3).p=p1;
	LHparticles.at(3).PID=aptype;
	LHparticles.at(3).MOT1=1;
	LHparticles.at(3).MOT2=2;
	LHparticles.at(3).ISTUP=1;	
      }
    }
    else if((aptype>0)&&(bptype<0)) { // If a q-Qx initial state
      if(flags.LHoutput) {
	LHparticles.at(0).COL1=0;
	LHparticles.at(0).COL2=501;
	LHparticles.at(1).COL1=501;
	LHparticles.at(1).COL2=0;
	LHparticles.at(2).COL1=0;
	LHparticles.at(2).COL2=503;
	LHparticles.at(2).p=p0;
	LHparticles.at(2).PID=bptype;
	LHparticles.at(2).MOT1=1;
	LHparticles.at(2).MOT2=2;
	LHparticles.at(2).ISTUP=1;
	LHparticles.at(3).COL1=503;
	LHparticles.at(3).COL2=0;
	LHparticles.at(3).p=p1;
	LHparticles.at(3).PID=aptype;
	LHparticles.at(3).MOT1=1;
	LHparticles.at(3).MOT2=2;
	LHparticles.at(3).ISTUP=1;	
      }
    }   

    smatrixudthud_(p12[0],&me2);

  }

  else if((bptype==21)&&(aptype>0)&&(aptype!=21)) { // If q-g initial state
    if(flags.LHoutput) {
      LHparticles.at(0).COL1=502;
      LHparticles.at(0).COL2=501;
      LHparticles.at(1).COL1=501;
      LHparticles.at(1).COL2=500;
      LHparticles.at(2).COL1=502;
      LHparticles.at(2).COL2=503;
      LHparticles.at(2).p=p0;
      LHparticles.at(2).PID=bptype;
      LHparticles.at(2).MOT1=1;
      LHparticles.at(2).MOT2=2;
      LHparticles.at(2).ISTUP=1;
      LHparticles.at(3).COL1=503;
      LHparticles.at(3).COL2=0;
      LHparticles.at(3).p=p1;
      LHparticles.at(3).PID=aptype;
      LHparticles.at(3).MOT1=1;
      LHparticles.at(3).MOT2=2;
      LHparticles.at(3).ISTUP=1; 
    }
    smatrixugthug_(p22[0],&me2);
  }
  else if((bptype==21)&&(aptype<0)) { // If qx-g initial state
    if(flags.LHoutput) {
      LHparticles.at(0).COL1=502;
      LHparticles.at(0).COL2=503;
      LHparticles.at(1).COL1=0;
      LHparticles.at(1).COL2=501;
      LHparticles.at(2).COL1=502;
      LHparticles.at(2).COL2=501;
      LHparticles.at(2).p=p0;
      LHparticles.at(2).PID=bptype;
      LHparticles.at(2).MOT1=1;
      LHparticles.at(2).MOT2=2;
      LHparticles.at(2).ISTUP=1;
      LHparticles.at(3).COL1=0;
      LHparticles.at(3).COL2=503;
      LHparticles.at(3).p=p1;
      LHparticles.at(3).PID=aptype;
      LHparticles.at(3).MOT1=1;
      LHparticles.at(3).MOT2=2;
      LHparticles.at(3).ISTUP=1;      
    }
    smatrixugthug_(p22[0],&me2);
  }
  else if((aptype==21)&&(bptype>0)&&(bptype!=21)) { // If g-q initial state
    if(flags.LHoutput) {
      LHparticles.at(0).COL1=502;
      LHparticles.at(0).COL2=0;
      LHparticles.at(1).COL1=501;
      LHparticles.at(1).COL2=502;
      LHparticles.at(2).COL1=503;
      LHparticles.at(2).COL2=0;
      LHparticles.at(2).p=p0;
      LHparticles.at(2).PID=bptype;
      LHparticles.at(2).MOT1=1;
      LHparticles.at(2).MOT2=2;
      LHparticles.at(2).ISTUP=1;
      LHparticles.at(3).COL1=501;
      LHparticles.at(3).COL2=503;
      LHparticles.at(3).p=p1;
      LHparticles.at(3).PID=aptype;
      LHparticles.at(3).MOT1=1;
      LHparticles.at(3).MOT2=2;
      LHparticles.at(3).ISTUP=1;      
    }
    smatrixugthug_(p[0],&me2);
  }
  else if((aptype==21)&&(bptype<0)) { // If g-qx initial state
    if(flags.LHoutput) {
      LHparticles.at(0).COL1=0;
      LHparticles.at(0).COL2=503;
      LHparticles.at(1).COL1=501;
      LHparticles.at(1).COL2=502;
      LHparticles.at(2).COL1=0;
      LHparticles.at(2).COL2=502;
      LHparticles.at(2).p=p0;
      LHparticles.at(2).PID=bptype;
      LHparticles.at(2).MOT1=1;
      LHparticles.at(2).MOT2=2;
      LHparticles.at(2).ISTUP=1;
      LHparticles.at(3).COL1=501;
      LHparticles.at(3).COL2=503;
      LHparticles.at(3).p=p1;
      LHparticles.at(3).PID=aptype;
      LHparticles.at(3).MOT1=1;
      LHparticles.at(3).MOT2=2;
      LHparticles.at(3).ISTUP=1;            
    }
    smatrixugthug_(p[0],&me2);
  }

  else if((aptype!=21)&&(aptype==bptype)) { // If a qq or qx-qx initial state
    setstatus(14);
    weight=0.;
    return 0.; // There are no non-FKL contributions to this process!
  }


  else { // if no appropriate process was found
    weight=0.;
    cout << "Error in 2 jet matching~"<<endl;
    cout << "Faulty initial state generated"<<endl;
    cout << "aptype, bptype= "<<aptype<<" "<<bptype<<endl;
    setstatus(15);
    return 0.;
  }

  if (flags.LHoutput) {
    particle tparticle; // a temp. particle
    // setup info for Higgs boson
    tparticle.PID=25;// Higgs boson
    tparticle.COL1=0;
    tparticle.COL2=0;
    tparticle.MOT1=1;
    tparticle.MOT2=2;
    tparticle.ISTUP=1;
    tparticle.p=getpHv();
    LHparticles.push_back(tparticle);
  }  

  double shat2=xa*xa*xb*xb*pow(getcmsenergy(),4);
  
//   cout <<"generateetc pdfga pdfga : "<<pdfta<<" "<<pdftb<<endl;
  weight*=pow(Halphas(flags.mH),2)*pow(Halphas(renscale),2)/(256.*shat2*pow(M_PI,5))*me2*pdfta*pdftb;

  if (flags.logcorrect==1)
    weight*=(1.-Halphas(renscale)*beta0*log(partonmomenta.front().perp()/renscale)/(2.*M_PI))*(1.-Halphas(renscale)*beta0*log(partonmomenta.back().perp()/renscale)/(2.*M_PI));
  // double tau = pHv.m2()/(4.*mt*mt);
  // double series_t = 1. + tau*(7./30. + tau*(2./21. + tau*26./525.));
  // weight*=series_t*series_t;

  return weight;
}

double CHMultijet::Generate3jNonFKLflavourconf_and_eval()
{
  double weight=1.;
  double ufa,ufb;
  double me2; // the square of the matrix element
  HepLorentzVector p0,p1,p2;
  FREAL p11[6][4],p21[6][4],p12[6][4],p13[6][4],p14[6][4],p15[6][4],p16[6][4],p22[6][4],p23[6][4],p24[6][4],p25[6][4],p26[6][4];

  // choose incoming flavours based on pdfs only
  ufb=ufa=renscale; // Set to PTMAX in Generate3jphsppoint() if necessary
  weight*=GenerateIncomingParticles(xa,ufa,xb,ufb);
  if (status!=0) { // if we failed in picking a sensible incoming state
    wt=0.;
    return wt;
  }

  p11[0][0]=pb.e();   // Incoming parton with positive momentum
  p11[0][1]=pb.x();
  p11[0][2]=pb.y();
  p11[0][3]=pb.z();

  p11[1][0]=pa.e();   // Incoming parton with negative momentum
  p11[1][1]=pa.x();
  p11[1][2]=pa.y();
  p11[1][3]=pa.z();

  p11[2][0]=pHv.e();   // Higgs
  p11[2][1]=pHv.x();
  p11[2][2]=pHv.y();
  p11[2][3]=pHv.z();

  p11[3][0]=partonmomenta.at(2).e();   // Positive rapidity final state parton
  p11[3][1]=partonmomenta.at(2).x();
  p11[3][2]=partonmomenta.at(2).y();
  p11[3][3]=partonmomenta.at(2).z();

  p11[5][0]=partonmomenta.at(1).e();   // Extra parton from BFKL radiation
  p11[5][1]=partonmomenta.at(1).x();
  p11[5][2]=partonmomenta.at(1).y();
  p11[5][3]=partonmomenta.at(1).z();

  p11[4][0]=partonmomenta.at(0).e();   // Negative rapidity final state parton
  p11[4][1]=partonmomenta.at(0).x();
  p11[4][2]=partonmomenta.at(0).y();
  p11[4][3]=partonmomenta.at(0).z();	

  // Momentum assignment for gq channel:
  for (int ii=0;ii<4;ii++) {
    p21[0][ii]=p11[1][ii];
    p21[1][ii]=p11[0][ii];
    p21[2][ii]=p11[2][ii];
    p21[3][ii]=p11[4][ii];
    p21[5][ii]=p11[5][ii];
    p21[4][ii]=p11[3][ii];
  }	

  // Momentum assignments for other rapidity orderings:
  for (int ii=0;ii<4;ii++) {
    p12[0][ii]=p11[0][ii];
    p12[1][ii]=p11[1][ii];
    p12[2][ii]=p11[2][ii];
    p12[3][ii]=p11[5][ii];
    p12[4][ii]=p11[3][ii];
    p12[5][ii]=p11[4][ii];

    p13[0][ii]=p11[0][ii];
    p13[1][ii]=p11[1][ii];
    p13[2][ii]=p11[2][ii];
    p13[3][ii]=p11[4][ii];
    p13[4][ii]=p11[5][ii];
    p13[5][ii]=p11[3][ii];

    p14[0][ii]=p11[0][ii];
    p14[1][ii]=p11[1][ii];
    p14[2][ii]=p11[2][ii];
    p14[3][ii]=p11[3][ii];
    p14[4][ii]=p11[5][ii];
    p14[5][ii]=p11[4][ii];

    p15[0][ii]=p11[0][ii];
    p15[1][ii]=p11[1][ii];
    p15[2][ii]=p11[2][ii];
    p15[3][ii]=p11[5][ii];
    p15[4][ii]=p11[4][ii];
    p15[5][ii]=p11[3][ii];

    p16[0][ii]=p11[0][ii];
    p16[1][ii]=p11[1][ii];
    p16[2][ii]=p11[2][ii];
    p16[3][ii]=p11[4][ii];
    p16[4][ii]=p11[3][ii];
    p16[5][ii]=p11[5][ii];	  
  }	

  // Other rapidity orderings for the gq channel::
  for (int ii=0;ii<4;ii++) {
    p22[0][ii]=p21[0][ii];
    p22[1][ii]=p21[1][ii];
    p22[2][ii]=p21[2][ii];
    p22[3][ii]=p21[5][ii];
    p22[4][ii]=p21[3][ii];
    p22[5][ii]=p21[4][ii];

    p23[0][ii]=p21[0][ii];
    p23[1][ii]=p21[1][ii];
    p23[2][ii]=p21[2][ii];
    p23[3][ii]=p21[4][ii];
    p23[4][ii]=p21[5][ii];
    p23[5][ii]=p21[3][ii];

    p24[0][ii]=p21[0][ii];
    p24[1][ii]=p21[1][ii];
    p24[2][ii]=p21[2][ii];
    p24[3][ii]=p21[3][ii];
    p24[4][ii]=p21[5][ii];
    p24[5][ii]=p21[4][ii];

    p25[0][ii]=p21[0][ii];
    p25[1][ii]=p21[1][ii];
    p25[2][ii]=p21[2][ii];
    p25[3][ii]=p21[5][ii];
    p25[4][ii]=p21[4][ii];
    p25[5][ii]=p21[3][ii];

    p26[0][ii]=p21[0][ii];
    p26[1][ii]=p21[1][ii];
    p26[2][ii]=p21[2][ii];
    p26[3][ii]=p21[4][ii];
    p26[4][ii]=p21[3][ii];
    p26[5][ii]=p21[5][ii];	  
  }	

  p0=partonmomenta.at(0); // Negative rapidity final state parton
  p1=partonmomenta.at(1); // "Extra" parton
  p2=partonmomenta.at(2); // Positive rapidity final state parton


  if (flags.LHoutput) {
    particle tparticle; // a temp. particle
    {
      // setup vector of particles
      tparticle.PID=0;
      tparticle.COL1=0;
      tparticle.COL2=0;
      tparticle.MOT1=0;
      tparticle.MOT2=0;
      tparticle.ISTUP=0;
      tparticle.p.set(0,0,0,0);
      for (unsigned int i=0;i<partonmomenta.size()+2;i++)
	LHparticles.push_back(tparticle);
    }
    
    // incoming particles
    LHparticles.at(0).p=pb;
    LHparticles.at(0).PID=bptype;
    LHparticles.at(0).MOT1=0;
    LHparticles.at(0).MOT2=0;
    LHparticles.at(0).ISTUP=-1;
    LHparticles.at(1).p=pa;
    LHparticles.at(1).PID=aptype;
    LHparticles.at(1).MOT1=0;
    LHparticles.at(1).MOT2=0;
    LHparticles.at(1).ISTUP=-1;
  }

  
  double shat2=xa*xa*xb*xb*pow(getcmsenergy(),4);


  if(aptype==21&&bptype==21) { // If gg initial state
    int pflag=int(6.*ran.flat()); // labels 6 rapidity configuration (0,..,5)
    int flav=int(5.*ran.flat())+1; // 5 final state quark flavours (1,..,5)
    weight*=6.*5.;

    if(pflag==0) {
      if(flags.LHoutput) {
	LHparticles.at(0).COL1=501;    // Incoming particle with +ve rapidity
	LHparticles.at(0).COL2=502;
	LHparticles.at(1).COL1=503;    // Incoming particle with -ve rapidity
	LHparticles.at(1).COL2=504;
	LHparticles.at(2).COL1=0;    // Outgoing particle with -ve rapidity
	LHparticles.at(2).COL2=504;
	LHparticles.at(2).p=p0;
	LHparticles.at(2).PID=-flav;
	LHparticles.at(2).MOT1=1;
	LHparticles.at(2).MOT2=2;
	LHparticles.at(2).ISTUP=1;
	LHparticles.at(3).COL1=503;
	LHparticles.at(3).COL2=502;
	LHparticles.at(3).p=p1;
	LHparticles.at(3).PID=21;     // "Extra" parton
	LHparticles.at(3).MOT1=1;
	LHparticles.at(3).MOT2=2;
	LHparticles.at(3).ISTUP=1;
        LHparticles.at(4).COL1=501;    // Outgoing parton with +ve rapidity 
 	LHparticles.at(4).COL2=0;
	LHparticles.at(4).p=p2;
	LHparticles.at(4).PID=flav;
	LHparticles.at(4).MOT1=1;
	LHparticles.at(4).MOT2=2;
	LHparticles.at(4).ISTUP=1;       
      }
      smatrixggthuuxg_(p11[0],&me2);
    }
    else if (pflag==1) {
      if(flags.LHoutput) {
	LHparticles.at(0).COL1=501;    // Incoming particle with +ve rapidity
	LHparticles.at(0).COL2=502;
	LHparticles.at(1).COL1=503;    // Incoming particle with -ve rapidity
	LHparticles.at(1).COL2=504;
	LHparticles.at(2).COL1=501;    // Outgoing particle with -ve rapidity
	LHparticles.at(2).COL2=504;
	LHparticles.at(2).p=p0;
	LHparticles.at(2).PID=21;
	LHparticles.at(2).MOT1=1;
	LHparticles.at(2).MOT2=2;
	LHparticles.at(2).ISTUP=1;
	LHparticles.at(3).COL1=503;       // "Extra" parton
	LHparticles.at(3).COL2=0;
	LHparticles.at(3).p=p1;
	LHparticles.at(3).PID=flav;     
	LHparticles.at(3).MOT1=1;
	LHparticles.at(3).MOT2=2;
	LHparticles.at(3).ISTUP=1;
        LHparticles.at(4).COL1=0;    // Outgoing parton with +ve rapidity 
 	LHparticles.at(4).COL2=502;
	LHparticles.at(4).p=p2;
	LHparticles.at(4).PID=-flav;
	LHparticles.at(4).MOT1=1;
	LHparticles.at(4).MOT2=2;
	LHparticles.at(4).ISTUP=1; 
      }
      smatrixggthuuxg_(p12[0],&me2);
    }
    else if (pflag==2) {
      if (flags.LHoutput) {

	LHparticles.at(0).COL1=501;    // Incoming particle with +ve rapidity
	LHparticles.at(0).COL2=502;
	LHparticles.at(1).COL1=503;    // Incoming particle with -ve rapidity
	LHparticles.at(1).COL2=504;
	LHparticles.at(2).COL1=503;    // Outgoing particle with -ve rapidity
	LHparticles.at(2).COL2=0;
	LHparticles.at(2).p=p0;
	LHparticles.at(2).PID=flav;
	LHparticles.at(2).MOT1=1;
	LHparticles.at(2).MOT2=2;
	LHparticles.at(2).ISTUP=1;
	LHparticles.at(3).COL1=0;     // "Extra" parton
	LHparticles.at(3).COL2=502;
	LHparticles.at(3).p=p1;
	LHparticles.at(3).PID=-flav;    
	LHparticles.at(3).MOT1=1;
	LHparticles.at(3).MOT2=2;
	LHparticles.at(3).ISTUP=1;
        LHparticles.at(4).COL1=501;    // Outgoing parton with +ve rapidity 
 	LHparticles.at(4).COL2=504;
	LHparticles.at(4).p=p2;
	LHparticles.at(4).PID=21;
	LHparticles.at(4).MOT1=1;
	LHparticles.at(4).MOT2=2;
	LHparticles.at(4).ISTUP=1;       
      }
      smatrixggthuuxg_(p13[0],&me2);      
    }
    else if (pflag==3) {
      if (flags.LHoutput) {

	LHparticles.at(0).COL1=501;    // Incoming particle with +ve rapidity
	LHparticles.at(0).COL2=502;
	LHparticles.at(1).COL1=503;    // Incoming particle with -ve rapidity
	LHparticles.at(1).COL2=504;
	LHparticles.at(2).COL1=501;    // Outgoing particle with -ve rapidity
	LHparticles.at(2).COL2=504;
	LHparticles.at(2).p=p0;
	LHparticles.at(2).PID=21;
	LHparticles.at(2).MOT1=1;
	LHparticles.at(2).MOT2=2;
	LHparticles.at(2).ISTUP=1;
	LHparticles.at(3).COL1=0;     // "Extra" parton
	LHparticles.at(3).COL2=502;
	LHparticles.at(3).p=p1;
	LHparticles.at(3).PID=-flav;    
	LHparticles.at(3).MOT1=1;
	LHparticles.at(3).MOT2=2;
	LHparticles.at(3).ISTUP=1;
        LHparticles.at(4).COL1=503;    // Outgoing parton with +ve rapidity 
 	LHparticles.at(4).COL2=0;
	LHparticles.at(4).p=p2;
	LHparticles.at(4).PID=flav;
	LHparticles.at(4).MOT1=1;
	LHparticles.at(4).MOT2=2;
	LHparticles.at(4).ISTUP=1;       
      }

      smatrixggthuuxg_(p14[0],&me2);
    }
    else if (pflag==4) {
      if (flags.LHoutput) {

	LHparticles.at(0).COL1=501;    // Incoming particle with +ve rapidity
	LHparticles.at(0).COL2=502;
	LHparticles.at(1).COL1=503;    // Incoming particle with -ve rapidity
	LHparticles.at(1).COL2=504;
	LHparticles.at(2).COL1=0;    // Outgoing particle with -ve rapidity
	LHparticles.at(2).COL2=502;
	LHparticles.at(2).p=p0;
	LHparticles.at(2).PID=-flav;
	LHparticles.at(2).MOT1=1;
	LHparticles.at(2).MOT2=2;
	LHparticles.at(2).ISTUP=1;
	LHparticles.at(3).COL1=503;     // "Extra" parton
	LHparticles.at(3).COL2=0;
	LHparticles.at(3).p=p1;
	LHparticles.at(3).PID=flav;    
	LHparticles.at(3).MOT1=1;
	LHparticles.at(3).MOT2=2;
	LHparticles.at(3).ISTUP=1;
        LHparticles.at(4).COL1=501;    // Outgoing parton with +ve rapidity 
 	LHparticles.at(4).COL2=504;
	LHparticles.at(4).p=p2;
	LHparticles.at(4).PID=21;
	LHparticles.at(4).MOT1=1;
	LHparticles.at(4).MOT2=2;
	LHparticles.at(4).ISTUP=1;       
      }  
      smatrixggthuuxg_(p15[0],&me2);
    }
    else if (pflag==5) {
      if (flags.LHoutput) {
	LHparticles.at(0).COL1=501;    // Incoming particle with +ve rapidity
	LHparticles.at(0).COL2=502;
	LHparticles.at(1).COL1=503;    // Incoming particle with -ve rapidity
	LHparticles.at(1).COL2=504;
	LHparticles.at(2).COL1=503;    // Outgoing particle with -ve rapidity
	LHparticles.at(2).COL2=0;
	LHparticles.at(2).p=p0;
	LHparticles.at(2).PID=flav;
	LHparticles.at(2).MOT1=1;
	LHparticles.at(2).MOT2=2;
	LHparticles.at(2).ISTUP=1;
	LHparticles.at(3).COL1=501;     // "Extra" parton
	LHparticles.at(3).COL2=504;
	LHparticles.at(3).p=p1;
	LHparticles.at(3).PID=21;    
	LHparticles.at(3).MOT1=1;
	LHparticles.at(3).MOT2=2;
	LHparticles.at(3).ISTUP=1;
        LHparticles.at(4).COL1=0;    // Outgoing parton with +ve rapidity 
 	LHparticles.at(4).COL2=502;
	LHparticles.at(4).p=p2;
	LHparticles.at(4).PID=-flav;
	LHparticles.at(4).MOT1=1;
	LHparticles.at(4).MOT2=2;
	LHparticles.at(4).ISTUP=1;       
      }
      smatrixggthuuxg_(p16[0],&me2);      
    }
  }
  else if (aptype==21&&bptype>0&&bptype!=21) { // if gq initial state
    int process=int(3.*ran.flat()); // labels partonic subprocess (0,1,2)
    weight*=3.;

    if(process==0){ // If ugthuddx
      int pflag=int(6.*ran.flat()); // labels 6 rapidity configuration (0,..,5)
      int flav=((int(4.*ran.flat())+bptype)%5)+1; // Choose a random flavour that is not the input flavour
      weight*=6.*4.; // 6 rapidity configs and 4 possible final state flavour choices
      if (pflag==0) {
	if (flags.LHoutput) {
	  LHparticles.at(0).COL1=501;    // Incoming particle with +ve rapidity
	  LHparticles.at(0).COL2=0;
	  LHparticles.at(1).COL1=502;    // Incoming particle with -ve rapidity
	  LHparticles.at(1).COL2=501;
	  LHparticles.at(2).COL1=502;    // Outgoing particle with -ve rapidity
	  LHparticles.at(2).COL2=0;
	  LHparticles.at(2).p=p0;
	  LHparticles.at(2).PID=flav;
	  LHparticles.at(2).MOT1=1;
	  LHparticles.at(2).MOT2=2;
	  LHparticles.at(2).ISTUP=1;
	  LHparticles.at(3).COL1=503;     // "Extra" parton
	  LHparticles.at(3).COL2=0;
	  LHparticles.at(3).p=p1;
	  LHparticles.at(3).PID=bptype;    
	  LHparticles.at(3).MOT1=1;
	  LHparticles.at(3).MOT2=2;
	  LHparticles.at(3).ISTUP=1;
	  LHparticles.at(4).COL1=0;    // Outgoing parton with +ve rapidity 
	  LHparticles.at(4).COL2=503;
	  LHparticles.at(4).p=p2;
	  LHparticles.at(4).PID=-flav;
	  LHparticles.at(4).MOT1=1;
	  LHparticles.at(4).MOT2=2;
	  LHparticles.at(4).ISTUP=1;       
	}
	smatrixugthuddx_(p15[0],&me2);
      }
      else if (pflag==1) {
	if (flags.LHoutput) {
	  LHparticles.at(0).COL1=501;    // Incoming particle with +ve rapidity
	  LHparticles.at(0).COL2=0;
	  LHparticles.at(1).COL1=502;    // Incoming particle with -ve rapidity
	  LHparticles.at(1).COL2=501;
	  LHparticles.at(2).COL1=0;    // Outgoing particle with -ve rapidity
	  LHparticles.at(2).COL2=503;
	  LHparticles.at(2).p=p0;
	  LHparticles.at(2).PID=-flav;
	  LHparticles.at(2).MOT1=1;
	  LHparticles.at(2).MOT2=2;
	  LHparticles.at(2).ISTUP=1;
	  LHparticles.at(3).COL1=502;     // "Extra" parton
	  LHparticles.at(3).COL2=0;
	  LHparticles.at(3).p=p1;
	  LHparticles.at(3).PID=flav;    
	  LHparticles.at(3).MOT1=1;
	  LHparticles.at(3).MOT2=2;
	  LHparticles.at(3).ISTUP=1;
	  LHparticles.at(4).COL1=503;    // Outgoing parton with +ve rapidity 
	  LHparticles.at(4).COL2=0;
	  LHparticles.at(4).p=p2;
	  LHparticles.at(4).PID=bptype;
	  LHparticles.at(4).MOT1=1;
	  LHparticles.at(4).MOT2=2;
	  LHparticles.at(4).ISTUP=1;       
	}	
	smatrixugthuddx_(p14[0],&me2);
      }
      else if (pflag==2) {
	if (flags.LHoutput) {
	  LHparticles.at(0).COL1=501;    // Incoming particle with +ve rapidity
	  LHparticles.at(0).COL2=0;
	  LHparticles.at(1).COL1=502;    // Incoming particle with -ve rapidity
	  LHparticles.at(1).COL2=501;
	  LHparticles.at(2).COL1=503;    // Outgoing particle with -ve rapidity
	  LHparticles.at(2).COL2=0;
	  LHparticles.at(2).p=p0;
	  LHparticles.at(2).PID=bptype;
	  LHparticles.at(2).MOT1=1;
	  LHparticles.at(2).MOT2=2;
	  LHparticles.at(2).ISTUP=1;
	  LHparticles.at(3).COL1=0;     // "Extra" parton
	  LHparticles.at(3).COL2=503;
	  LHparticles.at(3).p=p1;
	  LHparticles.at(3).PID=-flav;    
	  LHparticles.at(3).MOT1=1;
	  LHparticles.at(3).MOT2=2;
	  LHparticles.at(3).ISTUP=1;
	  LHparticles.at(4).COL1=502;    // Outgoing parton with +ve rapidity 
	  LHparticles.at(4).COL2=0;
	  LHparticles.at(4).p=p2;
	  LHparticles.at(4).PID=flav;
	  LHparticles.at(4).MOT1=1;
	  LHparticles.at(4).MOT2=2;
	  LHparticles.at(4).ISTUP=1;       
	}	
	smatrixugthuddx_(p16[0],&me2);	
      }
      else if (pflag==3) {
	if (flags.LHoutput) {
	  LHparticles.at(0).COL1=501;    // Incoming particle with +ve rapidity
	  LHparticles.at(0).COL2=0;
	  LHparticles.at(1).COL1=502;    // Incoming particle with -ve rapidity
	  LHparticles.at(1).COL2=501;
	  LHparticles.at(2).COL1=502;    // Outgoing particle with -ve rapidity
	  LHparticles.at(2).COL2=0;
	  LHparticles.at(2).p=p0;
	  LHparticles.at(2).PID=flav;
	  LHparticles.at(2).MOT1=1;
	  LHparticles.at(2).MOT2=2;
	  LHparticles.at(2).ISTUP=1;
	  LHparticles.at(3).COL1=0;     // "Extra" parton
	  LHparticles.at(3).COL2=503;
	  LHparticles.at(3).p=p1;
	  LHparticles.at(3).PID=-flav;    
	  LHparticles.at(3).MOT1=1;
	  LHparticles.at(3).MOT2=2;
	  LHparticles.at(3).ISTUP=1;
	  LHparticles.at(4).COL1=503;    // Outgoing parton with +ve rapidity 
	  LHparticles.at(4).COL2=0;
	  LHparticles.at(4).p=p2;
	  LHparticles.at(4).PID=bptype;
	  LHparticles.at(4).MOT1=1;
	  LHparticles.at(4).MOT2=2;
	  LHparticles.at(4).ISTUP=1;       
	}
	smatrixugthuddx_(p11[0],&me2);
      }
      else if (pflag==4) {
	if (flags.LHoutput) {
	  LHparticles.at(0).COL1=501;    // Incoming particle with +ve rapidity
	  LHparticles.at(0).COL2=0;
	  LHparticles.at(1).COL1=502;    // Incoming particle with -ve rapidity
	  LHparticles.at(1).COL2=501;
	  LHparticles.at(2).COL1=503;    // Outgoing particle with -ve rapidity
	  LHparticles.at(2).COL2=0;
	  LHparticles.at(2).p=p0;
	  LHparticles.at(2).PID=bptype;
	  LHparticles.at(2).MOT1=1;
	  LHparticles.at(2).MOT2=2;
	  LHparticles.at(2).ISTUP=1;
	  LHparticles.at(3).COL1=502;     // "Extra" parton
	  LHparticles.at(3).COL2=0;
	  LHparticles.at(3).p=p1;
	  LHparticles.at(3).PID=flav;    
	  LHparticles.at(3).MOT1=1;
	  LHparticles.at(3).MOT2=2;
	  LHparticles.at(3).ISTUP=1;
	  LHparticles.at(4).COL1=0;    // Outgoing parton with +ve rapidity 
	  LHparticles.at(4).COL2=503;
	  LHparticles.at(4).p=p2;
	  LHparticles.at(4).PID=-flav;
	  LHparticles.at(4).MOT1=1;
	  LHparticles.at(4).MOT2=2;
	  LHparticles.at(4).ISTUP=1;       
	}
      smatrixugthuddx_(p13[0],&me2);
      }
      else if (pflag==5) {
	if (flags.LHoutput) {
	  LHparticles.at(0).COL1=501;    // Incoming particle with +ve rapidity
	  LHparticles.at(0).COL2=0;
	  LHparticles.at(1).COL1=502;    // Incoming particle with -ve rapidity
	  LHparticles.at(1).COL2=501;
	  LHparticles.at(2).COL1=0;    // Outgoing particle with -ve rapidity
	  LHparticles.at(2).COL2=503;
	  LHparticles.at(2).p=p0;
	  LHparticles.at(2).PID=-flav;
	  LHparticles.at(2).MOT1=1;
	  LHparticles.at(2).MOT2=2;
	  LHparticles.at(2).ISTUP=1;
	  LHparticles.at(3).COL1=503;     // "Extra" parton
	  LHparticles.at(3).COL2=0;
	  LHparticles.at(3).p=p1;
	  LHparticles.at(3).PID=bptype;    
	  LHparticles.at(3).MOT1=1;
	  LHparticles.at(3).MOT2=2;
	  LHparticles.at(3).ISTUP=1;
	  LHparticles.at(4).COL1=502;    // Outgoing parton with +ve rapidity 
	  LHparticles.at(4).COL2=0;
	  LHparticles.at(4).p=p2;
	  LHparticles.at(4).PID=flav;
	  LHparticles.at(4).MOT1=1;
	  LHparticles.at(4).MOT2=2;
	  LHparticles.at(4).ISTUP=1;       
	}
	smatrixugthuddx_(p12[0],&me2);
      }	
    }
    else if (process==1) { // If ugthuuux
      int pflag=int(3.*ran.flat()); // Labels rapidity configuration (0,1,2)
      weight*=3.;
      if (pflag==0) {
	if (flags.LHoutput) {
	  LHparticles.at(0).COL1=501;    // Incoming particle with +ve rapidity
	  LHparticles.at(0).COL2=0;
	  LHparticles.at(1).COL1=502;    // Incoming particle with -ve rapidity
	  LHparticles.at(1).COL2=501;
	  LHparticles.at(2).COL1=502;    // Outgoing particle with -ve rapidity
	  LHparticles.at(2).COL2=0;
	  LHparticles.at(2).p=p0;
	  LHparticles.at(2).PID=bptype;
	  LHparticles.at(2).MOT1=1;
	  LHparticles.at(2).MOT2=2;
	  LHparticles.at(2).ISTUP=1;
	  LHparticles.at(3).COL1=0;     // "Extra" parton
	  LHparticles.at(3).COL2=503;
	  LHparticles.at(3).p=p1;
	  LHparticles.at(3).PID=-bptype;    
	  LHparticles.at(3).MOT1=1;
	  LHparticles.at(3).MOT2=2;
	  LHparticles.at(3).ISTUP=1;
	  LHparticles.at(4).COL1=503;    // Outgoing parton with +ve rapidity 
	  LHparticles.at(4).COL2=0;
	  LHparticles.at(4).p=p2;
	  LHparticles.at(4).PID=bptype;
	  LHparticles.at(4).MOT1=1;
	  LHparticles.at(4).MOT2=2;
	  LHparticles.at(4).ISTUP=1;       
	}	
	smatrixugthuuux_(p16[0],&me2);
	me2*=2.; // 2 identical quarks in final state
      }
      else if (pflag==1) {
	if (flags.LHoutput) {
	  LHparticles.at(0).COL1=501;    // Incoming particle with +ve rapidity
	  LHparticles.at(0).COL2=0;
	  LHparticles.at(1).COL1=502;    // Incoming particle with -ve rapidity
	  LHparticles.at(1).COL2=501;
	  LHparticles.at(2).COL1=503;    // Outgoing particle with -ve rapidity
	  LHparticles.at(2).COL2=0;
	  LHparticles.at(2).p=p0;
	  LHparticles.at(2).PID=bptype;
	  LHparticles.at(2).MOT1=1;
	  LHparticles.at(2).MOT2=2;
	  LHparticles.at(2).ISTUP=1;
	  LHparticles.at(3).COL1=502;     // "Extra" parton
	  LHparticles.at(3).COL2=0;
	  LHparticles.at(3).p=p1;
	  LHparticles.at(3).PID=bptype;    
	  LHparticles.at(3).MOT1=1;
	  LHparticles.at(3).MOT2=2;
	  LHparticles.at(3).ISTUP=1;
	  LHparticles.at(4).COL1=0;    // Outgoing parton with +ve rapidity 
	  LHparticles.at(4).COL2=503;
	  LHparticles.at(4).p=p2;
	  LHparticles.at(4).PID=-bptype;
	  LHparticles.at(4).MOT1=1;
	  LHparticles.at(4).MOT2=2;
	  LHparticles.at(4).ISTUP=1;       
	}		
	smatrixugthuuux_(p15[0],&me2);
	me2*=2.; // 2 identical quarks in final state
      }
      else if (pflag==2) {
	if (flags.LHoutput) {
	  LHparticles.at(0).COL1=501;    // Incoming particle with +ve rapidity
	  LHparticles.at(0).COL2=0;
	  LHparticles.at(1).COL1=502;    // Incoming particle with -ve rapidity
	  LHparticles.at(1).COL2=501;
	  LHparticles.at(2).COL1=0;    // Outgoing particle with -ve rapidity
	  LHparticles.at(2).COL2=503;
	  LHparticles.at(2).p=p0;
	  LHparticles.at(2).PID=-bptype;
	  LHparticles.at(2).MOT1=1;
	  LHparticles.at(2).MOT2=2;
	  LHparticles.at(2).ISTUP=1;
	  LHparticles.at(3).COL1=503;     // "Extra" parton
	  LHparticles.at(3).COL2=0;
	  LHparticles.at(3).p=p1;
	  LHparticles.at(3).PID=bptype;    
	  LHparticles.at(3).MOT1=1;
	  LHparticles.at(3).MOT2=2;
	  LHparticles.at(3).ISTUP=1;
	  LHparticles.at(4).COL1=502;    // Outgoing parton with +ve rapidity 
	  LHparticles.at(4).COL2=0;
	  LHparticles.at(4).p=p2;
	  LHparticles.at(4).PID=bptype;
	  LHparticles.at(4).MOT1=1;
	  LHparticles.at(4).MOT2=2;
	  LHparticles.at(4).ISTUP=1;       
	}	
	smatrixugthuuux_(p14[0],&me2);
	me2*=2.; // 2 identical quarks in final state	
      }
    }
    else if (process==2) { // If ugthugg
      int pflag=int(2.*ran.flat()); // 2 rapidity configurations that are non-FKL like
      weight*=2.;
      if (pflag==0) {
	if (flags.LHoutput) {
	  LHparticles.at(0).COL1=501;    // Incoming particle with +ve rapidity
	  LHparticles.at(0).COL2=0;
	  LHparticles.at(1).COL1=502;    // Incoming particle with -ve rapidity
	  LHparticles.at(1).COL2=501;
	  LHparticles.at(2).COL1=503;    // Outgoing particle with -ve rapidity
	  LHparticles.at(2).COL2=0;
	  LHparticles.at(2).p=p0;
	  LHparticles.at(2).PID=bptype;
	  LHparticles.at(2).MOT1=1;
	  LHparticles.at(2).MOT2=2;
	  LHparticles.at(2).ISTUP=1;
	  LHparticles.at(3).COL1=502;     // "Extra" parton
	  LHparticles.at(3).COL2=504;
	  LHparticles.at(3).p=p1;
	  LHparticles.at(3).PID=21;    
	  LHparticles.at(3).MOT1=1;
	  LHparticles.at(3).MOT2=2;
	  LHparticles.at(3).ISTUP=1;
	  LHparticles.at(4).COL1=504;    // Outgoing parton with +ve rapidity 
	  LHparticles.at(4).COL2=503;
	  LHparticles.at(4).p=p2;
	  LHparticles.at(4).PID=21;
	  LHparticles.at(4).MOT1=1;
	  LHparticles.at(4).MOT2=2;
	  LHparticles.at(4).ISTUP=1;       
	}	
	smatrixugthugg_(p16[0],&me2);
	me2*=2.; // 2 identical gluons in final state
      }
      else if (pflag==1) {
	if (flags.LHoutput) {
	  LHparticles.at(0).COL1=501;    // Incoming particle with +ve rapidity
	  LHparticles.at(0).COL2=0;
	  LHparticles.at(1).COL1=502;    // Incoming particle with -ve rapidity
	  LHparticles.at(1).COL2=501;
	  LHparticles.at(2).COL1=503;    // Outgoing particle with -ve rapidity
	  LHparticles.at(2).COL2=504;
	  LHparticles.at(2).p=p0;
	  LHparticles.at(2).PID=21;
	  LHparticles.at(2).MOT1=1;
	  LHparticles.at(2).MOT2=2;
	  LHparticles.at(2).ISTUP=1;
	  LHparticles.at(3).COL1=504;     // "Extra" parton
	  LHparticles.at(3).COL2=0;
	  LHparticles.at(3).p=p1;
	  LHparticles.at(3).PID=bptype;    
	  LHparticles.at(3).MOT1=1;
	  LHparticles.at(3).MOT2=2;
	  LHparticles.at(3).ISTUP=1;
	  LHparticles.at(4).COL1=502;    // Outgoing parton with +ve rapidity 
	  LHparticles.at(4).COL2=503;
	  LHparticles.at(4).p=p2;
	  LHparticles.at(4).PID=21;
	  LHparticles.at(4).MOT1=1;
	  LHparticles.at(4).MOT2=2;
	  LHparticles.at(4).ISTUP=1;       
	}	
	smatrixugthugg_(p15[0],&me2);
	me2*=2.; // 2 identical gluons in final state      	
      }
    }
  }
  else if (bptype==21&&aptype>0&&aptype!=21) { // if qg initial state
    int process=int(3.*ran.flat()); // labels partonic subprocess (0,1,2)
    weight*=3.;
    if(process==0){ // If ugthuddx
      int pflag=int(6.*ran.flat()); // labels 6 rapidity configuration (0,..,5)
      int flav=((int(4.*ran.flat())+bptype)%5)+1; // Choose a random flavour that is not the input flavour
      weight*=6.*4.; // 6 rapidity configs and 4 possible final state flavour choices
      if(pflag==0) {
	if (flags.LHoutput) {
	  LHparticles.at(0).COL1=501;    // Incoming particle with +ve rapidity
	  LHparticles.at(0).COL2=502;
	  LHparticles.at(1).COL1=502;    // Incoming particle with -ve rapidity
	  LHparticles.at(1).COL2=0;
	  LHparticles.at(2).COL1=503;    // Outgoing particle with -ve rapidity
	  LHparticles.at(2).COL2=0;
	  LHparticles.at(2).p=p0;
	  LHparticles.at(2).PID=aptype;
	  LHparticles.at(2).MOT1=1;
	  LHparticles.at(2).MOT2=2;
	  LHparticles.at(2).ISTUP=1;
	  LHparticles.at(3).COL1=0;     // "Extra" parton
	  LHparticles.at(3).COL2=503;
	  LHparticles.at(3).p=p1;
	  LHparticles.at(3).PID=-flav;    
	  LHparticles.at(3).MOT1=1;
	  LHparticles.at(3).MOT2=2;
	  LHparticles.at(3).ISTUP=1;
	  LHparticles.at(4).COL1=501;    // Outgoing parton with +ve rapidity 
	  LHparticles.at(4).COL2=0;
	  LHparticles.at(4).p=p2;
	  LHparticles.at(4).PID=flav;
	  LHparticles.at(4).MOT1=1;
	  LHparticles.at(4).MOT2=2;
	  LHparticles.at(4).ISTUP=1;       
	}	
	smatrixugthuddx_(p21[0],&me2);
      }
      else if (pflag==1) {
	if (flags.LHoutput) {
	  LHparticles.at(0).COL1=501;    // Incoming particle with +ve rapidity
	  LHparticles.at(0).COL2=502;
	  LHparticles.at(1).COL1=502;    // Incoming particle with -ve rapidity
	  LHparticles.at(1).COL2=0;
	  LHparticles.at(2).COL1=501;    // Outgoing particle with -ve rapidity
	  LHparticles.at(2).COL2=0;
	  LHparticles.at(2).p=p0;
	  LHparticles.at(2).PID=flav;
	  LHparticles.at(2).MOT1=1;
	  LHparticles.at(2).MOT2=2;
	  LHparticles.at(2).ISTUP=1;
	  LHparticles.at(3).COL1=503;     // "Extra" parton
	  LHparticles.at(3).COL2=0;
	  LHparticles.at(3).p=p1;
	  LHparticles.at(3).PID=aptype;    
	  LHparticles.at(3).MOT1=1;
	  LHparticles.at(3).MOT2=2;
	  LHparticles.at(3).ISTUP=1;
	  LHparticles.at(4).COL1=0;    // Outgoing parton with +ve rapidity 
	  LHparticles.at(4).COL2=503;
	  LHparticles.at(4).p=p2;
	  LHparticles.at(4).PID=-flav;
	  LHparticles.at(4).MOT1=1;
	  LHparticles.at(4).MOT2=2;
	  LHparticles.at(4).ISTUP=1;       
	}		
	smatrixugthuddx_(p22[0],&me2);
      }
      else if (pflag==2) {
	if (flags.LHoutput) {
	  LHparticles.at(0).COL1=501;    // Incoming particle with +ve rapidity
	  LHparticles.at(0).COL2=502;
	  LHparticles.at(1).COL1=502;    // Incoming particle with -ve rapidity
	  LHparticles.at(1).COL2=0;
	  LHparticles.at(2).COL1=0;    // Outgoing particle with -ve rapidity
	  LHparticles.at(2).COL2=503;
	  LHparticles.at(2).p=p0;
	  LHparticles.at(2).PID=-flav;
	  LHparticles.at(2).MOT1=1;
	  LHparticles.at(2).MOT2=2;
	  LHparticles.at(2).ISTUP=1;
	  LHparticles.at(3).COL1=501;     // "Extra" parton
	  LHparticles.at(3).COL2=0;
	  LHparticles.at(3).p=p1;
	  LHparticles.at(3).PID=flav;    
	  LHparticles.at(3).MOT1=1;
	  LHparticles.at(3).MOT2=2;
	  LHparticles.at(3).ISTUP=1;
	  LHparticles.at(4).COL1=503;    // Outgoing parton with +ve rapidity 
	  LHparticles.at(4).COL2=0;
	  LHparticles.at(4).p=p2;
	  LHparticles.at(4).PID=aptype;
	  LHparticles.at(4).MOT1=1;
	  LHparticles.at(4).MOT2=2;
	  LHparticles.at(4).ISTUP=1;       
	}		
	smatrixugthuddx_(p23[0],&me2);
      }
      else if (pflag==3) {
	if (flags.LHoutput) {
	  LHparticles.at(0).COL1=501;    // Incoming particle with +ve rapidity
	  LHparticles.at(0).COL2=502;
	  LHparticles.at(1).COL1=502;    // Incoming particle with -ve rapidity
	  LHparticles.at(1).COL2=0;
	  LHparticles.at(2).COL1=503;    // Outgoing particle with -ve rapidity
	  LHparticles.at(2).COL2=0;
	  LHparticles.at(2).p=p0;
	  LHparticles.at(2).PID=aptype;
	  LHparticles.at(2).MOT1=1;
	  LHparticles.at(2).MOT2=2;
	  LHparticles.at(2).ISTUP=1;
	  LHparticles.at(3).COL1=501;     // "Extra" parton
	  LHparticles.at(3).COL2=0;
	  LHparticles.at(3).p=p1;
	  LHparticles.at(3).PID=flav;    
	  LHparticles.at(3).MOT1=1;
	  LHparticles.at(3).MOT2=2;
	  LHparticles.at(3).ISTUP=1;
	  LHparticles.at(4).COL1=0;    // Outgoing parton with +ve rapidity 
	  LHparticles.at(4).COL2=503;
	  LHparticles.at(4).p=p2;
	  LHparticles.at(4).PID=-flav;
	  LHparticles.at(4).MOT1=1;
	  LHparticles.at(4).MOT2=2;
	  LHparticles.at(4).ISTUP=1;       
	}	
	smatrixugthuddx_(p24[0],&me2);
      }
      else if (pflag==4) {
	if (flags.LHoutput) {
	  LHparticles.at(0).COL1=501;    // Incoming particle with +ve rapidity
	  LHparticles.at(0).COL2=502;
	  LHparticles.at(1).COL1=502;    // Incoming particle with -ve rapidity
	  LHparticles.at(1).COL2=0;
	  LHparticles.at(2).COL1=0;    // Outgoing particle with -ve rapidity
	  LHparticles.at(2).COL2=503;
	  LHparticles.at(2).p=p0;
	  LHparticles.at(2).PID=-flav;
	  LHparticles.at(2).MOT1=1;
	  LHparticles.at(2).MOT2=2;
	  LHparticles.at(2).ISTUP=1;
	  LHparticles.at(3).COL1=503;     // "Extra" parton
	  LHparticles.at(3).COL2=0;
	  LHparticles.at(3).p=p1;
	  LHparticles.at(3).PID=aptype;    
	  LHparticles.at(3).MOT1=1;
	  LHparticles.at(3).MOT2=2;
	  LHparticles.at(3).ISTUP=1;
	  LHparticles.at(4).COL1=501;    // Outgoing parton with +ve rapidity 
	  LHparticles.at(4).COL2=0;
	  LHparticles.at(4).p=p2;
	  LHparticles.at(4).PID=flav;
	  LHparticles.at(4).MOT1=1;
	  LHparticles.at(4).MOT2=2;
	  LHparticles.at(4).ISTUP=1;       
	}		
	smatrixugthuddx_(p25[0],&me2);
      }
      else if (pflag==5) {
	if (flags.LHoutput) {
	  LHparticles.at(0).COL1=501;    // Incoming particle with +ve rapidity
	  LHparticles.at(0).COL2=502;
	  LHparticles.at(1).COL1=502;    // Incoming particle with -ve rapidity
	  LHparticles.at(1).COL2=0;
	  LHparticles.at(2).COL1=501;    // Outgoing particle with -ve rapidity
	  LHparticles.at(2).COL2=0;
	  LHparticles.at(2).p=p0;
	  LHparticles.at(2).PID=flav;
	  LHparticles.at(2).MOT1=1;
	  LHparticles.at(2).MOT2=2;
	  LHparticles.at(2).ISTUP=1;
	  LHparticles.at(3).COL1=0;     // "Extra" parton
	  LHparticles.at(3).COL2=503;
	  LHparticles.at(3).p=p1;
	  LHparticles.at(3).PID=-flav;    
	  LHparticles.at(3).MOT1=1;
	  LHparticles.at(3).MOT2=2;
	  LHparticles.at(3).ISTUP=1;
	  LHparticles.at(4).COL1=503;    // Outgoing parton with +ve rapidity 
	  LHparticles.at(4).COL2=0;
	  LHparticles.at(4).p=p2;
	  LHparticles.at(4).PID=aptype;
	  LHparticles.at(4).MOT1=1;
	  LHparticles.at(4).MOT2=2;
	  LHparticles.at(4).ISTUP=1;       
	}	
	smatrixugthuddx_(p26[0],&me2);

      }
    }
    else if (process==1) {  // If ugthuuux
      int pflag=int(3.*ran.flat()); // Labels rapidity configuration (0,1,2)
      weight*=3.;
      if (pflag==0) {
	if (flags.LHoutput) {
	  LHparticles.at(0).COL1=501;    // Incoming particle with +ve rapidity
	  LHparticles.at(0).COL2=502;
	  LHparticles.at(1).COL1=502;    // Incoming particle with -ve rapidity
	  LHparticles.at(1).COL2=0;
	  LHparticles.at(2).COL1=503;    // Outgoing particle with -ve rapidity
	  LHparticles.at(2).COL2=0;
	  LHparticles.at(2).p=p0;
	  LHparticles.at(2).PID=aptype;
	  LHparticles.at(2).MOT1=1;
	  LHparticles.at(2).MOT2=2;
	  LHparticles.at(2).ISTUP=1;
	  LHparticles.at(3).COL1=0;     // "Extra" parton
	  LHparticles.at(3).COL2=503;
	  LHparticles.at(3).p=p1;
	  LHparticles.at(3).PID=-aptype;    
	  LHparticles.at(3).MOT1=1;
	  LHparticles.at(3).MOT2=2;
	  LHparticles.at(3).ISTUP=1;
	  LHparticles.at(4).COL1=501;    // Outgoing parton with +ve rapidity 
	  LHparticles.at(4).COL2=0;
	  LHparticles.at(4).p=p2;
	  LHparticles.at(4).PID=aptype;
	  LHparticles.at(4).MOT1=1;
	  LHparticles.at(4).MOT2=2;
	  LHparticles.at(4).ISTUP=1;       
	}	

	smatrixugthuuux_(p21[0],&me2);
	me2*=2.; // 2 identical quarks in final state
      }
      else if (pflag==1) {
	if (flags.LHoutput) {
	  LHparticles.at(0).COL1=501;    // Incoming particle with +ve rapidity
	  LHparticles.at(0).COL2=502;
	  LHparticles.at(1).COL1=502;    // Incoming particle with -ve rapidity
	  LHparticles.at(1).COL2=0;
	  LHparticles.at(2).COL1=501;    // Outgoing particle with -ve rapidity
	  LHparticles.at(2).COL2=0;
	  LHparticles.at(2).p=p0;
	  LHparticles.at(2).PID=aptype;
	  LHparticles.at(2).MOT1=1;
	  LHparticles.at(2).MOT2=2;
	  LHparticles.at(2).ISTUP=1;
	  LHparticles.at(3).COL1=503;     // "Extra" parton
	  LHparticles.at(3).COL2=0;
	  LHparticles.at(3).p=p1;
	  LHparticles.at(3).PID=aptype;    
	  LHparticles.at(3).MOT1=1;
	  LHparticles.at(3).MOT2=2;
	  LHparticles.at(3).ISTUP=1;
	  LHparticles.at(4).COL1=0;    // Outgoing parton with +ve rapidity 
	  LHparticles.at(4).COL2=503;
	  LHparticles.at(4).p=p2;
	  LHparticles.at(4).PID=-aptype;
	  LHparticles.at(4).MOT1=1;
	  LHparticles.at(4).MOT2=2;
	  LHparticles.at(4).ISTUP=1;       
	}	
	smatrixugthuuux_(p22[0],&me2);
	me2*=2.; // 2 identical quarks in final state
      }
      else if (pflag==2) {
	if (flags.LHoutput) {
	  LHparticles.at(0).COL1=501;    // Incoming particle with +ve rapidity
	  LHparticles.at(0).COL2=502;
	  LHparticles.at(1).COL1=502;    // Incoming particle with -ve rapidity
	  LHparticles.at(1).COL2=0;
	  LHparticles.at(2).COL1=0;    // Outgoing particle with -ve rapidity
	  LHparticles.at(2).COL2=503;
	  LHparticles.at(2).p=p0;
	  LHparticles.at(2).PID=-aptype;
	  LHparticles.at(2).MOT1=1;
	  LHparticles.at(2).MOT2=2;
	  LHparticles.at(2).ISTUP=1;
	  LHparticles.at(3).COL1=501;     // "Extra" parton
	  LHparticles.at(3).COL2=0;
	  LHparticles.at(3).p=p1;
	  LHparticles.at(3).PID=aptype;    
	  LHparticles.at(3).MOT1=1;
	  LHparticles.at(3).MOT2=2;
	  LHparticles.at(3).ISTUP=1;
	  LHparticles.at(4).COL1=503;    // Outgoing parton with +ve rapidity 
	  LHparticles.at(4).COL2=0;
	  LHparticles.at(4).p=p2;
	  LHparticles.at(4).PID=aptype;
	  LHparticles.at(4).MOT1=1;
	  LHparticles.at(4).MOT2=2;
	  LHparticles.at(4).ISTUP=1;       
	}	
	smatrixugthuuux_(p23[0],&me2);
	me2*=2; // 2 identical quarks in final state
      }

    }
    else if (process==2) { // if ugthugg
      int pflag=int(2.*ran.flat()); // 2 rapidity configurations that are non-FKL like
      weight*=2.;
      if(pflag==0) {
	if (flags.LHoutput) {
	  LHparticles.at(0).COL1=501;    // Incoming particle with +ve rapidity
	  LHparticles.at(0).COL2=502;
	  LHparticles.at(1).COL1=502;    // Incoming particle with -ve rapidity
	  LHparticles.at(1).COL2=0;
	  LHparticles.at(2).COL1=503;    // Outgoing particle with -ve rapidity
	  LHparticles.at(2).COL2=504;
	  LHparticles.at(2).p=p0;
	  LHparticles.at(2).PID=21;
	  LHparticles.at(2).MOT1=1;
	  LHparticles.at(2).MOT2=2;
	  LHparticles.at(2).ISTUP=1;
	  LHparticles.at(3).COL1=504;     // "Extra" parton
	  LHparticles.at(3).COL2=0;
	  LHparticles.at(3).p=p1;
	  LHparticles.at(3).PID=aptype;    
	  LHparticles.at(3).MOT1=1;
	  LHparticles.at(3).MOT2=2;
	  LHparticles.at(3).ISTUP=1;
	  LHparticles.at(4).COL1=501;    // Outgoing parton with +ve rapidity 
	  LHparticles.at(4).COL2=503;
	  LHparticles.at(4).p=p2;
	  LHparticles.at(4).PID=21;
	  LHparticles.at(4).MOT1=1;
	  LHparticles.at(4).MOT2=2;
	  LHparticles.at(4).ISTUP=1;       
	}
	smatrixugthugg_(p22[0],&me2);
	me2*=2.; // 2 identical gluons in final state
      }
      else if (pflag==1) {
	if (flags.LHoutput) {
	  LHparticles.at(0).COL1=501;    // Incoming particle with +ve rapidity
	  LHparticles.at(0).COL2=502;
	  LHparticles.at(1).COL1=502;    // Incoming particle with -ve rapidity
	  LHparticles.at(1).COL2=0;
	  LHparticles.at(2).COL1=501;    // Outgoing particle with -ve rapidity
	  LHparticles.at(2).COL2=503;
	  LHparticles.at(2).p=p0;
	  LHparticles.at(2).PID=21;
	  LHparticles.at(2).MOT1=1;
	  LHparticles.at(2).MOT2=2;
	  LHparticles.at(2).ISTUP=1;
	  LHparticles.at(3).COL1=503;     // "Extra" parton
	  LHparticles.at(3).COL2=504;
	  LHparticles.at(3).p=p1;
	  LHparticles.at(3).PID=21;    
	  LHparticles.at(3).MOT1=1;
	  LHparticles.at(3).MOT2=2;
	  LHparticles.at(3).ISTUP=1;
	  LHparticles.at(4).COL1=504;    // Outgoing parton with +ve rapidity 
	  LHparticles.at(4).COL2=0;
	  LHparticles.at(4).p=p2;
	  LHparticles.at(4).PID=aptype;
	  LHparticles.at(4).MOT1=1;
	  LHparticles.at(4).MOT2=2;
	  LHparticles.at(4).ISTUP=1;       
	}	
	smatrixugthugg_(p23[0],&me2);
	me2*=2.;
      }
    }
  }
  else if (aptype==21&&bptype<0) { // if gqx initial state
    int process=int(3.*ran.flat()); // labels partonic subprocess (0,1,2)
    weight*=3.;

    if(process==0){ // If ugthuddx
      int pflag=int(6.*ran.flat()); // labels 6 rapidity configuration (0,..,5)
      int flav=((int(4.*ran.flat())+abs(bptype))%5)+1; // Choose a random flavour that is not the input flavour
      weight*=6.*4.; // 6 rapidity configs and 4 possible final state flavour choices
      if (pflag==0) {
	if (flags.LHoutput) {
	  LHparticles.at(0).COL1=0;    // Incoming particle with +ve rapidity
	  LHparticles.at(0).COL2=501;
	  LHparticles.at(1).COL1=501;    // Incoming particle with -ve rapidity
	  LHparticles.at(1).COL2=502;
	  LHparticles.at(2).COL1=503;    // Outgoing particle with -ve rapidity
	  LHparticles.at(2).COL2=0;
	  LHparticles.at(2).p=p0;
	  LHparticles.at(2).PID=flav;
	  LHparticles.at(2).MOT1=1;
	  LHparticles.at(2).MOT2=2;
	  LHparticles.at(2).ISTUP=1;
	  LHparticles.at(3).COL1=0;     // "Extra" parton
	  LHparticles.at(3).COL2=502;
	  LHparticles.at(3).p=p1;
	  LHparticles.at(3).PID=-flav;    
	  LHparticles.at(3).MOT1=1;
	  LHparticles.at(3).MOT2=2;
	  LHparticles.at(3).ISTUP=1;
	  LHparticles.at(4).COL1=0;    // Outgoing parton with +ve rapidity 
	  LHparticles.at(4).COL2=503;
	  LHparticles.at(4).p=p2;
	  LHparticles.at(4).PID=bptype;
	  LHparticles.at(4).MOT1=1;
	  LHparticles.at(4).MOT2=2;
	  LHparticles.at(4).ISTUP=1;       
	}	
	smatrixuxgthuxddx_(p11[0],&me2);
      }
      else if (pflag==1) {
	if (flags.LHoutput) {
	  LHparticles.at(0).COL1=0;    // Incoming particle with +ve rapidity
	  LHparticles.at(0).COL2=501;
	  LHparticles.at(1).COL1=501;    // Incoming particle with -ve rapidity
	  LHparticles.at(1).COL2=502;
	  LHparticles.at(2).COL1=0;    // Outgoing particle with -ve rapidity
	  LHparticles.at(2).COL2=502;
	  LHparticles.at(2).p=p0;
	  LHparticles.at(2).PID=-flav;
	  LHparticles.at(2).MOT1=1;
	  LHparticles.at(2).MOT2=2;
	  LHparticles.at(2).ISTUP=1;
	  LHparticles.at(3).COL1=0;     // "Extra" parton
	  LHparticles.at(3).COL2=503;
	  LHparticles.at(3).p=p1;
	  LHparticles.at(3).PID=bptype;    
	  LHparticles.at(3).MOT1=1;
	  LHparticles.at(3).MOT2=2;
	  LHparticles.at(3).ISTUP=1;
	  LHparticles.at(4).COL1=503;    // Outgoing parton with +ve rapidity 
	  LHparticles.at(4).COL2=0;
	  LHparticles.at(4).p=p2;
	  LHparticles.at(4).PID=flav;
	  LHparticles.at(4).MOT1=1;
	  LHparticles.at(4).MOT2=2;
	  LHparticles.at(4).ISTUP=1;       
	}		  
	smatrixuxgthuxddx_(p12[0],&me2);
      }
      else if (pflag==2) {
	if (flags.LHoutput) {
	  LHparticles.at(0).COL1=0;    // Incoming particle with +ve rapidity
	  LHparticles.at(0).COL2=501;
	  LHparticles.at(1).COL1=501;    // Incoming particle with -ve rapidity
	  LHparticles.at(1).COL2=502;
	  LHparticles.at(2).COL1=0;    // Outgoing particle with -ve rapidity
	  LHparticles.at(2).COL2=503;
	  LHparticles.at(2).p=p0;
	  LHparticles.at(2).PID=bptype;
	  LHparticles.at(2).MOT1=1;
	  LHparticles.at(2).MOT2=2;
	  LHparticles.at(2).ISTUP=1;
	  LHparticles.at(3).COL1=503;     // "Extra" parton
	  LHparticles.at(3).COL2=0;
	  LHparticles.at(3).p=p1;
	  LHparticles.at(3).PID=flav;    
	  LHparticles.at(3).MOT1=1;
	  LHparticles.at(3).MOT2=2;
	  LHparticles.at(3).ISTUP=1;
	  LHparticles.at(4).COL1=0;    // Outgoing parton with +ve rapidity 
	  LHparticles.at(4).COL2=502;
	  LHparticles.at(4).p=p2;
	  LHparticles.at(4).PID=-flav;
	  LHparticles.at(4).MOT1=1;
	  LHparticles.at(4).MOT2=2;
	  LHparticles.at(4).ISTUP=1;       
	}	
	smatrixuxgthuxddx_(p13[0],&me2);
      }
      else if (pflag==3) {
	if (flags.LHoutput) {
	  LHparticles.at(0).COL1=0;    // Incoming particle with +ve rapidity
	  LHparticles.at(0).COL2=501;
	  LHparticles.at(1).COL1=501;    // Incoming particle with -ve rapidity
	  LHparticles.at(1).COL2=502;
	  LHparticles.at(2).COL1=0;    // Outgoing particle with -ve rapidity
	  LHparticles.at(2).COL2=502;
	  LHparticles.at(2).p=p0;
	  LHparticles.at(2).PID=-flav;
	  LHparticles.at(2).MOT1=1;
	  LHparticles.at(2).MOT2=2;
	  LHparticles.at(2).ISTUP=1;
	  LHparticles.at(3).COL1=503;     // "Extra" parton
	  LHparticles.at(3).COL2=0;
	  LHparticles.at(3).p=p1;
	  LHparticles.at(3).PID=flav;    
	  LHparticles.at(3).MOT1=1;
	  LHparticles.at(3).MOT2=2;
	  LHparticles.at(3).ISTUP=1;
	  LHparticles.at(4).COL1=0;    // Outgoing parton with +ve rapidity 
	  LHparticles.at(4).COL2=503;
	  LHparticles.at(4).p=p2;
	  LHparticles.at(4).PID=bptype;
	  LHparticles.at(4).MOT1=1;
	  LHparticles.at(4).MOT2=2;
	  LHparticles.at(4).ISTUP=1;       
	}	
	smatrixuxgthuxddx_(p14[0],&me2);
      }
      else if (pflag==4) {
	if (flags.LHoutput) {
	  LHparticles.at(0).COL1=0;    // Incoming particle with +ve rapidity
	  LHparticles.at(0).COL2=501;
	  LHparticles.at(1).COL1=501;    // Incoming particle with -ve rapidity
	  LHparticles.at(1).COL2=502;
	  LHparticles.at(2).COL1=503;    // Outgoing particle with -ve rapidity
	  LHparticles.at(2).COL2=0;
	  LHparticles.at(2).p=p0;
	  LHparticles.at(2).PID=flav;
	  LHparticles.at(2).MOT1=1;
	  LHparticles.at(2).MOT2=2;
	  LHparticles.at(2).ISTUP=1;
	  LHparticles.at(3).COL1=0;     // "Extra" parton
	  LHparticles.at(3).COL2=503;
	  LHparticles.at(3).p=p1;
	  LHparticles.at(3).PID=bptype;    
	  LHparticles.at(3).MOT1=1;
	  LHparticles.at(3).MOT2=2;
	  LHparticles.at(3).ISTUP=1;
	  LHparticles.at(4).COL1=0;    // Outgoing parton with +ve rapidity 
	  LHparticles.at(4).COL2=502;
	  LHparticles.at(4).p=p2;
	  LHparticles.at(4).PID=-flav;
	  LHparticles.at(4).MOT1=1;
	  LHparticles.at(4).MOT2=2;
	  LHparticles.at(4).ISTUP=1;       
	}	
	smatrixuxgthuxddx_(p15[0],&me2);
      }
      else if (pflag==5) {
	if (flags.LHoutput) {
	  LHparticles.at(0).COL1=0;    // Incoming particle with +ve rapidity
	  LHparticles.at(0).COL2=501;
	  LHparticles.at(1).COL1=501;    // Incoming particle with -ve rapidity
	  LHparticles.at(1).COL2=502;
	  LHparticles.at(2).COL1=0;    // Outgoing particle with -ve rapidity
	  LHparticles.at(2).COL2=503;
	  LHparticles.at(2).p=p0;
	  LHparticles.at(2).PID=bptype;
	  LHparticles.at(2).MOT1=1;
	  LHparticles.at(2).MOT2=2;
	  LHparticles.at(2).ISTUP=1;
	  LHparticles.at(3).COL1=0;     // "Extra" parton
	  LHparticles.at(3).COL2=502;
	  LHparticles.at(3).p=p1;
	  LHparticles.at(3).PID=-flav;    
	  LHparticles.at(3).MOT1=1;
	  LHparticles.at(3).MOT2=2;
	  LHparticles.at(3).ISTUP=1;
	  LHparticles.at(4).COL1=503;    // Outgoing parton with +ve rapidity 
	  LHparticles.at(4).COL2=0;
	  LHparticles.at(4).p=p2;
	  LHparticles.at(4).PID=flav;
	  LHparticles.at(4).MOT1=1;
	  LHparticles.at(4).MOT2=2;
	  LHparticles.at(4).ISTUP=1;       
	}	
	smatrixuxgthuxddx_(p16[0],&me2);
      }
    }
    else if (process==1) { // if uxgtuuxux
      int pflag=int(3.*ran.flat()); // Labels rapidity configuration (0,1,2)
      weight*=3.;
      if(pflag==0) {
	if (flags.LHoutput) {
	  LHparticles.at(0).COL1=0;    // Incoming particle with +ve rapidity
	  LHparticles.at(0).COL2=501;
	  LHparticles.at(1).COL1=501;    // Incoming particle with -ve rapidity
	  LHparticles.at(1).COL2=502;
	  LHparticles.at(2).COL1=0;    // Outgoing particle with -ve rapidity
	  LHparticles.at(2).COL2=503;
	  LHparticles.at(2).p=p0;
	  LHparticles.at(2).PID=bptype;
	  LHparticles.at(2).MOT1=1;
	  LHparticles.at(2).MOT2=2;
	  LHparticles.at(2).ISTUP=1;
	  LHparticles.at(3).COL1=0;     // "Extra" parton
	  LHparticles.at(3).COL2=502;
	  LHparticles.at(3).p=p1;
	  LHparticles.at(3).PID=bptype;    
	  LHparticles.at(3).MOT1=1;
	  LHparticles.at(3).MOT2=2;
	  LHparticles.at(3).ISTUP=1;
	  LHparticles.at(4).COL1=503;    // Outgoing parton with +ve rapidity 
	  LHparticles.at(4).COL2=0;
	  LHparticles.at(4).p=p2;
	  LHparticles.at(4).PID=-bptype;
	  LHparticles.at(4).MOT1=1;
	  LHparticles.at(4).MOT2=2;
	  LHparticles.at(4).ISTUP=1;       
	}	
	smatrixuxgthuuxux_(p11[0],&me2);
	me2*=2.; // 2 identical antiquarks in final state
      }
      else if (pflag==1) {
	if (flags.LHoutput) {
	  LHparticles.at(0).COL1=0;    // Incoming particle with +ve rapidity
	  LHparticles.at(0).COL2=501;
	  LHparticles.at(1).COL1=501;    // Incoming particle with -ve rapidity
	  LHparticles.at(1).COL2=502;
	  LHparticles.at(2).COL1=503;    // Outgoing particle with -ve rapidity
	  LHparticles.at(2).COL2=0;
	  LHparticles.at(2).p=p0;
	  LHparticles.at(2).PID=-bptype;
	  LHparticles.at(2).MOT1=1;
	  LHparticles.at(2).MOT2=2;
	  LHparticles.at(2).ISTUP=1;
	  LHparticles.at(3).COL1=0;     // "Extra" parton
	  LHparticles.at(3).COL2=503;
	  LHparticles.at(3).p=p1;
	  LHparticles.at(3).PID=bptype;    
	  LHparticles.at(3).MOT1=1;
	  LHparticles.at(3).MOT2=2;
	  LHparticles.at(3).ISTUP=1;
	  LHparticles.at(4).COL1=0;    // Outgoing parton with +ve rapidity 
	  LHparticles.at(4).COL2=502;
	  LHparticles.at(4).p=p2;
	  LHparticles.at(4).PID=bptype;
	  LHparticles.at(4).MOT1=1;
	  LHparticles.at(4).MOT2=2;
	  LHparticles.at(4).ISTUP=1;       
	}	
	smatrixuxgthuuxux_(p12[0],&me2);
	me2*=2.;
      }
      else  { // pflag==2
	if (flags.LHoutput) {
	  LHparticles.at(0).COL1=0;    // Incoming particle with +ve rapidity
	  LHparticles.at(0).COL2=501;
	  LHparticles.at(1).COL1=501;    // Incoming particle with -ve rapidity
	  LHparticles.at(1).COL2=502;
	  LHparticles.at(2).COL1=0;    // Outgoing particle with -ve rapidity
	  LHparticles.at(2).COL2=502;
	  LHparticles.at(2).p=p0;
	  LHparticles.at(2).PID=bptype;
	  LHparticles.at(2).MOT1=1;
	  LHparticles.at(2).MOT2=2;
	  LHparticles.at(2).ISTUP=1;
	  LHparticles.at(3).COL1=503;     // "Extra" parton
	  LHparticles.at(3).COL2=0;
	  LHparticles.at(3).p=p1;
	  LHparticles.at(3).PID=-bptype;    
	  LHparticles.at(3).MOT1=1;
	  LHparticles.at(3).MOT2=2;
	  LHparticles.at(3).ISTUP=1;
	  LHparticles.at(4).COL1=0;    // Outgoing parton with +ve rapidity 
	  LHparticles.at(4).COL2=503;
	  LHparticles.at(4).p=p2;
	  LHparticles.at(4).PID=bptype;
	  LHparticles.at(4).MOT1=1;
	  LHparticles.at(4).MOT2=2;
	  LHparticles.at(4).ISTUP=1;       
	}		  
	smatrixuxgthuuxux_(p13[0],&me2);
	me2*=2.;
      }
    }
    else if (process==2) { // If uxgthuxgg
      int pflag=int(2.*ran.flat()); // 2 rapidity configurations that are non-FKL like
      weight*=2.;
      if (pflag==0) {
	if (flags.LHoutput) {
	  LHparticles.at(0).COL1=0;    // Incoming particle with +ve rapidity
	  LHparticles.at(0).COL2=501;
	  LHparticles.at(1).COL1=501;    // Incoming particle with -ve rapidity
	  LHparticles.at(1).COL2=502;
	  LHparticles.at(2).COL1=503;    // Outgoing particle with -ve rapidity
	  LHparticles.at(2).COL2=502;
	  LHparticles.at(2).p=p0;
	  LHparticles.at(2).PID=21;
	  LHparticles.at(2).MOT1=1;
	  LHparticles.at(2).MOT2=2;
	  LHparticles.at(2).ISTUP=1;
	  LHparticles.at(3).COL1=0;     // "Extra" parton
	  LHparticles.at(3).COL2=504;
	  LHparticles.at(3).p=p1;
	  LHparticles.at(3).PID=bptype;    
	  LHparticles.at(3).MOT1=1;
	  LHparticles.at(3).MOT2=2;
	  LHparticles.at(3).ISTUP=1;
	  LHparticles.at(4).COL1=504;    // Outgoing parton with +ve rapidity 
	  LHparticles.at(4).COL2=503;
	  LHparticles.at(4).p=p2;
	  LHparticles.at(4).PID=21;
	  LHparticles.at(4).MOT1=1;
	  LHparticles.at(4).MOT2=2;
	  LHparticles.at(4).ISTUP=1;       
	}	
	smatrixuxgthuxgg_(p12[0],&me2);
	me2*=2.;
      }
      else if (pflag==1) {
	if (flags.LHoutput) {
	  LHparticles.at(0).COL1=0;    // Incoming particle with +ve rapidity
	  LHparticles.at(0).COL2=501;
	  LHparticles.at(1).COL1=501;    // Incoming particle with -ve rapidity
	  LHparticles.at(1).COL2=502;
	  LHparticles.at(2).COL1=0;    // Outgoing particle with -ve rapidity
	  LHparticles.at(2).COL2=503;
	  LHparticles.at(2).p=p0;
	  LHparticles.at(2).PID=bptype;
	  LHparticles.at(2).MOT1=1;
	  LHparticles.at(2).MOT2=2;
	  LHparticles.at(2).ISTUP=1;
	  LHparticles.at(3).COL1=503;     // "Extra" parton
	  LHparticles.at(3).COL2=504;
	  LHparticles.at(3).p=p1;
	  LHparticles.at(3).PID=21;    
	  LHparticles.at(3).MOT1=1;
	  LHparticles.at(3).MOT2=2;
	  LHparticles.at(3).ISTUP=1;
	  LHparticles.at(4).COL1=504;    // Outgoing parton with +ve rapidity 
	  LHparticles.at(4).COL2=502;
	  LHparticles.at(4).p=p2;
	  LHparticles.at(4).PID=21;
	  LHparticles.at(4).MOT1=1;
	  LHparticles.at(4).MOT2=2;
	  LHparticles.at(4).ISTUP=1;       
	}		  
	smatrixuxgthuxgg_(p13[0],&me2);
	me2*=2.;
      }
      
    }
  }
  else if (bptype==21&&aptype<0) { // if qxg initial state
    int process=int(3.*ran.flat()); // labels partonic subprocess (0,1,2)
    weight*=3.;
    if(process==0){ // If uxgthuxddx
      int pflag=int(6.*ran.flat()); // labels 6 rapidity configuration (0,..,5)
      int flav=((int(4.*ran.flat())+bptype)%5)+1; // Choose a random flavour that is not the input flavour
      weight*=6.*4.; // 6 rapidity configs and 4 possible final state flavour choices
      
      if (pflag==0) {
	if (flags.LHoutput) {
	  LHparticles.at(0).COL1=501;    // Incoming particle with +ve rapidity
	  LHparticles.at(0).COL2=502;
	  LHparticles.at(1).COL1=0;    // Incoming particle with -ve rapidity
	  LHparticles.at(1).COL2=501;
	  LHparticles.at(2).COL1=0;    // Outgoing particle with -ve rapidity
	  LHparticles.at(2).COL2=503;
	  LHparticles.at(2).p=p0;
	  LHparticles.at(2).PID=aptype;
	  LHparticles.at(2).MOT1=1;
	  LHparticles.at(2).MOT2=2;
	  LHparticles.at(2).ISTUP=1;
	  LHparticles.at(3).COL1=0;     // "Extra" parton
	  LHparticles.at(3).COL2=502;
	  LHparticles.at(3).p=p1;
	  LHparticles.at(3).PID=-flav;    
	  LHparticles.at(3).MOT1=1;
	  LHparticles.at(3).MOT2=2;
	  LHparticles.at(3).ISTUP=1;
	  LHparticles.at(4).COL1=503;    // Outgoing parton with +ve rapidity 
	  LHparticles.at(4).COL2=0;
	  LHparticles.at(4).p=p2;
	  LHparticles.at(4).PID=flav;
	  LHparticles.at(4).MOT1=1;
	  LHparticles.at(4).MOT2=2;
	  LHparticles.at(4).ISTUP=1;       
	}	
	smatrixuxgthuxddx_(p21[0],&me2);
      }
      else if (pflag==1) {
	if (flags.LHoutput) {
	  LHparticles.at(0).COL1=501;    // Incoming particle with +ve rapidity
	  LHparticles.at(0).COL2=502;
	  LHparticles.at(1).COL1=0;    // Incoming particle with -ve rapidity
	  LHparticles.at(1).COL2=501;
	  LHparticles.at(2).COL1=503;    // Outgoing particle with -ve rapidity
	  LHparticles.at(2).COL2=0;
	  LHparticles.at(2).p=p0;
	  LHparticles.at(2).PID=flav;
	  LHparticles.at(2).MOT1=1;
	  LHparticles.at(2).MOT2=2;
	  LHparticles.at(2).ISTUP=1;
	  LHparticles.at(3).COL1=0;     // "Extra" parton
	  LHparticles.at(3).COL2=503;
	  LHparticles.at(3).p=p1;
	  LHparticles.at(3).PID=aptype;    
	  LHparticles.at(3).MOT1=1;
	  LHparticles.at(3).MOT2=2;
	  LHparticles.at(3).ISTUP=1;
	  LHparticles.at(4).COL1=0;    // Outgoing parton with +ve rapidity 
	  LHparticles.at(4).COL2=502;
	  LHparticles.at(4).p=p2;
	  LHparticles.at(4).PID=-flav;
	  LHparticles.at(4).MOT1=1;
	  LHparticles.at(4).MOT2=2;
	  LHparticles.at(4).ISTUP=1;       
	}	
	smatrixuxgthuxddx_(p22[0],&me2);	  
      }
      else if (pflag==2) {
	if (flags.LHoutput) {
	  LHparticles.at(0).COL1=501;    // Incoming particle with +ve rapidity
	  LHparticles.at(0).COL2=502;
	  LHparticles.at(1).COL1=0;    // Incoming particle with -ve rapidity
	  LHparticles.at(1).COL2=501;
	  LHparticles.at(2).COL1=0;    // Outgoing particle with -ve rapidity
	  LHparticles.at(2).COL2=502;
	  LHparticles.at(2).p=p0;
	  LHparticles.at(2).PID=-flav;
	  LHparticles.at(2).MOT1=1;
	  LHparticles.at(2).MOT2=2;
	  LHparticles.at(2).ISTUP=1;
	  LHparticles.at(3).COL1=503;     // "Extra" parton
	  LHparticles.at(3).COL2=0;
	  LHparticles.at(3).p=p1;
	  LHparticles.at(3).PID=flav;    
	  LHparticles.at(3).MOT1=1;
	  LHparticles.at(3).MOT2=2;
	  LHparticles.at(3).ISTUP=1;
	  LHparticles.at(4).COL1=0;    // Outgoing parton with +ve rapidity 
	  LHparticles.at(4).COL2=503;
	  LHparticles.at(4).p=p2;
	  LHparticles.at(4).PID=aptype;
	  LHparticles.at(4).MOT1=1;
	  LHparticles.at(4).MOT2=2;
	  LHparticles.at(4).ISTUP=1;       
	}	
	smatrixuxgthuxddx_(p23[0],&me2);
      }
      else if (pflag==3) {
	if (flags.LHoutput) {
	  LHparticles.at(0).COL1=501;    // Incoming particle with +ve rapidity
	  LHparticles.at(0).COL2=502;
	  LHparticles.at(1).COL1=0;    // Incoming particle with -ve rapidity
	  LHparticles.at(1).COL2=501;
	  LHparticles.at(2).COL1=0;    // Outgoing particle with -ve rapidity
	  LHparticles.at(2).COL2=503;
	  LHparticles.at(2).p=p0;
	  LHparticles.at(2).PID=aptype;
	  LHparticles.at(2).MOT1=1;
	  LHparticles.at(2).MOT2=2;
	  LHparticles.at(2).ISTUP=1;
	  LHparticles.at(3).COL1=503;     // "Extra" parton
	  LHparticles.at(3).COL2=0;
	  LHparticles.at(3).p=p1;
	  LHparticles.at(3).PID=flav;    
	  LHparticles.at(3).MOT1=1;
	  LHparticles.at(3).MOT2=2;
	  LHparticles.at(3).ISTUP=1;
	  LHparticles.at(4).COL1=0;    // Outgoing parton with +ve rapidity 
	  LHparticles.at(4).COL2=502;
	  LHparticles.at(4).p=p2;
	  LHparticles.at(4).PID=-flav;
	  LHparticles.at(4).MOT1=1;
	  LHparticles.at(4).MOT2=2;
	  LHparticles.at(4).ISTUP=1;       
	}	
	smatrixuxgthuxddx_(p24[0],&me2);
      }
      else if (pflag==4) {
	if (flags.LHoutput) {
	  LHparticles.at(0).COL1=501;    // Incoming particle with +ve rapidity
	  LHparticles.at(0).COL2=502;
	  LHparticles.at(1).COL1=0;    // Incoming particle with -ve rapidity
	  LHparticles.at(1).COL2=501;
	  LHparticles.at(2).COL1=0;    // Outgoing particle with -ve rapidity
	  LHparticles.at(2).COL2=502;
	  LHparticles.at(2).p=p0;
	  LHparticles.at(2).PID=-flav;
	  LHparticles.at(2).MOT1=1;
	  LHparticles.at(2).MOT2=2;
	  LHparticles.at(2).ISTUP=1;
	  LHparticles.at(3).COL1=0;     // "Extra" parton
	  LHparticles.at(3).COL2=503;
	  LHparticles.at(3).p=p1;
	  LHparticles.at(3).PID=aptype;    
	  LHparticles.at(3).MOT1=1;
	  LHparticles.at(3).MOT2=2;
	  LHparticles.at(3).ISTUP=1;
	  LHparticles.at(4).COL1=503;    // Outgoing parton with +ve rapidity 
	  LHparticles.at(4).COL2=0;
	  LHparticles.at(4).p=p2;
	  LHparticles.at(4).PID=flav;
	  LHparticles.at(4).MOT1=1;
	  LHparticles.at(4).MOT2=2;
	  LHparticles.at(4).ISTUP=1;       
	}	
	smatrixuxgthuxddx_(p25[0],&me2);
      }
      else if (pflag==5) {
	if (flags.LHoutput) {
	  LHparticles.at(0).COL1=501;    // Incoming particle with +ve rapidity
	  LHparticles.at(0).COL2=502;
	  LHparticles.at(1).COL1=0;    // Incoming particle with -ve rapidity
	  LHparticles.at(1).COL2=501;
	  LHparticles.at(2).COL1=503;    // Outgoing particle with -ve rapidity
	  LHparticles.at(2).COL2=0;
	  LHparticles.at(2).p=p0;
	  LHparticles.at(2).PID=flav;
	  LHparticles.at(2).MOT1=1;
	  LHparticles.at(2).MOT2=2;
	  LHparticles.at(2).ISTUP=1;
	  LHparticles.at(3).COL1=0;     // "Extra" parton
	  LHparticles.at(3).COL2=502;
	  LHparticles.at(3).p=p1;
	  LHparticles.at(3).PID=-flav;    
	  LHparticles.at(3).MOT1=1;
	  LHparticles.at(3).MOT2=2;
	  LHparticles.at(3).ISTUP=1;
	  LHparticles.at(4).COL1=0;    // Outgoing parton with +ve rapidity 
	  LHparticles.at(4).COL2=503;
	  LHparticles.at(4).p=p2;
	  LHparticles.at(4).PID=aptype;
	  LHparticles.at(4).MOT1=1;
	  LHparticles.at(4).MOT2=2;
	  LHparticles.at(4).ISTUP=1;       
	}	
	smatrixuxgthuxddx_(p26[0],&me2);
      }
    }
    else if (process==1) { // if uxgtuuxux
      int pflag=int(3.*ran.flat()); // Labels rapidity configuration (0,1,2)
      weight*=3.;
      if (pflag==0) {
	if(flags.LHoutput) {
	  LHparticles.at(0).COL1=501;    // Incoming particle with +ve rapidity
	  LHparticles.at(0).COL2=502;
	  LHparticles.at(1).COL1=0;    // Incoming particle with -ve rapidity
	  LHparticles.at(1).COL2=501;
	  LHparticles.at(2).COL1=503;    // Outgoing particle with -ve rapidity
	  LHparticles.at(2).COL2=0;
	  LHparticles.at(2).p=p0;
	  LHparticles.at(2).PID=-aptype;
	  LHparticles.at(2).MOT1=1;
	  LHparticles.at(2).MOT2=2;
	  LHparticles.at(2).ISTUP=1;
	  LHparticles.at(3).COL1=0;     // "Extra" parton
	  LHparticles.at(3).COL2=503;
	  LHparticles.at(3).p=p1;
	  LHparticles.at(3).PID=aptype;    
	  LHparticles.at(3).MOT1=1;
	  LHparticles.at(3).MOT2=2;
	  LHparticles.at(3).ISTUP=1;
	  LHparticles.at(4).COL1=0;    // Outgoing parton with +ve rapidity 
	  LHparticles.at(4).COL2=502;
	  LHparticles.at(4).p=p2;
	  LHparticles.at(4).PID=aptype;
	  LHparticles.at(4).MOT1=1;
	  LHparticles.at(4).MOT2=2;
	  LHparticles.at(4).ISTUP=1;       
	}	
	smatrixuxgthuuxux_(p21[0],&me2);
	me2*=2.;
	}
	else if (pflag==1) {
	  if (flags.LHoutput) {
	    LHparticles.at(0).COL1=501;    // Incoming particle with +ve rapidity
	    LHparticles.at(0).COL2=502;
	    LHparticles.at(1).COL1=0;    // Incoming particle with -ve rapidity
	    LHparticles.at(1).COL2=501;
	    LHparticles.at(2).COL1=0;    // Outgoing particle with -ve rapidity
	    LHparticles.at(2).COL2=502;
	    LHparticles.at(2).p=p0;
	    LHparticles.at(2).PID=aptype;
	    LHparticles.at(2).MOT1=1;
	    LHparticles.at(2).MOT2=2;
	    LHparticles.at(2).ISTUP=1;
	    LHparticles.at(3).COL1=503;     // "Extra" parton
	    LHparticles.at(3).COL2=0;
	    LHparticles.at(3).p=p1;
	    LHparticles.at(3).PID=-aptype;    
	    LHparticles.at(3).MOT1=1;
	    LHparticles.at(3).MOT2=2;
	    LHparticles.at(3).ISTUP=1;
	    LHparticles.at(4).COL1=0;    // Outgoing parton with +ve rapidity 
	    LHparticles.at(4).COL2=503;
	    LHparticles.at(4).p=p2;
	    LHparticles.at(4).PID=aptype;
	    LHparticles.at(4).MOT1=1;
	    LHparticles.at(4).MOT2=2;
	    LHparticles.at(4).ISTUP=1;       
	  }	
	  smatrixuxgthuuxux_(p22[0],&me2);
	  me2*=2.;
	}
	else if (pflag==2) {
	  if (flags.LHoutput) {
	    LHparticles.at(0).COL1=501;    // Incoming particle with +ve rapidity
	    LHparticles.at(0).COL2=502;
	    LHparticles.at(1).COL1=0;    // Incoming particle with -ve rapidity
	    LHparticles.at(1).COL2=501;
	    LHparticles.at(2).COL1=0;    // Outgoing particle with -ve rapidity
	    LHparticles.at(2).COL2=503;
	    LHparticles.at(2).p=p0;
	    LHparticles.at(2).PID=aptype;
	    LHparticles.at(2).MOT1=1;
	    LHparticles.at(2).MOT2=2;
	    LHparticles.at(2).ISTUP=1;
	    LHparticles.at(3).COL1=0;     // "Extra" parton
	    LHparticles.at(3).COL2=502;
	    LHparticles.at(3).p=p1;
	    LHparticles.at(3).PID=aptype;    
	    LHparticles.at(3).MOT1=1;
	    LHparticles.at(3).MOT2=2;
	    LHparticles.at(3).ISTUP=1;
	    LHparticles.at(4).COL1=503;    // Outgoing parton with +ve rapidity 
	    LHparticles.at(4).COL2=0;
	    LHparticles.at(4).p=p2;
	    LHparticles.at(4).PID=-aptype;
	    LHparticles.at(4).MOT1=1;
	    LHparticles.at(4).MOT2=2;
	    LHparticles.at(4).ISTUP=1;       
	  }		 
	  smatrixuxgthuuxux_(p23[0],&me2);
	  me2*=2.;
	}
      }
      else if (process==2) { // If uxgthuxgg
	int pflag=int(2.*ran.flat()); // 2 rapidity configurations that are non-FKL like
	weight*=2.;
	if (pflag==0) {
	  if (flags.LHoutput) {
	    LHparticles.at(0).COL1=501;    // Incoming particle with +ve rapidity
	    LHparticles.at(0).COL2=502;
	    LHparticles.at(1).COL1=0;    // Incoming particle with -ve rapidity
	    LHparticles.at(1).COL2=501;
	    LHparticles.at(2).COL1=503;    // Outgoing particle with -ve rapidity
	    LHparticles.at(2).COL2=504;
	    LHparticles.at(2).p=p0;
	    LHparticles.at(2).PID=21;
	    LHparticles.at(2).MOT1=1;
	    LHparticles.at(2).MOT2=2;
	    LHparticles.at(2).ISTUP=1;
	    LHparticles.at(3).COL1=0;     // "Extra" parton
	    LHparticles.at(3).COL2=503;
	    LHparticles.at(3).p=p1;
	    LHparticles.at(3).PID=aptype;    
	    LHparticles.at(3).MOT1=1;
	    LHparticles.at(3).MOT2=2;
	    LHparticles.at(3).ISTUP=1;
	    LHparticles.at(4).COL1=504;    // Outgoing parton with +ve rapidity 
	    LHparticles.at(4).COL2=502;
	    LHparticles.at(4).p=p2;
	    LHparticles.at(4).PID=21;
	    LHparticles.at(4).MOT1=1;
	    LHparticles.at(4).MOT2=2;
	    LHparticles.at(4).ISTUP=1;       
	  }	
	  smatrixuxgthuxgg_(p22[0],&me2);
	  me2*=2.;
	}
	else if (pflag==1) {
	  if (flags.LHoutput) {
	    LHparticles.at(0).COL1=501;    // Incoming particle with +ve rapidity
	    LHparticles.at(0).COL2=502;
	    LHparticles.at(1).COL1=0;    // Incoming particle with -ve rapidity
	    LHparticles.at(1).COL2=501;
	    LHparticles.at(2).COL1=503;    // Outgoing particle with -ve rapidity
	    LHparticles.at(2).COL2=502;
	    LHparticles.at(2).p=p0;
	    LHparticles.at(2).PID=21;
	    LHparticles.at(2).MOT1=1;
	    LHparticles.at(2).MOT2=2;
	    LHparticles.at(2).ISTUP=1;
	    LHparticles.at(3).COL1=504;     // "Extra" parton
	    LHparticles.at(3).COL2=503;
	    LHparticles.at(3).p=p1;
	    LHparticles.at(3).PID=21;    
	    LHparticles.at(3).MOT1=1;
	    LHparticles.at(3).MOT2=2;
	    LHparticles.at(3).ISTUP=1;
	    LHparticles.at(4).COL1=0;    // Outgoing parton with +ve rapidity 
	    LHparticles.at(4).COL2=504;
	    LHparticles.at(4).p=p2;
	    LHparticles.at(4).PID=aptype;
	    LHparticles.at(4).MOT1=1;
	    LHparticles.at(4).MOT2=2;
	    LHparticles.at(4).ISTUP=1;       
	  }		  
	  smatrixuxgthuxgg_(p23[0],&me2);
	  me2*=2.;
	}
      }
    }
    else if (aptype>0&&bptype>0&&aptype!=21&&bptype!=21&&aptype!=bptype) { // If ud initial state
      int pflag=int(5.*ran.flat()); // 5 different rapidity orderings (non-FKL)
      weight*=5.;
      if (pflag==0) {
	if (flags.LHoutput) {
	  LHparticles.at(0).COL1=501;    // Incoming particle with +ve rapidity
	  LHparticles.at(0).COL2=0;
	  LHparticles.at(1).COL1=502;    // Incoming particle with -ve rapidity
	  LHparticles.at(1).COL2=0;
	  LHparticles.at(2).COL1=502;    // Outgoing particle with -ve rapidity
	  LHparticles.at(2).COL2=0;
	  LHparticles.at(2).p=p0;
	  LHparticles.at(2).PID=bptype;
	  LHparticles.at(2).MOT1=1;
	  LHparticles.at(2).MOT2=2;
	  LHparticles.at(2).ISTUP=1;
	  LHparticles.at(3).COL1=503;     // "Extra" parton
	  LHparticles.at(3).COL2=0;
	  LHparticles.at(3).p=p1;
	  LHparticles.at(3).PID=aptype;    
	  LHparticles.at(3).MOT1=1;
	  LHparticles.at(3).MOT2=2;
	  LHparticles.at(3).ISTUP=1;
	  LHparticles.at(4).COL1=501;    // Outgoing parton with +ve rapidity 
	  LHparticles.at(4).COL2=503;
	  LHparticles.at(4).p=p2;
	  LHparticles.at(4).PID=21;
	  LHparticles.at(4).MOT1=1;
	  LHparticles.at(4).MOT2=2;
	  LHparticles.at(4).ISTUP=1;       
	}
	smatrixudthudg_(p22[0],&me2);
      }
      else if (pflag==1) {
	if (flags.LHoutput) {
	  LHparticles.at(0).COL1=501;    // Incoming particle with +ve rapidity
	  LHparticles.at(0).COL2=0;
	  LHparticles.at(1).COL1=502;    // Incoming particle with -ve rapidity
	  LHparticles.at(1).COL2=0;
	  LHparticles.at(2).COL1=502;    // Outgoing particle with -ve rapidity
	  LHparticles.at(2).COL2=503;
	  LHparticles.at(2).p=p0;
	  LHparticles.at(2).PID=21;
	  LHparticles.at(2).MOT1=1;
	  LHparticles.at(2).MOT2=2;
	  LHparticles.at(2).ISTUP=1;
	  LHparticles.at(3).COL1=501;     // "Extra" parton
	  LHparticles.at(3).COL2=0;
	  LHparticles.at(3).p=p1;
	  LHparticles.at(3).PID=bptype;    
	  LHparticles.at(3).MOT1=1;
	  LHparticles.at(3).MOT2=2;
	  LHparticles.at(3).ISTUP=1;
	  LHparticles.at(4).COL1=503;    // Outgoing parton with +ve rapidity 
	  LHparticles.at(4).COL2=0;
	  LHparticles.at(4).p=p2;
	  LHparticles.at(4).PID=aptype;
	  LHparticles.at(4).MOT1=1;
	  LHparticles.at(4).MOT2=2;
	  LHparticles.at(4).ISTUP=1;       
	}
	smatrixudthudg_(p23[0],&me2);
      }
      else if (pflag==2) {
	if (flags.LHoutput) {
	  LHparticles.at(0).COL1=501;    // Incoming particle with +ve rapidity
	  LHparticles.at(0).COL2=0;
	  LHparticles.at(1).COL1=502;    // Incoming particle with -ve rapidity
	  LHparticles.at(1).COL2=0;
	  LHparticles.at(2).COL1=503;    // Outgoing particle with -ve rapidity
	  LHparticles.at(2).COL2=0;
	  LHparticles.at(2).p=p0;
	  LHparticles.at(2).PID=aptype;
	  LHparticles.at(2).MOT1=1;
	  LHparticles.at(2).MOT2=2;
	  LHparticles.at(2).ISTUP=1;
	  LHparticles.at(3).COL1=501;     // "Extra" parton
	  LHparticles.at(3).COL2=0;
	  LHparticles.at(3).p=p1;
	  LHparticles.at(3).PID=bptype;    
	  LHparticles.at(3).MOT1=1;
	  LHparticles.at(3).MOT2=2;
	  LHparticles.at(3).ISTUP=1;
	  LHparticles.at(4).COL1=502;    // Outgoing parton with +ve rapidity 
	  LHparticles.at(4).COL2=503;
	  LHparticles.at(4).p=p2;
	  LHparticles.at(4).PID=21;
	  LHparticles.at(4).MOT1=1;
	  LHparticles.at(4).MOT2=2;
	  LHparticles.at(4).ISTUP=1;       
	}
	smatrixudthudg_(p24[0],&me2);
      }
      else if (pflag==3) {
	if (flags.LHoutput) {
	  LHparticles.at(0).COL1=501;    // Incoming particle with +ve rapidity
	  LHparticles.at(0).COL2=0;
	  LHparticles.at(1).COL1=502;    // Incoming particle with -ve rapidity
	  LHparticles.at(1).COL2=0;
	  LHparticles.at(2).COL1=502;    // Outgoing particle with -ve rapidity
	  LHparticles.at(2).COL2=503;
	  LHparticles.at(2).p=p0;
	  LHparticles.at(2).PID=21;
	  LHparticles.at(2).MOT1=1;
	  LHparticles.at(2).MOT2=2;
	  LHparticles.at(2).ISTUP=1;
	  LHparticles.at(3).COL1=503;     // "Extra" parton
	  LHparticles.at(3).COL2=0;
	  LHparticles.at(3).p=p1;
	  LHparticles.at(3).PID=aptype;    
	  LHparticles.at(3).MOT1=1;
	  LHparticles.at(3).MOT2=2;
	  LHparticles.at(3).ISTUP=1;
	  LHparticles.at(4).COL1=501;    // Outgoing parton with +ve rapidity 
	  LHparticles.at(4).COL2=0;
	  LHparticles.at(4).p=p2;
	  LHparticles.at(4).PID=bptype;
	  LHparticles.at(4).MOT1=1;
	  LHparticles.at(4).MOT2=2;
	  LHparticles.at(4).ISTUP=1;       
	}
	smatrixudthudg_(p25[0],&me2);
      }
      else if (pflag==4) {
	if (flags.LHoutput) {
	  LHparticles.at(0).COL1=501;    // Incoming particle with +ve rapidity
	  LHparticles.at(0).COL2=0;
	  LHparticles.at(1).COL1=502;    // Incoming particle with -ve rapidity
	  LHparticles.at(1).COL2=0;
	  LHparticles.at(2).COL1=501;    // Outgoing particle with -ve rapidity
	  LHparticles.at(2).COL2=0;
	  LHparticles.at(2).p=p0;
	  LHparticles.at(2).PID=bptype;
	  LHparticles.at(2).MOT1=1;
	  LHparticles.at(2).MOT2=2;
	  LHparticles.at(2).ISTUP=1;
	  LHparticles.at(3).COL1=502;     // "Extra" parton
	  LHparticles.at(3).COL2=503;
	  LHparticles.at(3).p=p1;
	  LHparticles.at(3).PID=21;    
	  LHparticles.at(3).MOT1=1;
	  LHparticles.at(3).MOT2=2;
	  LHparticles.at(3).ISTUP=1;
	  LHparticles.at(4).COL1=503;    // Outgoing parton with +ve rapidity 
	  LHparticles.at(4).COL2=0;
	  LHparticles.at(4).p=p2;
	  LHparticles.at(4).PID=aptype;
	  LHparticles.at(4).MOT1=1;
	  LHparticles.at(4).MOT2=2;
	  LHparticles.at(4).ISTUP=1;       
	}
	smatrixudthudg_(p26[0],&me2);
      }
    }
    else if (aptype<0&&bptype<0&&aptype!=bptype) { // if uxdx initial state
      int pflag=int(5.*ran.flat()); // 5 rapidity configurations (non-FKL)
      weight*=5.;
      if(pflag==0) {
	if (flags.LHoutput) {
	  LHparticles.at(0).COL1=0;    // Incoming particle with +ve rapidity
	  LHparticles.at(0).COL2=501;
	  LHparticles.at(1).COL1=0;    // Incoming particle with -ve rapidity
	  LHparticles.at(1).COL2=502;
	  LHparticles.at(2).COL1=0;    // Outgoing particle with -ve rapidity
	  LHparticles.at(2).COL2=501;
	  LHparticles.at(2).p=p0;
	  LHparticles.at(2).PID=bptype;
	  LHparticles.at(2).MOT1=1;
	  LHparticles.at(2).MOT2=2;
	  LHparticles.at(2).ISTUP=1;
	  LHparticles.at(3).COL1=0;     // "Extra" parton
	  LHparticles.at(3).COL2=503;
	  LHparticles.at(3).p=p1;
	  LHparticles.at(3).PID=aptype;    
	  LHparticles.at(3).MOT1=1;
	  LHparticles.at(3).MOT2=2;
	  LHparticles.at(3).ISTUP=1;
	  LHparticles.at(4).COL1=503;    // Outgoing parton with +ve rapidity 
	  LHparticles.at(4).COL2=502;
	  LHparticles.at(4).p=p2;
	  LHparticles.at(4).PID=21;
	  LHparticles.at(4).MOT1=1;
	  LHparticles.at(4).MOT2=2;
	  LHparticles.at(4).ISTUP=1;       
	}	
	smatrixuxdxthuxdxg_(p22[0],&me2);
      }
      else if (pflag==1) {
	if (flags.LHoutput) {
	  LHparticles.at(0).COL1=0;    // Incoming particle with +ve rapidity
	  LHparticles.at(0).COL2=501;
	  LHparticles.at(1).COL1=0;    // Incoming particle with -ve rapidity
	  LHparticles.at(1).COL2=502;
	  LHparticles.at(2).COL1=503;    // Outgoing particle with -ve rapidity
	  LHparticles.at(2).COL2=502;
	  LHparticles.at(2).p=p0;
	  LHparticles.at(2).PID=21;
	  LHparticles.at(2).MOT1=1;
	  LHparticles.at(2).MOT2=2;
	  LHparticles.at(2).ISTUP=1;
	  LHparticles.at(3).COL1=0;     // "Extra" parton
	  LHparticles.at(3).COL2=501;
	  LHparticles.at(3).p=p1;
	  LHparticles.at(3).PID=bptype;    
	  LHparticles.at(3).MOT1=1;
	  LHparticles.at(3).MOT2=2;
	  LHparticles.at(3).ISTUP=1;
	  LHparticles.at(4).COL1=0;    // Outgoing parton with +ve rapidity 
	  LHparticles.at(4).COL2=503;
	  LHparticles.at(4).p=p2;
	  LHparticles.at(4).PID=aptype;
	  LHparticles.at(4).MOT1=1;
	  LHparticles.at(4).MOT2=2;
	  LHparticles.at(4).ISTUP=1;       
	}		
	smatrixuxdxthuxdxg_(p23[0],&me2);
      }
      else if (pflag==2) {
	if (flags.LHoutput) {
	  LHparticles.at(0).COL1=0;    // Incoming particle with +ve rapidity
	  LHparticles.at(0).COL2=501;
	  LHparticles.at(1).COL1=0;    // Incoming particle with -ve rapidity
	  LHparticles.at(1).COL2=502;
	  LHparticles.at(2).COL1=0;    // Outgoing particle with -ve rapidity
	  LHparticles.at(2).COL2=503;
	  LHparticles.at(2).p=p0;
	  LHparticles.at(2).PID=aptype;
	  LHparticles.at(2).MOT1=1;
	  LHparticles.at(2).MOT2=2;
	  LHparticles.at(2).ISTUP=1;
	  LHparticles.at(3).COL1=0;     // "Extra" parton
	  LHparticles.at(3).COL2=501;
	  LHparticles.at(3).p=p1;
	  LHparticles.at(3).PID=bptype;    
	  LHparticles.at(3).MOT1=1;
	  LHparticles.at(3).MOT2=2;
	  LHparticles.at(3).ISTUP=1;
	  LHparticles.at(4).COL1=503;    // Outgoing parton with +ve rapidity 
	  LHparticles.at(4).COL2=502;
	  LHparticles.at(4).p=p2;
	  LHparticles.at(4).PID=21;
	  LHparticles.at(4).MOT1=1;
	  LHparticles.at(4).MOT2=2;
	  LHparticles.at(4).ISTUP=1;       
	}		
	smatrixuxdxthuxdxg_(p24[0],&me2);
      }
      else if (pflag==3) {
	if (flags.LHoutput) {
	  LHparticles.at(0).COL1=0;    // Incoming particle with +ve rapidity
	  LHparticles.at(0).COL2=501;
	  LHparticles.at(1).COL1=0;    // Incoming particle with -ve rapidity
	  LHparticles.at(1).COL2=502;
	  LHparticles.at(2).COL1=503;    // Outgoing particle with -ve rapidity
	  LHparticles.at(2).COL2=502;
	  LHparticles.at(2).p=p0;
	  LHparticles.at(2).PID=21;
	  LHparticles.at(2).MOT1=1;
	  LHparticles.at(2).MOT2=2;
	  LHparticles.at(2).ISTUP=1;
	  LHparticles.at(3).COL1=0;     // "Extra" parton
	  LHparticles.at(3).COL2=503;
	  LHparticles.at(3).p=p1;
	  LHparticles.at(3).PID=aptype;    
	  LHparticles.at(3).MOT1=1;
	  LHparticles.at(3).MOT2=2;
	  LHparticles.at(3).ISTUP=1;
	  LHparticles.at(4).COL1=0;    // Outgoing parton with +ve rapidity 
	  LHparticles.at(4).COL2=501;
	  LHparticles.at(4).p=p2;
	  LHparticles.at(4).PID=bptype;
	  LHparticles.at(4).MOT1=1;
	  LHparticles.at(4).MOT2=2;
	  LHparticles.at(4).ISTUP=1;       
	}		
	smatrixuxdxthuxdxg_(p25[0],&me2);
      }
      else if (pflag==4) {
	if (flags.LHoutput) {
	  LHparticles.at(0).COL1=0;    // Incoming particle with +ve rapidity
	  LHparticles.at(0).COL2=501;
	  LHparticles.at(1).COL1=0;    // Incoming particle with -ve rapidity
	  LHparticles.at(1).COL2=502;
	  LHparticles.at(2).COL1=0;    // Outgoing particle with -ve rapidity
	  LHparticles.at(2).COL2=501;
	  LHparticles.at(2).p=p0;
	  LHparticles.at(2).PID=bptype;
	  LHparticles.at(2).MOT1=1;
	  LHparticles.at(2).MOT2=2;
	  LHparticles.at(2).ISTUP=1;
	  LHparticles.at(3).COL1=503;     // "Extra" parton
	  LHparticles.at(3).COL2=502;
	  LHparticles.at(3).p=p1;
	  LHparticles.at(3).PID=21;    
	  LHparticles.at(3).MOT1=1;
	  LHparticles.at(3).MOT2=2;
	  LHparticles.at(3).ISTUP=1;
	  LHparticles.at(4).COL1=0;    // Outgoing parton with +ve rapidity 
	  LHparticles.at(4).COL2=503;
	  LHparticles.at(4).p=p2;
	  LHparticles.at(4).PID=aptype;
	  LHparticles.at(4).MOT1=1;
	  LHparticles.at(4).MOT2=2;
	  LHparticles.at(4).ISTUP=1;       
	}	
	smatrixuxdxthuxdxg_(p26[0],&me2);
      }    
    }
    else if (aptype>0&&aptype!=21&&bptype<0&&aptype!=-bptype) { // If udx initial state
      int pflag=int(5.*ran.flat()); // 5 rapidity orderings (non-FKL)
      weight*=5.;
      if(pflag==0) {
	if (flags.LHoutput) {
	  LHparticles.at(0).COL1=0;    // Incoming particle with +ve rapidity
	  LHparticles.at(0).COL2=501;
	  LHparticles.at(1).COL1=501;    // Incoming particle with -ve rapidity
	  LHparticles.at(1).COL2=0;
	  LHparticles.at(2).COL1=0;    // Outgoing particle with -ve rapidity
	  LHparticles.at(2).COL2=502;
	  LHparticles.at(2).p=p0;
	  LHparticles.at(2).PID=bptype;
	  LHparticles.at(2).MOT1=1;
	  LHparticles.at(2).MOT2=2;
	  LHparticles.at(2).ISTUP=1;
	  LHparticles.at(3).COL1=503;     // "Extra" parton
	  LHparticles.at(3).COL2=0;
	  LHparticles.at(3).p=p1;
	  LHparticles.at(3).PID=aptype;    
	  LHparticles.at(3).MOT1=1;
	  LHparticles.at(3).MOT2=2;
	  LHparticles.at(3).ISTUP=1;
	  LHparticles.at(4).COL1=502;    // Outgoing parton with +ve rapidity 
	  LHparticles.at(4).COL2=503;
	  LHparticles.at(4).p=p2;
	  LHparticles.at(4).PID=21;
	  LHparticles.at(4).MOT1=1;
	  LHparticles.at(4).MOT2=2;
	  LHparticles.at(4).ISTUP=1;       
	}	
	smatrixudxthudxg_(p22[0],&me2);
      }
      else if (pflag==1) {
	if (flags.LHoutput) {
	  LHparticles.at(0).COL1=0;    // Incoming particle with +ve rapidity
	  LHparticles.at(0).COL2=501;
	  LHparticles.at(1).COL1=501;    // Incoming particle with -ve rapidity
	  LHparticles.at(1).COL2=0;
	  LHparticles.at(2).COL1=503;    // Outgoing particle with -ve rapidity
	  LHparticles.at(2).COL2=504;
	  LHparticles.at(2).p=p0;
	  LHparticles.at(2).PID=21;
	  LHparticles.at(2).MOT1=1;
	  LHparticles.at(2).MOT2=2;
	  LHparticles.at(2).ISTUP=1;
	  LHparticles.at(3).COL1=0;     // "Extra" parton
	  LHparticles.at(3).COL2=503;
	  LHparticles.at(3).p=p1;
	  LHparticles.at(3).PID=bptype;    
	  LHparticles.at(3).MOT1=1;
	  LHparticles.at(3).MOT2=2;
	  LHparticles.at(3).ISTUP=1;
	  LHparticles.at(4).COL1=504;    // Outgoing parton with +ve rapidity 
	  LHparticles.at(4).COL2=0;
	  LHparticles.at(4).p=p2;
	  LHparticles.at(4).PID=aptype;
	  LHparticles.at(4).MOT1=1;
	  LHparticles.at(4).MOT2=2;
	  LHparticles.at(4).ISTUP=1;       
	}	
	smatrixudxthudxg_(p23[0],&me2);
      }
      else if (pflag==2) {
	if (flags.LHoutput) {
	  LHparticles.at(0).COL1=0;    // Incoming particle with +ve rapidity
	  LHparticles.at(0).COL2=501;
	  LHparticles.at(1).COL1=501;    // Incoming particle with -ve rapidity
	  LHparticles.at(1).COL2=0;
	  LHparticles.at(2).COL1=504;    // Outgoing particle with -ve rapidity
	  LHparticles.at(2).COL2=0;
	  LHparticles.at(2).p=p0;
	  LHparticles.at(2).PID=aptype;
	  LHparticles.at(2).MOT1=1;
	  LHparticles.at(2).MOT2=2;
	  LHparticles.at(2).ISTUP=1;
	  LHparticles.at(3).COL1=0;     // "Extra" parton
	  LHparticles.at(3).COL2=503;
	  LHparticles.at(3).p=p1;
	  LHparticles.at(3).PID=bptype;    
	  LHparticles.at(3).MOT1=1;
	  LHparticles.at(3).MOT2=2;
	  LHparticles.at(3).ISTUP=1;
	  LHparticles.at(4).COL1=503;    // Outgoing parton with +ve rapidity 
	  LHparticles.at(4).COL2=504;
	  LHparticles.at(4).p=p2;
	  LHparticles.at(4).PID=21;
	  LHparticles.at(4).MOT1=1;
	  LHparticles.at(4).MOT2=2;
	  LHparticles.at(4).ISTUP=1;       
	}		
	smatrixudxthudxg_(p24[0],&me2);
      }
      else if (pflag==3) {
	if (flags.LHoutput) {
	  LHparticles.at(0).COL1=0;    // Incoming particle with +ve rapidity
	  LHparticles.at(0).COL2=501;
	  LHparticles.at(1).COL1=501;    // Incoming particle with -ve rapidity
	  LHparticles.at(1).COL2=0;
	  LHparticles.at(2).COL1=503;    // Outgoing particle with -ve rapidity
	  LHparticles.at(2).COL2=504;
	  LHparticles.at(2).p=p0;
	  LHparticles.at(2).PID=21;
	  LHparticles.at(2).MOT1=1;
	  LHparticles.at(2).MOT2=2;
	  LHparticles.at(2).ISTUP=1;
	  LHparticles.at(3).COL1=504;     // "Extra" parton
	  LHparticles.at(3).COL2=0;
	  LHparticles.at(3).p=p1;
	  LHparticles.at(3).PID=aptype;    
	  LHparticles.at(3).MOT1=1;
	  LHparticles.at(3).MOT2=2;
	  LHparticles.at(3).ISTUP=1;
	  LHparticles.at(4).COL1=0;    // Outgoing parton with +ve rapidity 
	  LHparticles.at(4).COL2=503;
	  LHparticles.at(4).p=p2;
	  LHparticles.at(4).PID=bptype;
	  LHparticles.at(4).MOT1=1;
	  LHparticles.at(4).MOT2=2;
	  LHparticles.at(4).ISTUP=1;       
	}		
	smatrixudxthudxg_(p25[0],&me2);
      }
      else if (pflag==4) {
	if (flags.LHoutput) {
	  LHparticles.at(0).COL1=0;    // Incoming particle with +ve rapidity
	  LHparticles.at(0).COL2=501;
	  LHparticles.at(1).COL1=501;    // Incoming particle with -ve rapidity
	  LHparticles.at(1).COL2=0;
	  LHparticles.at(2).COL1=0;    // Outgoing particle with -ve rapidity
	  LHparticles.at(2).COL2=503;
	  LHparticles.at(2).p=p0;
	  LHparticles.at(2).PID=bptype;
	  LHparticles.at(2).MOT1=1;
	  LHparticles.at(2).MOT2=2;
	  LHparticles.at(2).ISTUP=1;
	  LHparticles.at(3).COL1=503;     // "Extra" parton
	  LHparticles.at(3).COL2=504;
	  LHparticles.at(3).p=p1;
	  LHparticles.at(3).PID=21;    
	  LHparticles.at(3).MOT1=1;
	  LHparticles.at(3).MOT2=2;
	  LHparticles.at(3).ISTUP=1;
	  LHparticles.at(4).COL1=504;    // Outgoing parton with +ve rapidity 
	  LHparticles.at(4).COL2=0;
	  LHparticles.at(4).p=p2;
	  LHparticles.at(4).PID=aptype;
	  LHparticles.at(4).MOT1=1;
	  LHparticles.at(4).MOT2=2;
	  LHparticles.at(4).ISTUP=1;       
	}	
	smatrixudxthudxg_(p26[0],&me2);
      }
    }
    else if (aptype<0&&bptype>0&&bptype!=21&&aptype!=-bptype) { // If uxd initial state
      int pflag=int(5.*ran.flat()); // 5 rapidity orderings (non-FKL)
      weight*=5.;
      if (pflag==0) {
	if (flags.LHoutput) {
	  LHparticles.at(0).COL1=501;    // Incoming particle with +ve rapidity
	  LHparticles.at(0).COL2=0;
	  LHparticles.at(1).COL1=0;    // Incoming particle with -ve rapidity
	  LHparticles.at(1).COL2=501;
	  LHparticles.at(2).COL1=503;    // Outgoing particle with -ve rapidity
	  LHparticles.at(2).COL2=0;
	  LHparticles.at(2).p=p0;
	  LHparticles.at(2).PID=bptype;
	  LHparticles.at(2).MOT1=1;
	  LHparticles.at(2).MOT2=2;
	  LHparticles.at(2).ISTUP=1;
	  LHparticles.at(3).COL1=0;     // "Extra" parton
	  LHparticles.at(3).COL2=504;
	  LHparticles.at(3).p=p1;
	  LHparticles.at(3).PID=aptype;    
	  LHparticles.at(3).MOT1=1;
	  LHparticles.at(3).MOT2=2;
	  LHparticles.at(3).ISTUP=1;
	  LHparticles.at(4).COL1=504;    // Outgoing parton with +ve rapidity 
	  LHparticles.at(4).COL2=503;
	  LHparticles.at(4).p=p2;
	  LHparticles.at(4).PID=21;
	  LHparticles.at(4).MOT1=1;
	  LHparticles.at(4).MOT2=2;
	  LHparticles.at(4).ISTUP=1;       
	}	
	smatrixuxdthuxdg_(p22[0],&me2);
      }
      else if (pflag==1) {
	if (flags.LHoutput) {
	  LHparticles.at(0).COL1=501;    // Incoming particle with +ve rapidity
	  LHparticles.at(0).COL2=0;
	  LHparticles.at(1).COL1=0;    // Incoming particle with -ve rapidity
	  LHparticles.at(1).COL2=501;
	  LHparticles.at(2).COL1=504;    // Outgoing particle with -ve rapidity
	  LHparticles.at(2).COL2=503;
	  LHparticles.at(2).p=p0;
	  LHparticles.at(2).PID=21;
	  LHparticles.at(2).MOT1=1;
	  LHparticles.at(2).MOT2=2;
	  LHparticles.at(2).ISTUP=1;
	  LHparticles.at(3).COL1=503;     // "Extra" parton
	  LHparticles.at(3).COL2=0;
	  LHparticles.at(3).p=p1;
	  LHparticles.at(3).PID=bptype;    
	  LHparticles.at(3).MOT1=1;
	  LHparticles.at(3).MOT2=2;
	  LHparticles.at(3).ISTUP=1;
	  LHparticles.at(4).COL1=0;    // Outgoing parton with +ve rapidity 
	  LHparticles.at(4).COL2=504;
	  LHparticles.at(4).p=p2;
	  LHparticles.at(4).PID=aptype;
	  LHparticles.at(4).MOT1=1;
	  LHparticles.at(4).MOT2=2;
	  LHparticles.at(4).ISTUP=1;       
	}	
	smatrixuxdthuxdg_(p23[0],&me2);
      }
      else if (pflag==2) {
	if (flags.LHoutput) {
	  LHparticles.at(0).COL1=501;    // Incoming particle with +ve rapidity
	  LHparticles.at(0).COL2=0;
	  LHparticles.at(1).COL1=0;    // Incoming particle with -ve rapidity
	  LHparticles.at(1).COL2=501;
	  LHparticles.at(2).COL1=0;    // Outgoing particle with -ve rapidity
	  LHparticles.at(2).COL2=504;
	  LHparticles.at(2).p=p0;
	  LHparticles.at(2).PID=aptype;
	  LHparticles.at(2).MOT1=1;
	  LHparticles.at(2).MOT2=2;
	  LHparticles.at(2).ISTUP=1;
	  LHparticles.at(3).COL1=503;     // "Extra" parton
	  LHparticles.at(3).COL2=0;
	  LHparticles.at(3).p=p1;
	  LHparticles.at(3).PID=bptype;    
	  LHparticles.at(3).MOT1=1;
	  LHparticles.at(3).MOT2=2;
	  LHparticles.at(3).ISTUP=1;
	  LHparticles.at(4).COL1=504;    // Outgoing parton with +ve rapidity 
	  LHparticles.at(4).COL2=503;
	  LHparticles.at(4).p=p2;
	  LHparticles.at(4).PID=21;
	  LHparticles.at(4).MOT1=1;
	  LHparticles.at(4).MOT2=2;
	  LHparticles.at(4).ISTUP=1;       
	}		
	smatrixuxdthuxdg_(p24[0],&me2);
      }
      else if (pflag==3) {
	if (flags.LHoutput) {
	  LHparticles.at(0).COL1=501;    // Incoming particle with +ve rapidity
	  LHparticles.at(0).COL2=0;
	  LHparticles.at(1).COL1=0;    // Incoming particle with -ve rapidity
	  LHparticles.at(1).COL2=501;
	  LHparticles.at(2).COL1=503;    // Outgoing particle with -ve rapidity
	  LHparticles.at(2).COL2=504;
	  LHparticles.at(2).p=p0;
	  LHparticles.at(2).PID=21;
	  LHparticles.at(2).MOT1=1;
	  LHparticles.at(2).MOT2=2;
	  LHparticles.at(2).ISTUP=1;
	  LHparticles.at(3).COL1=0;     // "Extra" parton
	  LHparticles.at(3).COL2=503;
	  LHparticles.at(3).p=p1;
	  LHparticles.at(3).PID=aptype;    
	  LHparticles.at(3).MOT1=1;
	  LHparticles.at(3).MOT2=2;
	  LHparticles.at(3).ISTUP=1;
	  LHparticles.at(4).COL1=504;    // Outgoing parton with +ve rapidity 
	  LHparticles.at(4).COL2=0;
	  LHparticles.at(4).p=p2;
	  LHparticles.at(4).PID=bptype;
	  LHparticles.at(4).MOT1=1;
	  LHparticles.at(4).MOT2=2;
	  LHparticles.at(4).ISTUP=1;       
	}		
	smatrixuxdthuxdg_(p25[0],&me2);
      }
      else if (pflag==4) {
	if (flags.LHoutput) {
	  LHparticles.at(0).COL1=501;    // Incoming particle with +ve rapidity
	  LHparticles.at(0).COL2=0;
	  LHparticles.at(1).COL1=0;    // Incoming particle with -ve rapidity
	  LHparticles.at(1).COL2=501;
	  LHparticles.at(2).COL1=504;    // Outgoing particle with -ve rapidity
	  LHparticles.at(2).COL2=0;
	  LHparticles.at(2).p=p0;
	  LHparticles.at(2).PID=bptype;
	  LHparticles.at(2).MOT1=1;
	  LHparticles.at(2).MOT2=2;
	  LHparticles.at(2).ISTUP=1;
	  LHparticles.at(3).COL1=503;     // "Extra" parton
	  LHparticles.at(3).COL2=504;
	  LHparticles.at(3).p=p1;
	  LHparticles.at(3).PID=21;    
	  LHparticles.at(3).MOT1=1;
	  LHparticles.at(3).MOT2=2;
	  LHparticles.at(3).ISTUP=1;
	  LHparticles.at(4).COL1=0;    // Outgoing parton with +ve rapidity 
	  LHparticles.at(4).COL2=503;
	  LHparticles.at(4).p=p2;
	  LHparticles.at(4).PID=aptype;
	  LHparticles.at(4).MOT1=1;
	  LHparticles.at(4).MOT2=2;
	  LHparticles.at(4).ISTUP=1;       
	}	
	smatrixuxdthuxdg_(p26[0],&me2);
      }
    }
    else if (aptype>0&&bptype>0&&aptype!=21&&bptype!=21&&aptype==bptype) { // if uu initial state
      int pflag=int(2.*ran.flat()); // 2 rapidity orderings (non-FKL)
      weight*=2.;
      if(pflag==0) {
	if (flags.LHoutput) {
	  LHparticles.at(0).COL1=501;    // Incoming particle with +ve rapidity
	  LHparticles.at(0).COL2=0;
	  LHparticles.at(1).COL1=502;    // Incoming particle with -ve rapidity
	  LHparticles.at(1).COL2=0;
	  LHparticles.at(2).COL1=502;    // Outgoing particle with -ve rapidity
	  LHparticles.at(2).COL2=0;
	  LHparticles.at(2).p=p0;
	  LHparticles.at(2).PID=bptype;
	  LHparticles.at(2).MOT1=1;
	  LHparticles.at(2).MOT2=2;
	  LHparticles.at(2).ISTUP=1;
	  LHparticles.at(3).COL1=503;     // "Extra" parton
	  LHparticles.at(3).COL2=0;
	  LHparticles.at(3).p=p1;
	  LHparticles.at(3).PID=aptype;    
	  LHparticles.at(3).MOT1=1;
	  LHparticles.at(3).MOT2=2;
	  LHparticles.at(3).ISTUP=1;
	  LHparticles.at(4).COL1=501;    // Outgoing parton with +ve rapidity 
	  LHparticles.at(4).COL2=503;
	  LHparticles.at(4).p=p2;
	  LHparticles.at(4).PID=21;
	  LHparticles.at(4).MOT1=1;
	  LHparticles.at(4).MOT2=2;
	  LHparticles.at(4).ISTUP=1;       
	}	
	smatrixuuthuug_(p22[0],&me2);
	me2*=2.;
      }
      else if (pflag==1) {
	if (flags.LHoutput) {
	  LHparticles.at(0).COL1=501;    // Incoming particle with +ve rapidity
	  LHparticles.at(0).COL2=0;
	  LHparticles.at(1).COL1=502;    // Incoming particle with -ve rapidity
	  LHparticles.at(1).COL2=0;
	  LHparticles.at(2).COL1=501;    // Outgoing particle with -ve rapidity
	  LHparticles.at(2).COL2=503;
	  LHparticles.at(2).p=p0;
	  LHparticles.at(2).PID=21;
	  LHparticles.at(2).MOT1=1;
	  LHparticles.at(2).MOT2=2;
	  LHparticles.at(2).ISTUP=1;
	  LHparticles.at(3).COL1=502;     // "Extra" parton
	  LHparticles.at(3).COL2=0;
	  LHparticles.at(3).p=p1;
	  LHparticles.at(3).PID=bptype;    
	  LHparticles.at(3).MOT1=1;
	  LHparticles.at(3).MOT2=2;
	  LHparticles.at(3).ISTUP=1;
	  LHparticles.at(4).COL1=503;    // Outgoing parton with +ve rapidity 
	  LHparticles.at(4).COL2=0;
	  LHparticles.at(4).p=p2;
	  LHparticles.at(4).PID=aptype;
	  LHparticles.at(4).MOT1=1;
	  LHparticles.at(4).MOT2=2;
	  LHparticles.at(4).ISTUP=1;       
	}	
	smatrixuuthuug_(p23[0],&me2);
	me2*=2.;
      }
    }
    else if (aptype<0&&bptype<0&&aptype==bptype) { // If uxux initial state
      int pflag=int(2.*ran.flat()); // 2 rapidity orderings (non-FKL)
      weight*=2.;
      if(pflag==0) {
	if (flags.LHoutput) {
	  LHparticles.at(0).COL1=0;    // Incoming particle with +ve rapidity
	  LHparticles.at(0).COL2=501;
	  LHparticles.at(1).COL1=0;    // Incoming particle with -ve rapidity
	  LHparticles.at(1).COL2=502;
	  LHparticles.at(2).COL1=0;    // Outgoing particle with -ve rapidity
	  LHparticles.at(2).COL2=502;
	  LHparticles.at(2).p=p0;
	  LHparticles.at(2).PID=bptype;
	  LHparticles.at(2).MOT1=1;
	  LHparticles.at(2).MOT2=2;
	  LHparticles.at(2).ISTUP=1;
	  LHparticles.at(3).COL1=0;     // "Extra" parton
	  LHparticles.at(3).COL2=503;
	  LHparticles.at(3).p=p1;
	  LHparticles.at(3).PID=aptype;    
	  LHparticles.at(3).MOT1=1;
	  LHparticles.at(3).MOT2=2;
	  LHparticles.at(3).ISTUP=1;
	  LHparticles.at(4).COL1=503;    // Outgoing parton with +ve rapidity 
	  LHparticles.at(4).COL2=501;
	  LHparticles.at(4).p=p2;
	  LHparticles.at(4).PID=21;
	  LHparticles.at(4).MOT1=1;
	  LHparticles.at(4).MOT2=2;
	  LHparticles.at(4).ISTUP=1;       
	}		
	smatrixuxuxthuxuxg_(p22[0],&me2);
	me2*=2.;
      }
      else if (pflag==1) {
	if (flags.LHoutput) {
	  LHparticles.at(0).COL1=0;    // Incoming particle with +ve rapidity
	  LHparticles.at(0).COL2=501;
	  LHparticles.at(1).COL1=0;    // Incoming particle with -ve rapidity
	  LHparticles.at(1).COL2=502;
	  LHparticles.at(2).COL1=503;    // Outgoing particle with -ve rapidity
	  LHparticles.at(2).COL2=501;
	  LHparticles.at(2).p=p0;
	  LHparticles.at(2).PID=21;
	  LHparticles.at(2).MOT1=1;
	  LHparticles.at(2).MOT2=2;
	  LHparticles.at(2).ISTUP=1;
	  LHparticles.at(3).COL1=0;     // "Extra" parton
	  LHparticles.at(3).COL2=502;
	  LHparticles.at(3).p=p1;
	  LHparticles.at(3).PID=bptype;    
	  LHparticles.at(3).MOT1=1;
	  LHparticles.at(3).MOT2=2;
	  LHparticles.at(3).ISTUP=1;
	  LHparticles.at(4).COL1=0;    // Outgoing parton with +ve rapidity 
	  LHparticles.at(4).COL2=503;
	  LHparticles.at(4).p=p2;
	  LHparticles.at(4).PID=aptype;
	  LHparticles.at(4).MOT1=1;
	  LHparticles.at(4).MOT2=2;
	  LHparticles.at(4).ISTUP=1;       
	}	
	smatrixuxuxthuxuxg_(p23[0],&me2);
	me2*=2.;
      }
    }
    else if (aptype>0&&aptype!=21&&bptype<0&&aptype==-bptype) { // If uux initial state
      int process=int(3.*ran.flat()); // labels partonic subchannel
      weight*=3.;

      if(process==0) { // if uuxthddxg
	int pflag=int(6.*ran.flat()); // 6 rapifity configs
	int flav=((int(4.*ran.flat())+abs(bptype))%5)+1; // Choose a random flavour that is not the input flavour
	weight*=6.*4.;
	if (pflag==0) {
	  if (flags.LHoutput) {
	    LHparticles.at(0).COL1=0;    // Incoming particle with +ve rapidity
	    LHparticles.at(0).COL2=501;
	    LHparticles.at(1).COL1=502;    // Incoming particle with -ve rapidity
	    LHparticles.at(1).COL2=0;
	    LHparticles.at(2).COL1=502;    // Outgoing particle with -ve rapidity
	    LHparticles.at(2).COL2=0;
	    LHparticles.at(2).p=p0;
	    LHparticles.at(2).PID=flav;
	    LHparticles.at(2).MOT1=1;
	    LHparticles.at(2).MOT2=2;
	    LHparticles.at(2).ISTUP=1;
	    LHparticles.at(3).COL1=503;     // "Extra" parton
	    LHparticles.at(3).COL2=501;
	    LHparticles.at(3).p=p1;
	    LHparticles.at(3).PID=21;    
	    LHparticles.at(3).MOT1=1;
	    LHparticles.at(3).MOT2=2;
	    LHparticles.at(3).ISTUP=1;
	    LHparticles.at(4).COL1=0;    // Outgoing parton with +ve rapidity 
	    LHparticles.at(4).COL2=503;
	    LHparticles.at(4).p=p2;
	    LHparticles.at(4).PID=-flav;
	    LHparticles.at(4).MOT1=1;
	    LHparticles.at(4).MOT2=2;
	    LHparticles.at(4).ISTUP=1;       
	  }	
	  smatrixuuxthddxg_(p21[0],&me2);
	}
	if (pflag==1) {
	  if (flags.LHoutput) {
	    LHparticles.at(0).COL1=0;    // Incoming particle with +ve rapidity
	    LHparticles.at(0).COL2=501;
	    LHparticles.at(1).COL1=502;    // Incoming particle with -ve rapidity
	    LHparticles.at(1).COL2=0;
	    LHparticles.at(2).COL1=0;    // Outgoing particle with -ve rapidity
	    LHparticles.at(2).COL2=503;
	    LHparticles.at(2).p=p0;
	    LHparticles.at(2).PID=-flav;
	    LHparticles.at(2).MOT1=1;
	    LHparticles.at(2).MOT2=2;
	    LHparticles.at(2).ISTUP=1;
	    LHparticles.at(3).COL1=502;     // "Extra" parton
	    LHparticles.at(3).COL2=0;
	    LHparticles.at(3).p=p1;
	    LHparticles.at(3).PID=flav;    
	    LHparticles.at(3).MOT1=1;
	    LHparticles.at(3).MOT2=2;
	    LHparticles.at(3).ISTUP=1;
	    LHparticles.at(4).COL1=503;    // Outgoing parton with +ve rapidity 
	    LHparticles.at(4).COL2=501;
	    LHparticles.at(4).p=p2;
	    LHparticles.at(4).PID=21;
	    LHparticles.at(4).MOT1=1;
	    LHparticles.at(4).MOT2=2;
	    LHparticles.at(4).ISTUP=1;       
	  }	
	  smatrixuuxthddxg_(p22[0],&me2);
	}
	else if (pflag==2) {
	  if (flags.LHoutput) {
	    LHparticles.at(0).COL1=0;    // Incoming particle with +ve rapidity
	    LHparticles.at(0).COL2=501;
	    LHparticles.at(1).COL1=502;    // Incoming particle with -ve rapidity
	    LHparticles.at(1).COL2=0;
	    LHparticles.at(2).COL1=503;    // Outgoing particle with -ve rapidity
	    LHparticles.at(2).COL2=501;
	    LHparticles.at(2).p=p0;
	    LHparticles.at(2).PID=21;
	    LHparticles.at(2).MOT1=1;
	    LHparticles.at(2).MOT2=2;
	    LHparticles.at(2).ISTUP=1;
	    LHparticles.at(3).COL1=0;     // "Extra" parton
	    LHparticles.at(3).COL2=503;
	    LHparticles.at(3).p=p1;
	    LHparticles.at(3).PID=-flav;    
	    LHparticles.at(3).MOT1=1;
	    LHparticles.at(3).MOT2=2;
	    LHparticles.at(3).ISTUP=1;
	    LHparticles.at(4).COL1=502;    // Outgoing parton with +ve rapidity 
	    LHparticles.at(4).COL2=0;
	    LHparticles.at(4).p=p2;
	    LHparticles.at(4).PID=flav;
	    LHparticles.at(4).MOT1=1;
	    LHparticles.at(4).MOT2=2;
	    LHparticles.at(4).ISTUP=1;       
	  }	
	  smatrixuuxthddxg_(p23[0],&me2);
	}
	else if (pflag==3) {
	  if (flags.LHoutput) {
	    LHparticles.at(0).COL1=0;    // Incoming particle with +ve rapidity
	    LHparticles.at(0).COL2=501;
	    LHparticles.at(1).COL1=502;    // Incoming particle with -ve rapidity
	    LHparticles.at(1).COL2=0;
	    LHparticles.at(2).COL1=502;    // Outgoing particle with -ve rapidity
	    LHparticles.at(2).COL2=0;
	    LHparticles.at(2).p=p0;
	    LHparticles.at(2).PID=flav;
	    LHparticles.at(2).MOT1=1;
	    LHparticles.at(2).MOT2=2;
	    LHparticles.at(2).ISTUP=1;
	    LHparticles.at(3).COL1=0;     // "Extra" parton
	    LHparticles.at(3).COL2=503;
	    LHparticles.at(3).p=p1;
	    LHparticles.at(3).PID=-flav;    
	    LHparticles.at(3).MOT1=1;
	    LHparticles.at(3).MOT2=2;
	    LHparticles.at(3).ISTUP=1;
	    LHparticles.at(4).COL1=503;    // Outgoing parton with +ve rapidity 
	    LHparticles.at(4).COL2=501;
	    LHparticles.at(4).p=p2;
	    LHparticles.at(4).PID=21;
	    LHparticles.at(4).MOT1=1;
	    LHparticles.at(4).MOT2=2;
	    LHparticles.at(4).ISTUP=1;       
	  }		
	  smatrixuuxthddxg_(p24[0],&me2);
	}
	else if (pflag==4) {
	  if (flags.LHoutput) {
	    LHparticles.at(0).COL1=0;    // Incoming particle with +ve rapidity
	    LHparticles.at(0).COL2=501;
	    LHparticles.at(1).COL1=502;    // Incoming particle with -ve rapidity
	    LHparticles.at(1).COL2=0;
	    LHparticles.at(2).COL1=503;    // Outgoing particle with -ve rapidity
	    LHparticles.at(2).COL2=501;
	    LHparticles.at(2).p=p0;
	    LHparticles.at(2).PID=21;
	    LHparticles.at(2).MOT1=1;
	    LHparticles.at(2).MOT2=2;
	    LHparticles.at(2).ISTUP=1;
	    LHparticles.at(3).COL1=502;     // "Extra" parton
	    LHparticles.at(3).COL2=0;
	    LHparticles.at(3).p=p1;
	    LHparticles.at(3).PID=flav;    
	    LHparticles.at(3).MOT1=1;
	    LHparticles.at(3).MOT2=2;
	    LHparticles.at(3).ISTUP=1;
	    LHparticles.at(4).COL1=0;    // Outgoing parton with +ve rapidity 
	    LHparticles.at(4).COL2=503;
	    LHparticles.at(4).p=p2;
	    LHparticles.at(4).PID=-flav;
	    LHparticles.at(4).MOT1=1;
	    LHparticles.at(4).MOT2=2;
	    LHparticles.at(4).ISTUP=1;       
	  }	
	  smatrixuuxthddxg_(p25[0],&me2);
	}
	else if (pflag==5) {
	  if (flags.LHoutput) {
	    LHparticles.at(0).COL1=0;    // Incoming particle with +ve rapidity
	    LHparticles.at(0).COL2=501;
	    LHparticles.at(1).COL1=502;    // Incoming particle with -ve rapidity
	    LHparticles.at(1).COL2=0;
	    LHparticles.at(2).COL1=0;    // Outgoing particle with -ve rapidity
	    LHparticles.at(2).COL2=503;
	    LHparticles.at(2).p=p0;
	    LHparticles.at(2).PID=-flav;
	    LHparticles.at(2).MOT1=1;
	    LHparticles.at(2).MOT2=2;
	    LHparticles.at(2).ISTUP=1;
	    LHparticles.at(3).COL1=503;     // "Extra" parton
	    LHparticles.at(3).COL2=501;
	    LHparticles.at(3).p=p1;
	    LHparticles.at(3).PID=21;    
	    LHparticles.at(3).MOT1=1;
	    LHparticles.at(3).MOT2=2;
	    LHparticles.at(3).ISTUP=1;
	    LHparticles.at(4).COL1=502;    // Outgoing parton with +ve rapidity 
	    LHparticles.at(4).COL2=0;
	    LHparticles.at(4).p=p2;
	    LHparticles.at(4).PID=flav;
	    LHparticles.at(4).MOT1=1;
	    LHparticles.at(4).MOT2=2;
	    LHparticles.at(4).ISTUP=1;       
	  }	
	  smatrixuuxthddxg_(p26[0],&me2);
	}
      }
      else if (process==1) { // if uuxthuuxg
	int pflag=int(5.*ran.flat()); // 5 rapidity configs (non-FKL)
	weight*=5.;
	if (pflag==0) {
	  if (flags.LHoutput) {
	    LHparticles.at(0).COL1=0;    // Incoming particle with +ve rapidity
	    LHparticles.at(0).COL2=502;
	    LHparticles.at(1).COL1=501;    // Incoming particle with -ve rapidity
	    LHparticles.at(1).COL2=0;
	    LHparticles.at(2).COL1=0;    // Outgoing particle with -ve rapidity
	    LHparticles.at(2).COL2=503;
	    LHparticles.at(2).p=p0;
	    LHparticles.at(2).PID=-aptype;
	    LHparticles.at(2).MOT1=1;
	    LHparticles.at(2).MOT2=2;
	    LHparticles.at(2).ISTUP=1;
	    LHparticles.at(3).COL1=501;     // "Extra" parton
	    LHparticles.at(3).COL2=0;
	    LHparticles.at(3).p=p1;
	    LHparticles.at(3).PID=aptype;    
	    LHparticles.at(3).MOT1=1;
	    LHparticles.at(3).MOT2=2;
	    LHparticles.at(3).ISTUP=1;
	    LHparticles.at(4).COL1=503;    // Outgoing parton with +ve rapidity 
	    LHparticles.at(4).COL2=502;
	    LHparticles.at(4).p=p2;
	    LHparticles.at(4).PID=21;
	    LHparticles.at(4).MOT1=1;
	    LHparticles.at(4).MOT2=2;
	    LHparticles.at(4).ISTUP=1;       
	  }	
	  smatrixuuxthuuxg_(p22[0],&me2);
	}
	else if (pflag==1) {
	  if (flags.LHoutput) {
	    LHparticles.at(0).COL1=0;    // Incoming particle with +ve rapidity
	    LHparticles.at(0).COL2=502;
	    LHparticles.at(1).COL1=501;    // Incoming particle with -ve rapidity
	    LHparticles.at(1).COL2=0;
	    LHparticles.at(2).COL1=503;    // Outgoing particle with -ve rapidity
	    LHparticles.at(2).COL2=502;
	    LHparticles.at(2).p=p0;
	    LHparticles.at(2).PID=21;
	    LHparticles.at(2).MOT1=1;
	    LHparticles.at(2).MOT2=2;
	    LHparticles.at(2).ISTUP=1;
	    LHparticles.at(3).COL1=0;     // "Extra" parton
	    LHparticles.at(3).COL2=503;
	    LHparticles.at(3).p=p1;
	    LHparticles.at(3).PID=-aptype;    
	    LHparticles.at(3).MOT1=1;
	    LHparticles.at(3).MOT2=2;
	    LHparticles.at(3).ISTUP=1;
	    LHparticles.at(4).COL1=501;    // Outgoing parton with +ve rapidity 
	    LHparticles.at(4).COL2=0;
	    LHparticles.at(4).p=p2;
	    LHparticles.at(4).PID=aptype;
	    LHparticles.at(4).MOT1=1;
	    LHparticles.at(4).MOT2=2;
	    LHparticles.at(4).ISTUP=1;       
	  }		
	  smatrixuuxthuuxg_(p23[0],&me2);
	}
	else if (pflag==2) {
	  if (flags.LHoutput) {
	    LHparticles.at(0).COL1=0;    // Incoming particle with +ve rapidity
	    LHparticles.at(0).COL2=502;
	    LHparticles.at(1).COL1=501;    // Incoming particle with -ve rapidity
	    LHparticles.at(1).COL2=0;
	    LHparticles.at(2).COL1=501;    // Outgoing particle with -ve rapidity
	    LHparticles.at(2).COL2=0;
	    LHparticles.at(2).p=p0;
	    LHparticles.at(2).PID=aptype;
	    LHparticles.at(2).MOT1=1;
	    LHparticles.at(2).MOT2=2;
	    LHparticles.at(2).ISTUP=1;
	    LHparticles.at(3).COL1=0;     // "Extra" parton
	    LHparticles.at(3).COL2=503;
	    LHparticles.at(3).p=p1;
	    LHparticles.at(3).PID=-aptype;    
	    LHparticles.at(3).MOT1=1;
	    LHparticles.at(3).MOT2=2;
	    LHparticles.at(3).ISTUP=1;
	    LHparticles.at(4).COL1=503;    // Outgoing parton with +ve rapidity 
	    LHparticles.at(4).COL2=502;
	    LHparticles.at(4).p=p2;
	    LHparticles.at(4).PID=21;
	    LHparticles.at(4).MOT1=1;
	    LHparticles.at(4).MOT2=2;
	    LHparticles.at(4).ISTUP=1;       
	  }	
	  smatrixuuxthuuxg_(p24[0],&me2);
	}
	else if (pflag==3) {
	  if (flags.LHoutput) {
	    LHparticles.at(0).COL1=0;    // Incoming particle with +ve rapidity
	    LHparticles.at(0).COL2=502;
	    LHparticles.at(1).COL1=501;    // Incoming particle with -ve rapidity
	    LHparticles.at(1).COL2=0;
	    LHparticles.at(2).COL1=503;    // Outgoing particle with -ve rapidity
	    LHparticles.at(2).COL2=502;
	    LHparticles.at(2).p=p0;
	    LHparticles.at(2).PID=21;
	    LHparticles.at(2).MOT1=1;
	    LHparticles.at(2).MOT2=2;
	    LHparticles.at(2).ISTUP=1;
	    LHparticles.at(3).COL1=501;     // "Extra" parton
	    LHparticles.at(3).COL2=0;
	    LHparticles.at(3).p=p1;
	    LHparticles.at(3).PID=aptype;    
	    LHparticles.at(3).MOT1=1;
	    LHparticles.at(3).MOT2=2;
	    LHparticles.at(3).ISTUP=1;
	    LHparticles.at(4).COL1=0;    // Outgoing parton with +ve rapidity 
	    LHparticles.at(4).COL2=503;
	    LHparticles.at(4).p=p2;
	    LHparticles.at(4).PID=-aptype;
	    LHparticles.at(4).MOT1=1;
	    LHparticles.at(4).MOT2=2;
	    LHparticles.at(4).ISTUP=1;       
	  }		  
	  smatrixuuxthuuxg_(p25[0],&me2);
	}
	else if (pflag==4) {
	  if (flags.LHoutput) {
	    LHparticles.at(0).COL1=0;    // Incoming particle with +ve rapidity
	    LHparticles.at(0).COL2=502;
	    LHparticles.at(1).COL1=501;    // Incoming particle with -ve rapidity
	    LHparticles.at(1).COL2=0;
	    LHparticles.at(2).COL1=0;    // Outgoing particle with -ve rapidity
	    LHparticles.at(2).COL2=503;
	    LHparticles.at(2).p=p0;
	    LHparticles.at(2).PID=-aptype;
	    LHparticles.at(2).MOT1=1;
	    LHparticles.at(2).MOT2=2;
	    LHparticles.at(2).ISTUP=1;
	    LHparticles.at(3).COL1=503;     // "Extra" parton
	    LHparticles.at(3).COL2=502;
	    LHparticles.at(3).p=p1;
	    LHparticles.at(3).PID=21;    
	    LHparticles.at(3).MOT1=1;
	    LHparticles.at(3).MOT2=2;
	    LHparticles.at(3).ISTUP=1;
	    LHparticles.at(4).COL1=501;    // Outgoing parton with +ve rapidity 
	    LHparticles.at(4).COL2=0;
	    LHparticles.at(4).p=p2;
	    LHparticles.at(4).PID=aptype;
	    LHparticles.at(4).MOT1=1;
	    LHparticles.at(4).MOT2=2;
	    LHparticles.at(4).ISTUP=1;       
	  }		  
	  smatrixuuxthuuxg_(p26[0],&me2);
	}
      }
      else if (process==2) { // If uuxthggg
	if (flags.LHoutput) {
	  LHparticles.at(0).COL1=0;    // Incoming particle with +ve rapidity
	  LHparticles.at(0).COL2=502;
	  LHparticles.at(1).COL1=501;    // Incoming particle with -ve rapidity
	  LHparticles.at(1).COL2=0;
	  LHparticles.at(2).COL1=501;    // Outgoing particle with -ve rapidity
	  LHparticles.at(2).COL2=503;
	  LHparticles.at(2).p=p0;
	  LHparticles.at(2).PID=21;
	  LHparticles.at(2).MOT1=1;
	  LHparticles.at(2).MOT2=2;
	  LHparticles.at(2).ISTUP=1;
	  LHparticles.at(3).COL1=503;     // "Extra" parton
	  LHparticles.at(3).COL2=504;
	  LHparticles.at(3).p=p1;
	  LHparticles.at(3).PID=21;    
	  LHparticles.at(3).MOT1=1;
	  LHparticles.at(3).MOT2=2;
	  LHparticles.at(3).ISTUP=1;
	  LHparticles.at(4).COL1=504;    // Outgoing parton with +ve rapidity 
	  LHparticles.at(4).COL2=502;
	  LHparticles.at(4).p=p2;
	  LHparticles.at(4).PID=21;
	  LHparticles.at(4).MOT1=1;
	  LHparticles.at(4).MOT2=2;
	  LHparticles.at(4).ISTUP=1;       
	}		
	smatrixuuxthggg_(p21[0],&me2);
	me2*=6.;
      }
    }      
    else if (aptype<0&&bptype>0&&bptype!=21&&bptype==-aptype) { // if uxu initial state
      int process=int(3.*ran.flat()); // labels partonic subchannel
      weight*=3.;

      if(process==0) {
	int pflag=int(6.*ran.flat()); // 6 rapidity configs
	int flav=((int(4.*ran.flat())+abs(bptype))%5)+1; // Choose a random flavour that is not the input flavour
	weight*=6.*4;
	if (pflag==0) { // If uuxthddxg
	  if (flags.LHoutput) {
	    LHparticles.at(0).COL1=501;    // Incoming particle with +ve rapidity
	    LHparticles.at(0).COL2=0;
	    LHparticles.at(1).COL1=0;    // Incoming particle with -ve rapidity
	    LHparticles.at(1).COL2=502;
	    LHparticles.at(2).COL1=0;    // Outgoing particle with -ve rapidity
	    LHparticles.at(2).COL2=503;
	    LHparticles.at(2).p=p0;
	    LHparticles.at(2).PID=-flav;
	    LHparticles.at(2).MOT1=1;
	    LHparticles.at(2).MOT2=2;
	    LHparticles.at(2).ISTUP=1;
	    LHparticles.at(3).COL1=503;     // "Extra" parton
	    LHparticles.at(3).COL2=502;
	    LHparticles.at(3).p=p1;
	    LHparticles.at(3).PID=21;    
	    LHparticles.at(3).MOT1=1;
	    LHparticles.at(3).MOT2=2;
	    LHparticles.at(3).ISTUP=1;
	    LHparticles.at(4).COL1=501;    // Outgoing parton with +ve rapidity 
	    LHparticles.at(4).COL2=0;
	    LHparticles.at(4).p=p2;
	    LHparticles.at(4).PID=flav;
	    LHparticles.at(4).MOT1=1;
	    LHparticles.at(4).MOT2=2;
	    LHparticles.at(4).ISTUP=1;       
	  }	
	  smatrixuuxthddxg_(p11[0],&me2);
	}
	else if (pflag==1) {
	  if (flags.LHoutput) {
	    LHparticles.at(0).COL1=501;    // Incoming particle with +ve rapidity
	    LHparticles.at(0).COL2=0;
	    LHparticles.at(1).COL1=0;    // Incoming particle with -ve rapidity
	    LHparticles.at(1).COL2=502;
	    LHparticles.at(2).COL1=503;    // Outgoing particle with -ve rapidity
	    LHparticles.at(2).COL2=502;
	    LHparticles.at(2).p=p0;
	    LHparticles.at(2).PID=21;
	    LHparticles.at(2).MOT1=1;
	    LHparticles.at(2).MOT2=2;
	    LHparticles.at(2).ISTUP=1;
	    LHparticles.at(3).COL1=501;     // "Extra" parton
	    LHparticles.at(3).COL2=0;
	    LHparticles.at(3).p=p1;
	    LHparticles.at(3).PID=flav;    
	    LHparticles.at(3).MOT1=1;
	    LHparticles.at(3).MOT2=2;
	    LHparticles.at(3).ISTUP=1;
	    LHparticles.at(4).COL1=0;    // Outgoing parton with +ve rapidity 
	    LHparticles.at(4).COL2=503;
	    LHparticles.at(4).p=p2;
	    LHparticles.at(4).PID=-flav;
	    LHparticles.at(4).MOT1=1;
	    LHparticles.at(4).MOT2=2;
	    LHparticles.at(4).ISTUP=1;       
	  }	
	  smatrixuuxthddxg_(p12[0],&me2);
	}
	else if (pflag==2) {
	  if (flags.LHoutput) {
	    LHparticles.at(0).COL1=501;    // Incoming particle with +ve rapidity
	    LHparticles.at(0).COL2=0;
	    LHparticles.at(1).COL1=0;    // Incoming particle with -ve rapidity
	    LHparticles.at(1).COL2=502;
	    LHparticles.at(2).COL1=501;    // Outgoing particle with -ve rapidity
	    LHparticles.at(2).COL2=0;
	    LHparticles.at(2).p=p0;
	    LHparticles.at(2).PID=flav;
	    LHparticles.at(2).MOT1=1;
	    LHparticles.at(2).MOT2=2;
	    LHparticles.at(2).ISTUP=1;
	    LHparticles.at(3).COL1=0;     // "Extra" parton
	    LHparticles.at(3).COL2=503;
	    LHparticles.at(3).p=p1;
	    LHparticles.at(3).PID=-flav;    
	    LHparticles.at(3).MOT1=1;
	    LHparticles.at(3).MOT2=2;
	    LHparticles.at(3).ISTUP=1;
	    LHparticles.at(4).COL1=503;    // Outgoing parton with +ve rapidity 
	    LHparticles.at(4).COL2=502;
	    LHparticles.at(4).p=p2;
	    LHparticles.at(4).PID=21;
	    LHparticles.at(4).MOT1=1;
	    LHparticles.at(4).MOT2=2;
	    LHparticles.at(4).ISTUP=1;       
	  }		
	  smatrixuuxthddxg_(p13[0],&me2);
	}
	else if (pflag==3) {
	  if (flags.LHoutput) {
	    LHparticles.at(0).COL1=501;    // Incoming particle with +ve rapidity
	    LHparticles.at(0).COL2=0;
	    LHparticles.at(1).COL1=0;    // Incoming particle with -ve rapidity
	    LHparticles.at(1).COL2=502;
	    LHparticles.at(2).COL1=503;    // Outgoing particle with -ve rapidity
	    LHparticles.at(2).COL2=502;
	    LHparticles.at(2).p=p0;
	    LHparticles.at(2).PID=21;
	    LHparticles.at(2).MOT1=1;
	    LHparticles.at(2).MOT2=2;
	    LHparticles.at(2).ISTUP=1;
	    LHparticles.at(3).COL1=0;     // "Extra" parton
	    LHparticles.at(3).COL2=503;
	    LHparticles.at(3).p=p1;
	    LHparticles.at(3).PID=-flav;    
	    LHparticles.at(3).MOT1=1;
	    LHparticles.at(3).MOT2=2;
	    LHparticles.at(3).ISTUP=1;
	    LHparticles.at(4).COL1=501;    // Outgoing parton with +ve rapidity 
	    LHparticles.at(4).COL2=0;
	    LHparticles.at(4).p=p2;
	    LHparticles.at(4).PID=flav;
	    LHparticles.at(4).MOT1=1;
	    LHparticles.at(4).MOT2=2;
	    LHparticles.at(4).ISTUP=1;       
	  }	     
	  smatrixuuxthddxg_(p14[0],&me2);
	}
	else if (pflag==4) {
	  if (flags.LHoutput) {
	    LHparticles.at(0).COL1=501;    // Incoming particle with +ve rapidity
	    LHparticles.at(0).COL2=0;
	    LHparticles.at(1).COL1=0;    // Incoming particle with -ve rapidity
	    LHparticles.at(1).COL2=502;
	    LHparticles.at(2).COL1=0;    // Outgoing particle with -ve rapidity
	    LHparticles.at(2).COL2=503;
	    LHparticles.at(2).p=p0;
	    LHparticles.at(2).PID=-flav;
	    LHparticles.at(2).MOT1=1;
	    LHparticles.at(2).MOT2=2;
	    LHparticles.at(2).ISTUP=1;
	    LHparticles.at(3).COL1=501;     // "Extra" parton
	    LHparticles.at(3).COL2=0;
	    LHparticles.at(3).p=p1;
	    LHparticles.at(3).PID=flav;    
	    LHparticles.at(3).MOT1=1;
	    LHparticles.at(3).MOT2=2;
	    LHparticles.at(3).ISTUP=1;
	    LHparticles.at(4).COL1=503;    // Outgoing parton with +ve rapidity 
	    LHparticles.at(4).COL2=502;
	    LHparticles.at(4).p=p2;
	    LHparticles.at(4).PID=21;
	    LHparticles.at(4).MOT1=1;
	    LHparticles.at(4).MOT2=2;
	    LHparticles.at(4).ISTUP=1;       
	  }	
	  smatrixuuxthddxg_(p15[0],&me2);
	}
	else if (pflag==5) {
	  if (flags.LHoutput) {
	    LHparticles.at(0).COL1=501;    // Incoming particle with +ve rapidity
	    LHparticles.at(0).COL2=0;
	    LHparticles.at(1).COL1=0;    // Incoming particle with -ve rapidity
	    LHparticles.at(1).COL2=502;
	    LHparticles.at(2).COL1=501;    // Outgoing particle with -ve rapidity
	    LHparticles.at(2).COL2=0;
	    LHparticles.at(2).p=p0;
	    LHparticles.at(2).PID=flav;
	    LHparticles.at(2).MOT1=1;
	    LHparticles.at(2).MOT2=2;
	    LHparticles.at(2).ISTUP=1;
	    LHparticles.at(3).COL1=503;     // "Extra" parton
	    LHparticles.at(3).COL2=502;
	    LHparticles.at(3).p=p1;
	    LHparticles.at(3).PID=21;    
	    LHparticles.at(3).MOT1=1;
	    LHparticles.at(3).MOT2=2;
	    LHparticles.at(3).ISTUP=1;
	    LHparticles.at(4).COL1=0;    // Outgoing parton with +ve rapidity 
	    LHparticles.at(4).COL2=503;
	    LHparticles.at(4).p=p2;
	    LHparticles.at(4).PID=-flav;
	    LHparticles.at(4).MOT1=1;
	    LHparticles.at(4).MOT2=2;
	    LHparticles.at(4).ISTUP=1;       
	  }	
	  smatrixuuxthddxg_(p16[0],&me2);
      }
    }
    else if (process==1) { // If uuxthuuxg
      int pflag=int(5.*ran.flat()); // 5 rapidity configs (non-FKL)
      weight*=5.;
      if (pflag==0) {
	if (flags.LHoutput) {
	  LHparticles.at(0).COL1=501;    // Incoming particle with +ve rapidity
	  LHparticles.at(0).COL2=0;
	  LHparticles.at(1).COL1=0;    // Incoming particle with -ve rapidity
	  LHparticles.at(1).COL2=502;
	  LHparticles.at(2).COL1=503;    // Outgoing particle with -ve rapidity
	  LHparticles.at(2).COL2=502;
	  LHparticles.at(2).p=p0;
	  LHparticles.at(2).PID=21;
	  LHparticles.at(2).MOT1=1;
	  LHparticles.at(2).MOT2=2;
	  LHparticles.at(2).ISTUP=1;
	  LHparticles.at(3).COL1=501;     // "Extra" parton
	  LHparticles.at(3).COL2=0;
	  LHparticles.at(3).p=p1;
	  LHparticles.at(3).PID=bptype;    
	  LHparticles.at(3).MOT1=1;
	  LHparticles.at(3).MOT2=2;
	  LHparticles.at(3).ISTUP=1;
	  LHparticles.at(4).COL1=0;    // Outgoing parton with +ve rapidity 
	  LHparticles.at(4).COL2=503;
	  LHparticles.at(4).p=p2;
	  LHparticles.at(4).PID=-bptype;
	  LHparticles.at(4).MOT1=1;
	  LHparticles.at(4).MOT2=2;
	  LHparticles.at(4).ISTUP=1;       
	}		
	smatrixuuxthuuxg_(p12[0],&me2);
      }
      else if (pflag==1) {
	if (flags.LHoutput) {
	  LHparticles.at(0).COL1=501;    // Incoming particle with +ve rapidity
	  LHparticles.at(0).COL2=0;
	  LHparticles.at(1).COL1=0;    // Incoming particle with -ve rapidity
	  LHparticles.at(1).COL2=502;
	  LHparticles.at(2).COL1=501;    // Outgoing particle with -ve rapidity
	  LHparticles.at(2).COL2=0;
	  LHparticles.at(2).p=p0;
	  LHparticles.at(2).PID=bptype;
	  LHparticles.at(2).MOT1=1;
	  LHparticles.at(2).MOT2=2;
	  LHparticles.at(2).ISTUP=1;
	  LHparticles.at(3).COL1=0;     // "Extra" parton
	  LHparticles.at(3).COL2=503;
	  LHparticles.at(3).p=p1;
	  LHparticles.at(3).PID=-bptype;    
	  LHparticles.at(3).MOT1=1;
	  LHparticles.at(3).MOT2=2;
	  LHparticles.at(3).ISTUP=1;
	  LHparticles.at(4).COL1=503;    // Outgoing parton with +ve rapidity 
	  LHparticles.at(4).COL2=502;
	  LHparticles.at(4).p=p2;
	  LHparticles.at(4).PID=21;
	  LHparticles.at(4).MOT1=1;
	  LHparticles.at(4).MOT2=2;
	  LHparticles.at(4).ISTUP=1;       
	}		
	smatrixuuxthuuxg_(p13[0],&me2);
      }
      else if (pflag==2) {
	if (flags.LHoutput) {
	  LHparticles.at(0).COL1=501;    // Incoming particle with +ve rapidity
	  LHparticles.at(0).COL2=0;
	  LHparticles.at(1).COL1=0;    // Incoming particle with -ve rapidity
	  LHparticles.at(1).COL2=502;
	  LHparticles.at(2).COL1=503;    // Outgoing particle with -ve rapidity
	  LHparticles.at(2).COL2=502;
	  LHparticles.at(2).p=p0;
	  LHparticles.at(2).PID=21;
	  LHparticles.at(2).MOT1=1;
	  LHparticles.at(2).MOT2=2;
	  LHparticles.at(2).ISTUP=1;
	  LHparticles.at(3).COL1=0;     // "Extra" parton
	  LHparticles.at(3).COL2=503;
	  LHparticles.at(3).p=p1;
	  LHparticles.at(3).PID=-bptype;    
	  LHparticles.at(3).MOT1=1;
	  LHparticles.at(3).MOT2=2;
	  LHparticles.at(3).ISTUP=1;
	  LHparticles.at(4).COL1=501;    // Outgoing parton with +ve rapidity 
	  LHparticles.at(4).COL2=0;
	  LHparticles.at(4).p=p2;
	  LHparticles.at(4).PID=bptype;
	  LHparticles.at(4).MOT1=1;
	  LHparticles.at(4).MOT2=2;
	  LHparticles.at(4).ISTUP=1;       
	}	
	smatrixuuxthuuxg_(p14[0],&me2);
      }
      else if (pflag==3) {
	if (flags.LHoutput) {
	  LHparticles.at(0).COL1=501;    // Incoming particle with +ve rapidity
	  LHparticles.at(0).COL2=0;
	  LHparticles.at(1).COL1=0;    // Incoming particle with -ve rapidity
	  LHparticles.at(1).COL2=502;
	  LHparticles.at(2).COL1=0;    // Outgoing particle with -ve rapidity
	  LHparticles.at(2).COL2=503;
	  LHparticles.at(2).p=p0;
	  LHparticles.at(2).PID=-bptype;
	  LHparticles.at(2).MOT1=1;
	  LHparticles.at(2).MOT2=2;
	  LHparticles.at(2).ISTUP=1;
	  LHparticles.at(3).COL1=501;     // "Extra" parton
	  LHparticles.at(3).COL2=0;
	  LHparticles.at(3).p=p1;
	  LHparticles.at(3).PID=bptype;    
	  LHparticles.at(3).MOT1=1;
	  LHparticles.at(3).MOT2=2;
	  LHparticles.at(3).ISTUP=1;
	  LHparticles.at(4).COL1=503;    // Outgoing parton with +ve rapidity 
	  LHparticles.at(4).COL2=502;
	  LHparticles.at(4).p=p2;
	  LHparticles.at(4).PID=21;
	  LHparticles.at(4).MOT1=1;
	  LHparticles.at(4).MOT2=2;
	  LHparticles.at(4).ISTUP=1;       
	}	
	smatrixuuxthuuxg_(p15[0],&me2);
      }
      else if (pflag==4) {
	if (flags.LHoutput) {
	  LHparticles.at(0).COL1=501;    // Incoming particle with +ve rapidity
	  LHparticles.at(0).COL2=0;
	  LHparticles.at(1).COL1=0;    // Incoming particle with -ve rapidity
	  LHparticles.at(1).COL2=502;
	  LHparticles.at(2).COL1=501;    // Outgoing particle with -ve rapidity
	  LHparticles.at(2).COL2=0;
	  LHparticles.at(2).p=p0;
	  LHparticles.at(2).PID=bptype;
	  LHparticles.at(2).MOT1=1;
	  LHparticles.at(2).MOT2=2;
	  LHparticles.at(2).ISTUP=1;
	  LHparticles.at(3).COL1=503;     // "Extra" parton
	  LHparticles.at(3).COL2=502;
	  LHparticles.at(3).p=p1;
	  LHparticles.at(3).PID=21;    
	  LHparticles.at(3).MOT1=1;
	  LHparticles.at(3).MOT2=2;
	  LHparticles.at(3).ISTUP=1;
	  LHparticles.at(4).COL1=0;    // Outgoing parton with +ve rapidity 
	  LHparticles.at(4).COL2=503;
	  LHparticles.at(4).p=p2;
	  LHparticles.at(4).PID=-bptype;
	  LHparticles.at(4).MOT1=1;
	  LHparticles.at(4).MOT2=2;
	  LHparticles.at(4).ISTUP=1;       
	}	
	smatrixuuxthuuxg_(p16[0],&me2);
      }
    }
    else if (process==2) { // If uuxthggg
      if (flags.LHoutput) {
	LHparticles.at(0).COL1=501;    // Incoming particle with +ve rapidity
	LHparticles.at(0).COL2=0;
	LHparticles.at(1).COL1=0;    // Incoming particle with -ve rapidity
	LHparticles.at(1).COL2=502;
	LHparticles.at(2).COL1=501;    // Outgoing particle with -ve rapidity
	LHparticles.at(2).COL2=503;
	LHparticles.at(2).p=p0;
	LHparticles.at(2).PID=21;
	LHparticles.at(2).MOT1=1;
	LHparticles.at(2).MOT2=2;
	LHparticles.at(2).ISTUP=1;
	LHparticles.at(3).COL1=503;     // "Extra" parton
	LHparticles.at(3).COL2=504;
	LHparticles.at(3).p=p1;
	LHparticles.at(3).PID=21;    
	LHparticles.at(3).MOT1=1;
	LHparticles.at(3).MOT2=2;
	LHparticles.at(3).ISTUP=1;
        LHparticles.at(4).COL1=504;    // Outgoing parton with +ve rapidity 
 	LHparticles.at(4).COL2=502;
	LHparticles.at(4).p=p2;
	LHparticles.at(4).PID=21;
	LHparticles.at(4).MOT1=1;
	LHparticles.at(4).MOT2=2;
	LHparticles.at(4).ISTUP=1;       
      }	     
      smatrixuuxthggg_(p11[0],&me2);
      me2*=6.;
    }
    }
    else {
      cout << "Unrecognised initial state in hjjj non-FKL matching code: "<<aptype <<" "<<bptype<<endl;
      setstatus(16);
    }

  if (flags.LHoutput) {
    particle tparticle; // a temp. particle
    // setup info for Higgs boson
    tparticle.PID=25;// Higgs boson
    tparticle.COL1=0;
    tparticle.COL2=0;
    tparticle.MOT1=1;
    tparticle.MOT2=2;
    tparticle.ISTUP=1;
    tparticle.p=getpHv();
    LHparticles.push_back(tparticle);
  }  

  weight*=pow(Halphas(renscale),3)*pow(Halphas(flags.mH),2)/(256.*shat2*pow(M_PI,5))*me2*pdfta*pdftb;

  if (flags.logcorrect==1)
    weight*=(1.-Halphas(renscale)*beta0*log(partonmomenta.front().perp()/renscale)/(2.*M_PI))*(1.-Halphas(renscale)*beta0*log(partonmomenta.back().perp()/renscale)/(2.*M_PI))*(1.-Halphas(renscale)*beta0*log(partonmomenta.at(1).perp()/renscale)/(2.*M_PI));
  // double tau = pHv.m2()/(4.*mt*mt);
  // double series_t = 1. + tau*(7./30. + tau*(2./21. + tau*26./525.));
  // weight*=series_t*series_t;

  return weight;
}
